#ifndef CHAP_H
#define CHAP_H

#define NUMBER_OF_ALGORITHMS		0x02
#define MD5_ALGORITHM				0x05
#define SHA1_ALGORITHM				0x07

#include "misc_func.h"

typedef struct dateunit
{
	int length;
	char * data;
} DataUnit, * DataUnit_P;

typedef struct chap_context
{
	int number_format;
	int hash_algorithm;
	char * name;
	char * secret;
	unsigned char identifier;
	DataUnit challenge;
	DataUnit response;
} CHAP_Context, * CHAP_Context_P;

extern CHAP_Context * CHAP_InitializeContext(void);

extern CHAP_Context * CHAP_CloneContext(CHAP_Context * ctx);

extern int CHAP_FinalizeContext(CHAP_Context * context);

extern void CHAP_PrintContext(CHAP_Context * context);

extern int CHAP_SetChallengeLength(int clength, 
			CHAP_Context * context);

extern int CHAP_SetNumberFormat(int format,
			CHAP_Context * context);


/*	Searches built-in "hash_algorithms[]" table to find "algorithm".
	If found, sets value in "hash_algorithm" field of "context" and returns 1.
	If not found, returns 0.
*/
extern int CHAP_SetAlgorithm( int algorithm, CHAP_Context * context );

extern int CHAP_SetName(char * name, 
		CHAP_Context * context);

extern char * CHAP_GetName(CHAP_Context * context);

extern int CHAP_SetSecret(char * secret,
		CHAP_Context * context);

extern int CHAP_SetIdentifier(unsigned char identifier, 
		CHAP_Context * context);

extern unsigned char CHAP_GetIdentifier(CHAP_Context * context);

extern char * CHAP_GetChallenge(CHAP_Context * context);

extern char * CHAP_GetResponse(unsigned char identifier, 
			char * challenge, 
			CHAP_Context * context);

extern int CHAP_CheckResponse(char * response, 
		CHAP_Context * context);


/*	Searches "list" to find first algorithm number we support.
	If found, return than number.
	If not found, return -1.
*/
extern int CHAP_SelectAlgorithm( char *list );


/*	Builds list of available CHAP hash algorithms as a character string
	starting at "list".  Returns strlen(final list).
*/
extern int CHAP_GetAlgorithmList( char *list );

#endif

