/*
 * IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING. By downloading, copying, installing or
 * using the software you agree to this license. If you do not agree to this license, do not download, install,
 * copy or use the software. 
 *
 * Intel License Agreement 
 *
 * Copyright (c) 2000, Intel Corporation
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 * the following conditions are met: 
 *
 * -Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *  following disclaimer. 
 *
 * -Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *  following disclaimer in the documentation and/or other materials provided with the distribution. 
 *
 * -The name of Intel Corporation may not be used to endorse or promote products derived from this software
 *  without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef _DEBUG_H
#define _DEBUG_H

#ifdef __KERNEL__
#else
#include <stdio.h>
#endif

/*
 * Debugging Support
 */

#define TRACE_DEBUG          0x00000001
#define TRACE_ISCSI          0x00000004
#define TRACE_NET            0x00000008
#define TRACE_BUFF           0x00000010
#define TRACE_FILE           0x00000020
#define TRACE_LINE           0x00000040
#define TRACE_SEM            0x00000080
#define TRACE_ENTER_LEAVE    0x00000100
#define TRACE_ISCSI_FULL     0x00000200
#define TRACE_MY_MEMORY      0x00000400
#define TRACE_ALL            0xffffffff
#define TRACE_ENDING         0x00000000

static char iscsi_trace_buff[4096];

#ifdef __KERNEL__
#define PRINT printk
#else
#define PRINT printf
#endif

#ifdef CONFIG_ISCSI_DEBUG
static unsigned iscsi_trace =
				/*** TRACE_DEBUG | ***/
				TRACE_ISCSI |
				/*** TRACE_NET | ***/
				/*** TRACE_BUFF | ***/
				/*** TRACE_FILE | ***/
				/*** TRACE_LINE | ***/
				/*** TRACE_SEM | ***/
				/*** TRACE_ENTER_LEAVE | ***/
				/*** TRACE_ISCSI_FULL |  ***/
				/*** TRACE_MY_MEMORY | ***/
				TRACE_ENDING;

#define TRACE(trace, args...)                                      \
do {								   \
if (iscsi_trace&trace) {                                           \
    sprintf(iscsi_trace_buff, args);                               \
  if (iscsi_trace&TRACE_FILE) {                                    \
    PRINT("%s::%d: %s", __FILE__, __LINE__,                        \
          iscsi_trace_buff);                                       \
  } else if (iscsi_trace & TRACE_LINE) {			   \
    PRINT("%d: %s", __LINE__, iscsi_trace_buff);                   \
  } else {                                                         \
    PRINT("%s", iscsi_trace_buff);                                 \
  }                                                                \
}								   \
   } while(0)

#define PRINT_BUFF(buff, len)                      \
  if(iscsi_trace&TRACE_BUFF) {                     \
    int zzz;                                       \
                                                   \
    PRINT("%d: \n", __LINE__);                     \
    for (zzz=0;zzz<len;zzz++) {                    \
      if (zzz%16==0) {                             \
        if (zzz) PRINT("\n");                      \
        PRINT("%4i: ", zzz);                       \
      }                                            \
      PRINT("%02x ",((unsigned char *)(buff))[zzz]); \
    }                                              \
    if ((len+1)%16)  PRINT("\n");                  \
  }

#define TRACE_TEST(trace)			   \
   (iscsi_trace&trace)
#else
#define TRACE(trace, args...)
#define PRINT_BUFF(buff, len)
#define TRACE_TEST(trace)		(0)
#endif

#define TRACE_ERROR(args...)                     \
sprintf(iscsi_trace_buff, args);                 \
PRINT("%s:%s:%i: ***ERROR*** %s",                \
       __FILE__, __FUNCTION__, __LINE__,         \
       iscsi_trace_buff);

#endif // _DEBUG_H
