/*	initiator_negotiate.h */

/*	The code for redirection during login is based on code first written
	by Scott Cranston of Equal Logic (scranston@equallogic.com).
*/

#include "../iscsi-unh-common/chap.h"
#include "../iscsi-unh-common/misc_func.h"

/* check the correctness of target ip address which should be in format of 
 TargetAddress=domainname[:port][,portal-group-tag] */
int check_target_address( char *value, int when)
{
	int retval = 0;	
	//if( when 	

	return retval;
}



/* when get here, we already know that opcode is login response */
/* returns 1 if valid redirection, 0 if other success, -1 if any error */
int check_login_response( unsigned int draft_number,
						struct generic_pdu *inputpdu,
						struct generic_pdu *outputpdu )
{
  int retval = 0;
  struct iscsi_targ_login_rsp	*login_rsp;
  struct iscsi_init_login_cmnd	*login_req;
  char *tsix;


  TRACE(TRACE_ENTER_LEAVE, "Enter check_login_response\n");

  if( draft_number < 12 )
    tsix = "TSID";
  else
    tsix = "TSIH";

  login_rsp = (struct iscsi_targ_login_rsp *)inputpdu;
  login_req = (struct iscsi_init_login_cmnd *)outputpdu;

  /* check that reserved fields in the login response are all zero */
  if(	((draft_number >= DRAFT13)  &&			/* draft 13 or after */
		( (login_rsp->opcode & OLD_X_BIT)  ||
			(login_rsp->flags & (W_BIT | BRO_BIT))  ||
			login_rsp->rsvd1 || login_rsp->rsvd2 || login_rsp->rsvd3 ))
    ||
 		((draft_number<DRAFT13)&&(draft_number>=DRAFT9)&& /* draft 9 or after */
		( (login_rsp->opcode & OLD_X_BIT)  ||
			(login_rsp->flags & (R_BIT | W_BIT | BRO_BIT))  ||
			login_rsp->rsvd1 || login_rsp->rsvd2 || login_rsp->rsvd3 ))
    ||
    	((draft_number <= DRAFT8)  &&			/* draft 8 or before */
		( (login_rsp->flags & (R_BIT | W_BIT | BRO_BIT))  ||
			login_rsp->isid[0] || login_rsp->isid[1] ||
			login_rsp->isid[2] || login_rsp->isid[3] ||
			login_rsp->rsvd1 || login_rsp->rsvd2 || login_rsp->rsvd3 ))
    )
    {
      TRACE_ERROR("Reserved fields in Login Response header not all 0\n");
    }

  if( (inputpdu->status_class != STAT_CLASS_SUCCESS) ) 
    {/* the target sent a login reject */

      TRACE(TRACE_ISCSI, "Login Response with non-zero StatusClass 0x%02x, "
      "StatusDetail 0x%02x\n", inputpdu->status_class, inputpdu->status_detail);

      /* Draft 18, Section 4.3.1 Login Phase Start
		 "-Login Response with Login Reject.  This is an immediate rejection
		 from the target that causes the connection to terminate and the session
		 to terminate if this is the first (or only) connection of a new
		 session.  The T bit and the CSG and NSG fields are reserved."
      */

      /* Draft 9, Section 5.1 Login Phase Start
		 "-Login Response with Login Reject.  This is an immediate rejection
		 from the target that causes the session to terminate.  The T bit of
		 the response MUST be set to 1 and the CSG and NSG are reserved."
      */

      if( (draft_number <= DRAFT9  &&  !(inputpdu->flags & T_BIT))  ||
      	  (draft_number >  DRAFT9  &&  (inputpdu->flags & T_BIT))  ||
		  (inputpdu->flags & CSG) != 0  ||  (inputpdu->flags & NSG) != 0 ) 
		{
		  /* protocol error */
		  TRACE_ERROR("Non-zero Status-Class MUST have T=%d, CSG=0 & NSG=0\n",
							  draft_number <= DRAFT9);
		  retval = -1;
		}

	  if( inputpdu->status_class == STAT_CLASS_REDIRECTION )
	  	{/* target is redirecting login to different IP address/Port number */
		  if( inputpdu->status_detail < STAT_DETAIL_TARG_MOVED_TEMP  &&
		  	  inputpdu->status_detail > STAT_DETAIL_TARG_MOVED_PERM )
			{
			  TRACE_ERROR("Invalid Status Detail 0x%02x for Redirection\n",
													inputpdu->status_detail);
			  retval = -1;
			}
		  else
		  	{/* valid redirection detail, there should be a TargetAddress key
				attached to this PDU that contains the redirection info */
			  retval = 1;
			}
		}
	  else if( inputpdu->status_class == STAT_CLASS_INITIATOR )
		{/* Initiator Error (not a format error) */	
		  if( inputpdu->status_detail > STAT_DETAIL_INVALID_DURING_LOGIN )
			  TRACE_ERROR("Invalid Status Detail 0x%02x for Initiator Error\n",
													inputpdu->status_detail);
		  retval = -1;
		}
	  else if( inputpdu->status_class == STAT_CLASS_TARGET )
		{/* Target Error */	
		  if( inputpdu->status_detail > STAT_DETAIL_OUT_OF_RESOURCE )
			  TRACE_ERROR("Invalid Status Detail 0x%02x for Target Error\n",
													inputpdu->status_detail);
		  retval = -1;
		}
	  else
		{/* Bad status class */	
		  TRACE_ERROR("Invalid Status Class 0x%02x\n", inputpdu->status_class);
		  retval = -1;
		}
    }

  /* status_class is 0, status_detail should also be 0 */
  else if( (inputpdu->status_detail != 0x0) )
    {
      TRACE_ERROR("Login Response with zero StatusClass but non-zero "
			      "StatusDetail 0x%02x\n", inputpdu->status_detail);
      retval = -1;
    }

  else if( (inputpdu->flags & CSG) != (outputpdu->flags & CSG) )
    {/* target's state not the same as initiator's state */
      /* protocol error */
      TRACE_ERROR("State inconsistant! Target in state %d, initiator in %d\n",
		  (inputpdu->flags & CSG)>>CSG_SHIFT,
		  (outputpdu->flags & CSG)>>CSG_SHIFT);
      retval = -1;
    }

  else if( (draft_number <= DRAFT8  &&
			  memcmp(login_rsp->isid, login_req->isid, 2) != 0)  ||
    	   (draft_number > DRAFT8  &&
			  memcmp(login_rsp->isid, login_req->isid, 6) != 0) )
    {/* target changed the isid */
      /* protocol error */
      TRACE_ERROR("Target changed the isid\n");
      retval = -1;
    }

  else if( !(inputpdu->flags & T_BIT) )
    {/* target does not want to change stage */
      if( inputpdu->flags & NSG )
		{
		  /* check reserved field, but only give warning if errors are caught */
		  TRACE(TRACE_ISCSI,
						"***Warning***: NSG %d should be 0 when T bit is 0\n",
		  				inputpdu->flags & NSG);
		}
    }

  else
    {/* input T bit is set to 1, target wants to change stage */
      if( !(outputpdu->flags & T_BIT) )
		{/* but initiator did not want to change stage yet */
		  /* protocol error */
		  TRACE_ERROR("Login Response with T bit set but Initiator's T bit \
							not set \n");
		  retval = -1;	 
		}

      else if( (inputpdu->flags & NSG) <= ((inputpdu->flags & CSG) >> CSG_SHIFT)
				      || (inputpdu->flags & NSG) == NSG2 )
		{/* target wants to go to an illegal stage */
		  /* protocol error */
		  TRACE_ERROR("Target trying to change to illegal stage %d\n",
				  inputpdu->flags & NSG );
		  retval = -1;
		}     

      else if( (inputpdu->flags & NSG) == NSG3 )
		{/* target going to FFP, can only do this if initiator asked for it */
		  /*Draft 18, Section 4.3 "Login Phase"
			"The Login Final-Response that accepts a Login Command can only
			come as a response to a Login request with the T bit set to 1,
			and both the request and response MUST indicate FullFeaturePhase
			as the next phase via the NSG field."
		  */
		  if( (outputpdu->flags & NSG) != NSG3 )
			{
			  TRACE_ERROR("Target trying to change to FFP but initiator wanted "
					  "change to Operational stage only\n");
			  retval = -1;
			}
		  else if( login_rsp->tsih == 0 )
			{/* protocol error */
			  TRACE_ERROR("Target's final login response has %s = 0\n", tsix);
			  retval = -1;
			}
		}
      else if( login_req->tsih != login_rsp->tsih )
        {
		  TRACE_ERROR("Target changed initiator's %s from %u to %u\n", tsix,
			be16_to_cpu(login_req->tsih), be16_to_cpu(login_rsp->tsih));
		}
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave check_login_response, retval = %d\n", retval);
  return retval;
}


/*	Called to send a login already set up in outputpdu
	and then to read back a login response into inputpdu */
/*	Returns 1 on redirection, 0 on other success, -1 on any error */
static int write_and_read( struct socket *sock, unsigned int draft_number,
					struct generic_pdu *inputpdu,
					struct generic_pdu *outputpdu,
					int *target_version_max,
					int *target_version_active,
					unsigned int flags )
{
  int			padding = 0;
  int			retval = 0;
  unsigned int	size;


  TRACE(TRACE_ENTER_LEAVE, "Enter write_and_read\n");

  /* send output */
  if( iscsi_send_msg(sock, draft_number, outputpdu, flags) < 0 )
    {
      TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
      return -1;
    }

  /* After sending the outputpdu, reset it's text len to 0 */
  outputpdu->text_length = 0;

  /* wait for input from target */
  size = ISCSI_HDR_LEN;

  /* NOTE: in draft 9 and after, no packets sent or received during login phase
		will have digests of any kind.
		However, in draft 8 and before, packets sent after the security
		negotiation stage may have digests.
  */
  if( flags & USE_HEADERDIGEST )
  	size += CRC_LEN;

  if( iscsi_recv_msg(sock, size, (char *)inputpdu, flags) < 0 )
    {
      TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
      return -1;
    }

  if( flags & USE_HEADERDIGEST )
  	{/* header digests are in use, compute the crc and check it */
	do_crc((unsigned char *)inputpdu, ISCSI_HDR_LEN, 3, &size);
	if( inputpdu->header_digest != size )
		{/* crc check should have been == and was not */
		TRACE_ERROR("Header CRC wrong: got 0x%08x, expected 0x%08x\n",
								ntohl(inputpdu->header_digest), ntohl(size));
		}
	}

  /* Also get the text associated with it */
  inputpdu->text_length = be32_to_cpu(inputpdu->length);

  /* Include the padding bytes too */
  padding = inputpdu->text_length + ((-inputpdu->text_length) & 3);

  if( inputpdu->text_length > MAX_TEXT_LEN )
    {
      TRACE_ERROR("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
										  inputpdu->text_length, MAX_TEXT_LEN);
      return -1;
    }

  if( padding > 0  &&  (flags & USE_DATADIGEST) )
  	{/* data digests are in use */
	padding += CRC_LEN;
	}

  if( padding > 0 && iscsi_recv_msg(sock, padding, inputpdu->text, flags) < 0 )
	{
	  TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
	  return -1;
	}

  if( padding > 0  &&  (flags & USE_DATADIGEST) )
  	{/* data digests are in use, compute and check it for this data */
	do_crc(inputpdu->text, padding-CRC_LEN, 3, &size);
	if( size != *(unsigned int *)(&inputpdu->text[padding-CRC_LEN]) )
		{
		TRACE_ERROR("Data CRC wrong: got 0x%08x, expected 0x%08x\n",
					ntohl(*(unsigned int *)(&inputpdu->text[padding-CRC_LEN])),
					ntohl(size));
		}
	}

  /* the I bit has to be set in all response PDUs for drafts 9 and before */
  if( draft_number <= DRAFT9  &&  !(inputpdu->opcode & I_BIT) )
    {
      TRACE_ERROR("I_BIT not set in response PDU\n");
    }

  /* the X bit has to be set in all response PDUs for drafts 8 and before */
  if( draft_number <= DRAFT8  &&  !(inputpdu->opcode & OLD_X_BIT) )
  	{
	  TRACE_ERROR("X_BIT not set in response PDU\n");
	}

  if( (inputpdu->opcode & ISCSI_OPCODE) == ISCSI_TARG_LOGIN_RSP )
    {/* as expected, this is a login response, check it */
	  TRACE(TRACE_ISCSI, "Got Login Rsp, ITT %u\n",
	  											ntohl(inputpdu->init_task_tag));
	  if( TRACE_TEST(TRACE_ISCSI_FULL) )
      	print_targ_login_rsp((struct iscsi_targ_login_rsp *)inputpdu);

  	  /* Check the initiator task tag */
  	  if( inputpdu->init_task_tag != outputpdu->init_task_tag )
		{
		  TRACE(TRACE_ISCSI,"***Warning***: received init_task_tag = %d, "
				"expected %d\n",
				 be32_to_cpu(inputpdu->init_task_tag),
				 be32_to_cpu(outputpdu->init_task_tag));
		}

  	  /* Check received version_max and version_active fields for consistency */
  	  if( inputpdu->version_max < inputpdu->version_active )
		{
		  TRACE(TRACE_ISCSI, "***Warning***: received version_max = 0x%02x "
							"less than received version_active 0x%02x\n",
							inputpdu->version_max,
							inputpdu->version_active);
		}

	  /* The received version_max and version_active fields must not change */

	  /* Draft 18, Section 9.13.1 Version-max (for Login Response)
	  	 "All Login responses within the Login Phase MUST carry the same
		 Version-max.
		 The initiator MUST use the value presented as a response to the
		 first login request."

		 Draft 18, Section 9.13.2 Version-active (for Login Response)
		 "All Login responses within the Login Phase MUST carry the same
		 Version-active.
		 The initiator MUST use the value presented as a response to the
		 first login request."
	  */
	  if( *target_version_max < 0 )
		{/* first login response, just remember its values */
	  	  *target_version_max = inputpdu->version_max;
	  	  *target_version_active = inputpdu->version_active;
		}
	  else if( *target_version_max != inputpdu->version_max )
		{
		  TRACE(TRACE_ISCSI, "***Warning***: received version_max = 0x%02x "
					"differs from previously received version_max 0x%02x\n",
					inputpdu->version_max, *target_version_max);
		}
	  else if( *target_version_active != inputpdu->version_active )
		{
		  TRACE(TRACE_ISCSI, "***Warning***: received version_active = 0x%02x "
					"differs from previously received version_active 0x%02x\n",
					inputpdu->version_active, *target_version_active);
		}

  	  /* Check the version_active field to see if we can use it */
  	  /* For now, we send version_min == version_max, so target has no choice */
  	  if( inputpdu->version_active != outputpdu->version_active )
		{
		  TRACE_ERROR("received version_active = 0x%02x, expected 0x%02x\n",
				 inputpdu->version_active, outputpdu->version_active);
		  retval = -1;
		}

	  /* Check the C bit, which we do not use yet */
	  else if( draft_number >= DRAFT13  &&  inputpdu->flags & C_BIT )
	    {/* C bit is set in Login Response */
		  /* Draft 18, Section 9.13.7 C (Continue) Bit
			 "A Login Response with the C bit set to 1 must have the
			 T bit set to 0."
		  */
		  if( inputpdu->flags & T_BIT )
		  	{
		  	TRACE_ERROR("C=1 in Login Response requires T=0, not T=1\n");
			}
		  else
		  	{/* for now */
			TRACE_ERROR("C=1 in Login Response not implemented yet\n");
			}
		  retval = -1;
		}
	  else
		retval = check_login_response(draft_number, inputpdu, outputpdu);
    }

  else if( (inputpdu->opcode & ISCSI_OPCODE) == ISCSI_TARG_RJT )
    {
      /* protocol error */
      TRACE_ERROR("got Reject PDU\n");
      print_targ_rjt( (struct iscsi_targ_rjt *)inputpdu );
      retval = -1;
    }
  else
  	{
	  TRACE_ERROR("received opcode = 0x%02x, expected 0x%02x\n",
  						inputpdu->opcode & ISCSI_OPCODE, ISCSI_TARG_LOGIN_RSP);
	  retval = -1;		/* this is fatal */
	}

  TRACE(TRACE_ENTER_LEAVE, "Leave write_and_read, retval %d\n", retval);

  return retval;
}



/*	When called, 1st login PDU is set up in outputpdu except for T, NSG, CSG */
/*	Returns 1 for redirection, 0 for all other successes, -1 for any error */
static int initiator_security_phase( struct socket *sock,  
					struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
					struct chap_parameter_type p_chap_param,
					__u32 *cur_cmd_sn,
					int *exp_stat_sn,
					int *max_send_length,
					int noperational,
					unsigned int	draft_number,
					unsigned int	who_called,
					struct generic_pdu *inputpdu,
					struct generic_pdu *outputpdu,
					int * target_version_max,
					int * target_version_active,
					unsigned int connection_flags,
					unsigned long long *login_flags,
					struct unknown_key ** unknown_key_list)
	{
	int add_length = 0;
	int retval = -1;			/* pessimistic assumption that we will fail */
	char * dummy_string;
	struct parameter_type * auth_p;
	int count;
	int temp_len;
	char * temp_str;
	unsigned char chap_i = 0;
	char * his_chap_c = NULL;
	char * our_chap_c = NULL;
	char * his_chap_r = NULL;
	char * our_chap_r = NULL;
	char * our_chap_n = NULL;
	int int_val;
	int neg_flag = 0;
	CHAP_Context * init_ctx = NULL;
	CHAP_Context * targ_ctx = NULL;
	int	security_state = 0;

	struct unknown_key	*key;
	

	TRACE(TRACE_ENTER_LEAVE, "Enter initiator_security_phase\n");
	TRACE(TRACE_DEBUG, "initiator_security_phase draft_number %u\n", draft_number);
	ALLOCATE_MAX_TEXT_LEN(dummy_string);

	/*	offer no transition on first login pdu sent to target */
	outputpdu->flags &= ~(CSG | NSG | T_BIT);
	
	add_length = scan_table_and_process(sock,
					p_param_tbl,
					SECURITY_PARAM | INFORMATIONAL_PARAM,
					INITIATOR,
					draft_number,
					inputpdu,
					outputpdu,
					connection_flags,
					login_flags);
	if( add_length < 0 )
		{
		retval = add_length;
		goto out;
		}
	outputpdu->text_length = add_length;

	for( count = 0;  count < LOOP_TIMES;  count++ )
		{
		TRACE(TRACE_DEBUG, "Send security login, key length %d\n", add_length);
		outputpdu->cmd_sn = cpu_to_be32(*cur_cmd_sn);
		TRACE(TRACE_ISCSI, "Send Login, ITT %u\n",
											ntohl(outputpdu->init_task_tag));

		if( write_and_read(sock, draft_number, inputpdu, outputpdu, 
			target_version_max, target_version_active, connection_flags) < 0 )
			{
			goto out;
			}

		if( inputpdu->status_class != 0x00 )
			{
			TRACE_ERROR(
					"Login Reject from target, Class 0x%02x, Detail 0x%02x\n",
					inputpdu->status_class, inputpdu->status_detail);
			goto out;
			}

		*exp_stat_sn = be32_to_cpu(((struct iscsi_targ_login_rsp *)
														inputpdu)->stat_sn)+1;
		outputpdu->exp_stat_sn = cpu_to_be32(*exp_stat_sn);

		/*	assume we will want no transition on our next output pdu */
		outputpdu->flags &= ~(CSG | NSG | T_BIT);

		add_length = scan_input_and_process(sock,
						p_param_tbl,
						SECURITY_PARAM | INFORMATIONAL_PARAM,
						TARGETPORTALGROUPTAG_FLAG,
						INITIATOR,
						max_send_length,
						draft_number,
						who_called,
						inputpdu,
						outputpdu,
						connection_flags,
						login_flags,
						unknown_key_list);	

		if( add_length < 0 )
			{
			retval = add_length;
			goto out;
			}
		outputpdu->text_length = add_length;

		if( security_state == 0 )
			{/* looking to find AuthMethod=CHAP */

			/* deal with any unknown keys: CHAP keys illegal in this state */
			for( key = *unknown_key_list;  key != NULL;  key = key->next )
				{
				if( !key->processed )
					{
					key->processed = 1;
					TRACE_ERROR("key \"%s\" not allowed now\n", key->keyname);
					goto out;
					}
				}
			
			if( (auth_p = find_flag_parameter(AUTHMETHOD_FLAG, p_param_tbl))
																	== NULL )
				{/* should NEVER happen */
				TRACE_ERROR("AuthMethod parameter not found\n");
				goto out;
				}

			if( !IS_KEY_GOT_FROM_OTHER_SIDE(auth_p->neg_info) )
				{/* AuthMethod key not negotiated yet, stay in state 0 */
				continue;
				}
			else if( !strcmp(auth_p->str_value, NONE) )
				{/* not going to do authentication, offer transition */
				security_state = 3;
				if( noperational == 0 )
					outputpdu->flags |= (NSG3|T_BIT);
				else
					outputpdu->flags |= (NSG1|T_BIT);
				}
			else if( !strcmp(auth_p->str_value, CHAP) )
				{/* start CHAP authentication */
				TRACE(TRACE_DEBUG, "Start initiator chap authentication\n");
				init_ctx = p_chap_param.local_ctx;
				targ_ctx = p_chap_param.peer_ctx;
				if( init_ctx == NULL  ||  targ_ctx == NULL )
					{
					TRACE_ERROR("context not initialized yet\n");
					goto out;
					}
				CHAP_SetAlgorithm(MD5_ALGORITHM, targ_ctx);
				CHAP_SetAlgorithm(MD5_ALGORITHM, init_ctx);

				temp_len = sprintf(dummy_string, "CHAP_A=");
				temp_len += CHAP_GetAlgorithmList(dummy_string+temp_len);
				TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
				strcpy(outputpdu->text+outputpdu->text_length, dummy_string);
				outputpdu->text_length += temp_len + 1;

				security_state = 1;
				}
			else
				{/* should NEVER happen */
				TRACE_ERROR("Unknown AuthMethod value \"%s\"\n",
															auth_p->str_value);
				goto out;
				}
			}
		else if( security_state == 1 )
			{/* previously sent CHAP_A list, expecting CHAP_A, CHAP_I, CHAP_C */
			for( key = *unknown_key_list;  key != NULL;  key = key->next )
				{
				if( !key->processed )
					{
					key->processed = 1;
					if( !strcmp(key->keyname, "CHAP_A") )
						{
						if( neg_flag & 0x01 )
							{
							TRACE_ERROR("Duplicate CHAP_A key, value \"%s\"\n",
																key->keyvalue);
							goto out;
							}
						TRACE(TRACE_DEBUG, "got chap_a: %s\n", key->keyvalue);
						neg_flag |= 0x01;
						int_val=(int)simple_strtoul(key->keyvalue,&temp_str,0);
						if( !(isdigit(*key->keyvalue)) || *temp_str != '\0' ||
										!CHAP_SetAlgorithm(int_val, targ_ctx) ||
										!CHAP_SetAlgorithm(int_val, init_ctx) )
							{
							TRACE_ERROR("Illegal chap_a algorithm \"%s\"\n",
																key->keyvalue);
							goto out;
							}
						}
					else if( !strcmp(key->keyname, "CHAP_I") )
						{
						if( neg_flag & 0x02 )
							{
							TRACE_ERROR("Duplicate CHAP_I key, value \"%s\"\n",
																key->keyvalue);
							goto out;
							}
						TRACE(TRACE_DEBUG, "got chap_i: %s\n", key->keyvalue);
						neg_flag |= 0x02;
						int_val=(int)simple_strtoul(key->keyvalue,&temp_str,0);
						if( !(isdigit(*key->keyvalue)) || *temp_str != '\0' ||
															(int_val > 255) )
							{
							TRACE_ERROR("Illegal chap_i number \"%s\"\n",
																key->keyvalue);
							goto out;
							}
						chap_i = (unsigned char) int_val;
						}
					else if( !strcmp(key->keyname, "CHAP_C") )
						{
						if( neg_flag & 0x04 )
							{
							TRACE_ERROR("Duplicate CHAP_C key, value \"%s\"\n",
																key->keyvalue);
							goto out;
							}
						TRACE(TRACE_DEBUG, "got chap_c: %s\n", key->keyvalue);
						neg_flag |= 0x04;
						his_chap_c = key->keyvalue;
						}
					else
						{
						TRACE_ERROR("Unknown key \"%s\" in security phase\n",
																key->keyname);
						goto out;
						}
					}
				}
			if( neg_flag == 7 )
				{/* have all target keys needed to respond to his challenge */
				our_chap_n = CHAP_GetName(targ_ctx);
				if( our_chap_n == NULL )
					{
					TRACE_ERROR("CHAP_N to Target not configured\n");
					goto out;
					}
				our_chap_r = CHAP_GetResponse(chap_i, his_chap_c, targ_ctx);
				if( our_chap_r == NULL )
					{
					TRACE_ERROR("CHAP_R to Target cannot be generated\n");
					goto out;
					}
				temp_len = sprintf(dummy_string, "CHAP_N=%s", our_chap_n);
				TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
				add_length = (temp_len + 1);
				temp_len = sprintf(dummy_string+add_length, "CHAP_R=%s",
																	our_chap_r);
				TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string+add_length);
				add_length += (temp_len + 1);
				if( p_chap_param.chap_flags & USE_TARGET_CONFIRMATION )
					{
					chap_i = CHAP_GetIdentifier(init_ctx);
					our_chap_c = CHAP_GetChallenge(init_ctx);
					if( our_chap_c == NULL )
						{
						TRACE_ERROR("Second exchange of chap failed\n");
						goto out;
						}
					temp_len = sprintf(dummy_string+add_length, "CHAP_I=%d",
																		chap_i);
					TRACE(TRACE_ISCSI, "Attach key: %s\n",
													dummy_string+add_length);
					add_length += (temp_len + 1);
					temp_len = sprintf(dummy_string+add_length, "CHAP_C=%s",
																	our_chap_c);
					TRACE(TRACE_ISCSI, "Attach key: %s\n",
													dummy_string+add_length);
					add_length += (temp_len + 1);	

					security_state = 2;	/* expect target reply to challenge */
					}
				else
					{
					security_state = 3;		/* we are not challenging target */
					}

				memcpy(outputpdu->text+outputpdu->text_length, dummy_string,
																	add_length);
				outputpdu->text_length += add_length;

				/*	regardless of whether we challenge target or not,
					this is the last pdu we need to send, so offer transition */
				if( noperational == 0 )
					outputpdu->flags |= (NSG3|T_BIT);
				else
					outputpdu->flags |= (NSG1|T_BIT);
				}
			}
		else if( security_state == 2 )
			{/* expecting target reply to our previously sent challenge */
			for( key = *unknown_key_list;  key != NULL;  key = key->next )
				{
				if ( !key->processed )
					{
					key->processed = 1;
					if( !strcmp(key->keyname, "CHAP_N") )
						{
						if( neg_flag & 0x10 )
							{
							TRACE_ERROR("Duplicate CHAP_N key, value \"%s\"\n",
																key->keyvalue);
							goto out;
							}
						neg_flag |= 0x10;
						if( strcmp(key->keyvalue, init_ctx->name) )
							{
							TRACE_ERROR("chap_n value not as expected\n");
							goto out;
							}
						}
					else if( !strcmp(key->keyname, "CHAP_R") )
						{
						if( neg_flag & 0x20 )
							{
							TRACE_ERROR("Duplicate CHAP_R key, value \"%s\"\n",
																key->keyvalue);
							goto out;
							}
						neg_flag |= 0x20;
						his_chap_r = key->keyvalue;
						}
					else
						{
						TRACE_ERROR("Invalid key in security phase : %s\n",
																key->keyname);
						goto out;
						}
					}
				}
			if( neg_flag == 0x37 )
				{/*	target sent complete reply to our challenge, check it now */
				if( CHAP_CheckResponse(his_chap_r, init_ctx) <= 0 )
					{
					TRACE_ERROR("CHAP failed to authenticate target\n");
					goto out;
					}

				TRACE(TRACE_ISCSI, "CHAP successfully authenticated target\n");

				if( inputpdu->flags & T_BIT )
					{/* target agreed to transition out of security stage */
					break;
					}

				TRACE(TRACE_ISCSI, "end up the security phase\n");
				security_state = 3;
				if( noperational == 0 )
					outputpdu->flags |= (NSG3 | T_BIT);
				else
					outputpdu->flags |= (NSG1 | T_BIT);
				}
			}
		else if( security_state == 3 )
			{/* done with authentication (may never have done any) */

			for( key = *unknown_key_list;  key != NULL;  key = key->next )
			{
				if( !key->processed )
				{
					key->processed = 1;
					TRACE_ERROR("key %s not allowed now\n", key->keyname);
					goto out;
				}
			}

			if( inputpdu->flags & T_BIT )
				{/* target agreed to transition out of security stage */
				break;
				}

			TRACE(TRACE_ISCSI, "end up the security phase\n");
			if( noperational == 0 )
				outputpdu->flags |= (NSG3|T_BIT);
			else
				outputpdu->flags |= (NSG1|T_BIT);
			}
		else
			{/* should NEVER happen */
			TRACE_ERROR("Unknown security state %d\n", security_state);
			goto out;
			}
		}

	if( count >= LOOP_TIMES )
		{
		TRACE_ERROR("Infinite loop in security phase negotiations\n");
		}
	else
		retval = 0;			/* our pessimism was unjustified, we succeeded */

out:
	my_kfree((void **)&our_chap_n, "CHAP Get Name");
	my_kfree((void **)&our_chap_r, "CHAP Response");
	my_kfree((void **)&our_chap_c, "CHAP Challenge");

	FREE_STRING(dummy_string);

	TRACE(TRACE_ENTER_LEAVE, "Leave initiator_security_phase, retval %d\n",
																		retval);
	return retval;
	}



/*	Returns 1 for redirection, 0 for all other successes, -1 for any error */
static int initiator_operational_phase( struct socket *sock,  
				 	struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
					__u32 *cur_cmd_sn,
					int *exp_stat_sn,
					int *max_send_length,
					unsigned int draft_number,
					unsigned int when_called,
					struct generic_pdu *inputpdu,
					struct generic_pdu *outputpdu,
					int *target_version_max,
					int *target_version_active,
					unsigned int connection_flags,
					unsigned long long *login_flags,
					struct unknown_key **unknown_key_list )
	{
	int		add_length = 0;
	int		count, retval = 0;


	TRACE(TRACE_ENTER_LEAVE, "Entering initiator_operational_phase\n");

	count = 0;
	do
		{
		TRACE(TRACE_DEBUG, "count %d, inputpdu->T_BIT 0x%02x\n",
											count, inputpdu->flags & T_BIT);

		/* assume there is only one message is to be sent in this phase */
		outputpdu->flags &= (~CSG|NSG);  
		outputpdu->flags |= CSG1|T_BIT|NSG3;  /* set T=1, CSG=1, NSG=3 */

		if( (*login_flags & MORE_TO_SEND_FLAG)  &&  count < 3 )
			{	
			/* set output.T = 0 */
			outputpdu->flags &= ~(T_BIT);
			}

		*login_flags &= ~MORE_TO_SEND_FLAG;

		/* scan through our table and process any unsent O or I keys */
		/*		-- if any unsent S keys then protocol error */

		add_length = scan_table_and_process( sock,
								p_param_tbl, 
								OPERATIONAL_PARAM | INFORMATIONAL_PARAM,
								INITIATOR,
								draft_number,
								inputpdu,
								outputpdu,
								connection_flags,
								login_flags);
		if( add_length < 0 )
			{
			retval = add_length;
			goto out;
			}

		outputpdu->text_length += add_length;

		/*	Draft 18, Section 9.12.8 CmdSN (for Login Request):
			"CmdSN is either the initial command sequence number of a session
			(for the first Login request of a session - the "leading" login),
			or the command sequence number in the command stream if the login
			is for a new connection in an existing session."
		*/

		/*	This means we always have to pick up the session's CmdSN and
			put it into the pdu just as we are about to send it on the wire.
			However, we do not advance it during login phase.
		*/
		outputpdu->cmd_sn = cpu_to_be32(*cur_cmd_sn);

		TRACE(TRACE_ISCSI, "Send Login, ITT %u\n",
  											ntohl(outputpdu->init_task_tag));
		if( (retval = write_and_read(sock, draft_number, inputpdu, outputpdu,
							target_version_max, target_version_active,
							connection_flags)) < 0 )
			{
			goto out;
			}

		/* pass back the exp_stat_sn from the last login response we get */
		/* exp_stat_sn is always 1 greater than last statsn received */
		*exp_stat_sn = 1 +
				be32_to_cpu(((struct iscsi_targ_login_rsp *)inputpdu)->stat_sn);

		/* set up the reply (if any) to have the correct ExpStatSN */
		outputpdu->exp_stat_sn = cpu_to_be32(*exp_stat_sn);

		/* process each key attached to the input packet */
		add_length = scan_input_and_process(sock, 
							p_param_tbl, 
							OPERATIONAL_PARAM | INFORMATIONAL_PARAM,
							TARGETPORTALGROUPTAG_FLAG,
							INITIATOR,
							max_send_length,
							draft_number,
							when_called,
							inputpdu,
							outputpdu,
							connection_flags,
							login_flags,
							unknown_key_list);
		if( add_length < 0 )
			{
			retval = add_length;
			goto out;
			}
		outputpdu->text_length = add_length;

		/* see if Login Request was a redirect */
		if( retval == STAT_CLASS_REDIRECTION )
			{
			goto out;
			}

		if( *login_flags & FIRST_FLAG )
			{/* this was first Login Request/Login Response pair */
			if( !(*login_flags & TARGETPORTALGROUPTAG_FLAG)  )
				{
					if( draft_number == DRAFT12 )
					{/* draft 12, 4.3.1, TargetPortalGroupTag SHOULD be set on first login response*/
		  				TRACE(TRACE_ISCSI,"***Warning***: \"TargetPortalGroupTag\" "
						"key not set on first Login Response from target\n");
					}
					else if( draft_number >= DRAFT18 )
					{/* draft 18, 4.3.1, TargetPortalGroupTag MUST be set on first login response*/
						TRACE(TRACE_ISCSI,"\"TargetPortalGroupTag\" "
						"key not set on first Login Response from target\n");
						retval = -1;
						goto out;
					}
				}
			*login_flags &= ~FIRST_FLAG;
			}

		count++;			/* to prevent infinite loops */
		if( count >= LOOP_TIMES )
			{
			TRACE_ERROR("Infinite loop in operational phase negotiations\n");
			retval = -1;
			goto out;
			}

		if( (inputpdu->flags & T_BIT ) && add_length > 0 )
			{/* protocol error  */
			TRACE_ERROR(
			"Target has T bit set, but initiator still needs to send more\n");
			retval = -1;
			goto out;
			}

		} while( add_length > 0  ||  !(inputpdu->flags & T_BIT) );

	/* When loop ends, operational parameter negotiation stage is finished */
	/* Check whether all parameters that were sent in this stage */
	/* have received their responses */
	if( check_neg_responses(p_param_tbl, 1) < 0 )
		{/* for now, ignore this (message already given) */
		}

	if( *login_flags & MORE_TO_SEND_FLAG )
		{/* protocol error */
		TRACE_ERROR("Not all parameters negotiated\n");
		}

out:

	TRACE(TRACE_ENTER_LEAVE, "Leaving initiator_operational_phase, retval %d\n",
																		retval);

	return retval;
	}



/************************************************************************/
/* This local function performs parameter negotiation on initiator side.
 * Called only from initiator_parameter_negotiate() after all local
 * working storage has been allocated.
 * Returns 1 if Redirection, 0 if otherwise successful, -1 on error.
 * If successful:
 *		login phase is complete,
 *		digests have been enabled as necessary,
 *		all parameters passed by pointer have been updated,
 *		we are now in FFP.
 ************************************************************************/

static int init_parameter_negotiate( struct socket *sock, 
					struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
					struct chap_parameter_type chap_param, /* chap support - CHONG */
					__u32 *cur_cmd_sn,
					unsigned int *connection_flags,
					unsigned int draft_number,
					unsigned int when_called,
					int *exp_stat_sn,
					int *max_send_length,
					int *max_recv_length,
					struct generic_pdu *inputpdu,
					struct generic_pdu *outputpdu,
					struct unknown_key **unknown_key_list )
	{
	int retval = 0;
	int	nsecurity, ninformational, noperational;
	int target_version_max = -1, target_version_active = -1;
	unsigned long long	login_flags;


	TRACE(TRACE_ENTER_LEAVE, "Enter init_parameter_negotiate\n");

	/* about to process the first Login Request/Response */
	login_flags = FIRST_FLAG;

	/* count the number of each kind of parameter still to be negotiated */
	scan_table_and_count(p_param_tbl, draft_number, &nsecurity,
											&ninformational, &noperational);

	/* to force a login phase, make it look like there is a param of that type*/
	if( *connection_flags & USE_SECURITY_PHASE )
		nsecurity++;
	if( *connection_flags & USE_OPERATIONAL_PHASE )
		noperational++;

	if( nsecurity+noperational+ninformational == 0 )
		{/* no parameters at all to negotiate, and no phases forced! */
		nsecurity++;		/* might as well start in security phase */
		}

	if( nsecurity > 0 )
		{/* have some security parameters to negotiate
			or want to force security phase in any case */
		/* chap support - CHONG */
		if( (retval = initiator_security_phase(sock, p_param_tbl, chap_param, cur_cmd_sn,
								exp_stat_sn, max_send_length, noperational,
								draft_number, when_called, inputpdu, outputpdu,
								&target_version_max, &target_version_active,
								*connection_flags, &login_flags,
								unknown_key_list)) < 0 )
			{
			goto out;
			}

		ninformational = 0;		/* these have been sent in security phase */
		}

	/*	for draft 8 and before, enable digests at end of security stage.
		Draft 8, Section 5.2: "All PDUs sent after the security negotiation
		stage MUST be built using the agreed security."
	*/
	if( draft_number <= DRAFT8  &&  retval == 0 )
		{/* draft 8 or before */
		set_digestflags(p_param_tbl, connection_flags);
		}

	if( (noperational+ninformational > 0 || (inputpdu->flags & NSG) == NSG1)  &&
																retval == 0 )
		{/* enter the operational negotiation phase */
		if( (retval = initiator_operational_phase(sock, p_param_tbl,
				cur_cmd_sn, exp_stat_sn, max_send_length, draft_number,
				when_called, inputpdu, outputpdu,
				&target_version_max, &target_version_active,
				*connection_flags, &login_flags, unknown_key_list)) < 0 )
			{
			goto out;
			}
		} 

	/*	Draft 18, Section 4.2 Text Mode Negotiation
		"Some parameters may be subject to integrity rules (e.g., parameter-x
		must not exceed parameter-y or parameter-u not 1 implies parameter-v
		be Yes).  Whenever required, integrity rules are specified with the
		keys.  Checking for compliance with the integrity rule must only be
		performed after all the parameters are available (the existent and
		newly negotiated).  An iSCSI target MUST perform integrity checking
		before the new parameters take effect.  An initiator MAY perform
		integrity checking."
	*/
	if( retval == 0 )
		{/* not a redirection */
		check_integrity_rules(p_param_tbl, draft_number);

		/*	for draft 9 and after, enable digests at end of the login phase.
			Draft 9 Appendix D 10: "When Initiator and Target agree on a digest,
			this digest MUST be used for every PDU in Full Feature Phase."
		*/
		if( draft_number >= DRAFT9 )
			{/* draft 9 and after */
			set_digestflags(p_param_tbl, connection_flags);
			}

		/*	if we (the initiator) sent a MaxRecvDataSegmentLength (draft 18)
			value to the target, we have to now start using it in FFP */
		set_connection_recv_length(p_param_tbl, max_recv_length, draft_number);
		}

out:

	TRACE(TRACE_ENTER_LEAVE, "Leave init_parameter_negotiate, retval %d\n",
																		retval);
	return retval;
	}


/************************************************************************/
/* This function is the entry point to this file to perform parameter
 * negotiation on initiator side.
 * Called only from iscsi_initiator_login() in iscsi_initiator.c.
 * Returns 1 on Redirection, 0 on all other successes, -1 on any error.
 ************************************************************************/

int initiator_parameter_negotiate( struct socket *sock, 
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS], 
			struct chap_parameter_type chap_param, /* chap support - CHONG */
			__u32 *cur_cmd_sn,
			__u32 login_task_tag,
			__u8 isid[6],
			__u32 *tsih,
			__u16 cid,
			__u32 *connection_flags,
			__u32 draft_number,
			__u32 version_number,
			__u32 when_called,
			int *exp_stat_sn,
			int *max_send_length,
			int *max_recv_length )
	{
	int								retval = 0;
	struct generic_pdu				*inputpdu, *outputpdu;
	struct iscsi_init_login_cmnd	*loginpdu;
	struct unknown_key				*uptr, *unknown_key_list = NULL;


	TRACE(TRACE_ENTER_LEAVE, "Enter initiator_parameter_negotiate\n");

	if( (inputpdu=my_kmalloc(sizeof(struct generic_pdu),"Input pdu")) == NULL )
		{
		retval = -1;
		goto out;
		}

	if( (outputpdu=my_kmalloc(sizeof(struct generic_pdu),"Output pdu")) == NULL)
		{
		my_kfree((void **)&inputpdu, "Input pdu");
		retval = -1;
		goto out;
		}

	/* Get memory for text portion of inputpdu and outputpdu */
	if( (inputpdu->text = my_kmalloc(MAX_TEXT_LEN, "Input text")) == NULL )
		{
		my_kfree((void **)&outputpdu, "Output pdu");
		my_kfree((void **)&inputpdu, "Input pdu");
		retval = -1;
		goto out;
		}
	inputpdu->text_length = 0;

	if( (outputpdu->text = my_kmalloc(MAX_TEXT_LEN, "Output text")) == NULL )
		{
		my_kfree((void **)&inputpdu->text, "Input text");
		my_kfree((void **)&outputpdu, "Output pdu");
		my_kfree((void **)&inputpdu, "Input pdu");
		retval = -1;
		goto out;
		}
	outputpdu->text_length = 0;

	/* Set up the output pdu */
	loginpdu = (struct iscsi_init_login_cmnd *)outputpdu;

	/* initialize everything to 0 so reserved fields will be 0 */
	memset(loginpdu, 0, ISCSI_HDR_LEN);

	/* then set up those fields in the Login Request PDU that will not change */
	loginpdu->opcode = ISCSI_INIT_LOGIN_CMND;		/* login request opcode */
	if( draft_number <= DRAFT9 || draft_number >= DRAFT18 )
		{/* for drafts 9 and before, or draft18 and after,
			Login command always marked immediate */
		loginpdu->opcode |= I_BIT;
		}
	loginpdu->version_max = version_number;
	loginpdu->version_min = version_number;
	memcpy(loginpdu->isid, isid, 6);
	loginpdu->tsih = cpu_to_be16(*tsih);
	loginpdu->init_task_tag = cpu_to_be32(login_task_tag);
	loginpdu->cid = cpu_to_be16(cid);

	TRACE(TRACE_DEBUG,"initiator_parameter_negotiate draft_number %u "
						"version_number %u\n", draft_number, version_number);

	/* chap support - CHONG */
	retval = init_parameter_negotiate(sock, p_param_tbl, chap_param,  cur_cmd_sn,
										connection_flags, draft_number,
										when_called, exp_stat_sn,
										max_send_length, max_recv_length,
										inputpdu, outputpdu, &unknown_key_list);

	if( retval == 0  &&  *tsih == 0 )
		{/* first login phase of a new session completed successfully,
			pass back the tsih assigned to this session by the target */
		*tsih = be16_to_cpu(((struct iscsi_targ_login_rsp *)inputpdu)->tsih);
		}

	/* free up memory allocated for any unknown keys we received */
	while( unknown_key_list != NULL )
		{
		if( !unknown_key_list->processed )
			{
			TRACE_ERROR("Unknown key \"%s=%s\"\n", unknown_key_list->keyname,
													unknown_key_list->keyvalue);
			}
		uptr = unknown_key_list->next;
		my_kfree((void **)&unknown_key_list->keyvalue, "Key Value");
		my_kfree((void **)&unknown_key_list->keyname, "Key Name");
		my_kfree((void **)&unknown_key_list, "Unknown key");
		unknown_key_list = uptr;
		}

	/* Free all allocated memory */
	my_kfree((void **)&outputpdu->text, "Output text");
	my_kfree((void **)&inputpdu->text, "Input text");
	my_kfree((void **)&outputpdu, "Output pdu");
	my_kfree((void **)&inputpdu, "Input pdu");

out:

	TRACE(TRACE_ENTER_LEAVE, "Leave initiator_parameter_negotiate, retval %d\n",
																	retval);

	return retval;
	}
