#include <linux/blk.h>
#include <linux/kernel.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/unistd.h>
#include <linux/spinlock.h>
#include <asm/io.h>   
#include <asm/irq.h>
#include <linux/smp_lock.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <asm/uaccess.h>

#include <net/sock.h>
#include <net/tcp.h>
#include <linux/module.h>

#include "iscsi_common.h"

#ifdef __KERNEL__

char *string_llx( unsigned long long x )
{
    int			i;
    static char		str[20];
    static char		convert[] = "0123456789abcdef";
    char		*ptr;
    unsigned int	c;
    union {
    	  unsigned long long	x;
	  unsigned char		c[8];
	  }	tmp;

    tmp.x = x;
    ptr = str;
    *ptr++ = '0';
    *ptr++ = 'x';
    for( i = 7;  i >= 0;  i-- )
    	{
	c = tmp.c[i];
	*ptr++ = convert[c>>4];
	*ptr++ = convert[c&0xf];
	}
    *ptr = '\0';
    return str;
}

void print_rsvd_u8( int n, unsigned char rsvd )
{
  if( rsvd != 0 )
    printk("    rsvd%d: 0x%.2x\n", n, rsvd);
}
void print_rsvd_u16( int n, unsigned short rsvd )
{
  if( rsvd != 0 )
    printk("    rsvd%d: 0x%.4x\n", n, rsvd);
}
void print_rsvd_u32( int n, unsigned int rsvd )
{
  if( rsvd != 0 )
    printk("    rsvd%d: 0x%.8x\n", n, rsvd);
}
void print_rsvd_u64( int n, unsigned long long rsvd )
{
    if( rsvd != 0ll )
      {
      printk("    rsvd%d: %s\n", n, string_llx(rsvd));
      }
}
void print_opcode( unsigned char opcode )
{
    printk("    Opcode: 0x%.2x,  I: %d\n", opcode&0x3f, (opcode&0x40)!=0 );
}
void print_flags( unsigned char flags )
{
    printk("    flags: 0x%.2x\n", flags);
}
void print_lun( unsigned long long lun )
{
    if( lun != 0ll )
      {
      printk("    LUN: %s\n", string_llx(lun));
      }
}
void print_isid_tsih( __u8 isid[6], __u16 tsih )
{
    printk("    ISID: 0x%.2x %.2x %.2x %.2x %.2x %.2x\n",
			isid[0], isid[1], isid[2], isid[3], isid[4], isid[5]);
	/*	need to go back and fix this to print TSID for drafts < 12 */
    printk("    TSIH: %u\n", be16_to_cpu(tsih));
}
void print_dsl( unsigned int length )
{
    if( length != 0 )
      {
      printk("    DSL: %u\n", be32_to_cpu(length));
      }
}
void print_itt( unsigned int init_task_tag )
{
	if( init_task_tag == 0xffffffff )
      printk("    ITT: 0x%08x\n", init_task_tag);
	else
      printk("    ITT: %u\n", be32_to_cpu(init_task_tag));
}
void print_ttt( unsigned int target_xfer_tag )
{
	if( target_xfer_tag == 0xffffffff )
      printk("    TTT: 0x%08x\n", target_xfer_tag);
	else
      printk("    TTT: %u\n", be32_to_cpu(target_xfer_tag));
}
void print_expstatsn( unsigned int exp_stat_sn )
{
    if( exp_stat_sn != 0 )
      {
      printk("    ExpStatSN: %u\n", be32_to_cpu(exp_stat_sn));
      }
}
void print_cmdsn_expstatsn( unsigned int cmd_sn, unsigned int exp_stat_sn )
{
    printk("    CmdSN: %u\n", be32_to_cpu(cmd_sn));
    print_expstatsn(exp_stat_sn);
}
void print_statsn_exp_max( unsigned int stat_sn, unsigned int exp_cmd_sn,
						unsigned int max_cmd_sn )
{
    if( stat_sn != 0 )
      printk("    StatSN: %u\n", be32_to_cpu(stat_sn));
    printk("    ExpCmdSN: %u\n", be32_to_cpu(exp_cmd_sn));
    printk("    MaxCmdSN: %u\n", be32_to_cpu(max_cmd_sn));
}
void print_residual( unsigned int resid )
{
    if( resid != 0 )
      {
      printk("    ResidualCount: %u\n", be32_to_cpu(resid));
      }
}
void print_datasn( unsigned int data_sn )
{
    if( data_sn != 0 )
      {
      printk("    DataSN: %u\n", be32_to_cpu(data_sn));
      }
}
void print_offset( unsigned int offset )
{
    if( offset != 0 )
      {
      printk("    BufferOffset: %u\n", be32_to_cpu(offset));
      }
}
void print_rtt( unsigned int ref_task_tag )
{
	if( ref_task_tag != 0 )
      printk("    RTT: 0x%.8x\n", be32_to_cpu(ref_task_tag));
}
void print_exp_data_sn( unsigned int exp_data_sn )
{
    if( exp_data_sn != 0 )
      printk("    ExpDataSN: %u\n", be32_to_cpu(exp_data_sn));
}

void print_init_scsi_cmnd(struct iscsi_init_scsi_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u8(1, cmd->rsvd1);
    /*	need to go back and fix this to print CRN only for drafts < 12 */
    if( cmd->crn != 0 )
      printk("    CRN: 0x%.2x\n", cmd->crn);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    printk("    EDTL: %u\n", be32_to_cpu(cmd->xfer_len));
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    printk("    CDB: 0x%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x\n",
		cmd->cdb[0], cmd->cdb[1], cmd->cdb[2], cmd->cdb[3], 
		cmd->cdb[4], cmd->cdb[5], cmd->cdb[6], cmd->cdb[7], 
		cmd->cdb[8], cmd->cdb[9], cmd->cdb[10], cmd->cdb[11],
		cmd->cdb[12], cmd->cdb[13], cmd->cdb[14], cmd->cdb[15]);
}

void print_targ_scsi_rsp(struct iscsi_targ_scsi_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    if( cmd->response != 0 )
      printk("    Response: 0x%.2x\n", cmd->response);
    if( cmd->status != 0 )
      printk("    Status: 0x%.2x\n", cmd->status);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_exp_data_sn(cmd->exp_data_sn);
    if( cmd->bidi_resid != 0 )
      printk("    BidiResidualCount: %u\n", be32_to_cpu(cmd->bidi_resid));
    print_residual(cmd->resid);
}

void print_init_text_cmnd(struct iscsi_init_text_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_targ_text_rsp(struct iscsi_targ_text_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_init_login_cmnd(struct iscsi_init_login_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    printk("    VersionMax: 0x%.2x\n", cmd->version_max);
    printk("    VersionMin: 0x%.2x\n", cmd->version_min);
    print_dsl(cmd->length);
    print_isid_tsih(cmd->isid, cmd->tsih);
    print_itt(cmd->init_task_tag);
    printk("    CID: %u\n", be16_to_cpu(cmd->cid));
    print_rsvd_u16(1, cmd->rsvd1);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_targ_login_rsp(struct iscsi_targ_login_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    printk("    VersionMax: 0x%.2x\n", cmd->version_max);
    printk("    VersionActive: 0x%.2x\n", cmd->version_active);
    print_dsl(cmd->length);
    print_isid_tsih(cmd->isid, cmd->tsih);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(1, cmd->rsvd1);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    if( cmd->status_class != 0 )
      printk("    StatusClass: 0x%.2x\n", cmd->status_class);
    if( cmd->status_detail != 0 )
      printk("    StatusDetail: 0x%.2x\n", cmd->status_detail);
    print_rsvd_u16(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_init_logout_cmnd(struct iscsi_init_logout_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    printk("reasoncod: 0x%.2x\n", cmd->flags);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    if( cmd->cid != 0 )
      printk("    CID: %u\n", be16_to_cpu(cmd->cid));
    print_rsvd_u8(2, cmd->rsvd2);
    print_rsvd_u8(3, cmd->rsvd3);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_targ_logout_rsp(struct iscsi_targ_logout_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    printk(" response: 0x%.2x\n", cmd->response);
    print_rsvd_u8(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(3, cmd->rsvd3);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    printk("Time2Wait: 0x%.8x\n", be16_to_cpu(cmd->time2wait));
    printk("Tm2Retain: 0x%.8x\n", be16_to_cpu(cmd->time2retain));
    print_rsvd_u32(5, cmd->rsvd5);
}

void print_init_scsi_data_out(struct iscsi_init_scsi_data_out *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_rsvd_u32(3, cmd->rsvd3);
    print_expstatsn(cmd->exp_stat_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    print_datasn(cmd->data_sn);
    print_offset(cmd->offset);
    print_rsvd_u32(5, cmd->rsvd5);
}

void print_targ_scsi_data_in(struct iscsi_targ_scsi_data_in *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u8(1, cmd->rsvd1);
    if( cmd->status != 0 )
      printk("    Status: 0x%.2x\n", cmd->status);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_datasn(cmd->data_sn);
    print_offset(cmd->offset);
    print_residual(cmd->resid);
}

void print_targ_rjt(struct iscsi_targ_rjt *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    if( cmd->reason != 0 )
      printk("    Reason: 0x%.2x\n", cmd->reason);
    print_rsvd_u8(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(4, cmd->rsvd4);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_datasn(cmd->data_sn);
    print_rsvd_u64(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_init_nopout(struct iscsi_init_nopout *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_targ_nopin(struct iscsi_targ_nopin *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max(cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_targ_r2t(struct iscsi_targ_r2t *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    printk("    R2TSN: %u\n", be32_to_cpu(cmd->r2t_sn));
    print_offset(cmd->offset);
    printk("    DDTL: %u\n", be32_to_cpu(cmd->xfer_len));
}

void print_targ_async_msg(struct iscsi_targ_async_msg *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(3, cmd->rsvd3);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    printk("AsyncEvnt: %u\n", cmd->async_event);
    printk("AsyncVCod: %u\n", cmd->async_vcode);
    if( cmd->parameter1 != 0 )
	printk("   Param1: %u\n", cmd->parameter1);
    if( cmd->parameter2 != 0 )
	printk("   Param2: %u\n", cmd->parameter2);
    if( cmd->parameter3 != 0 )
	printk("   Param3: %u\n", cmd->parameter3);
    print_rsvd_u32(5, cmd->rsvd5);
}

void print_init_task_mgt_command(struct iscsi_init_task_mgt_command *cmd)
{
    print_opcode(cmd->opcode);
    printk("    Function: 0x%.2x\n", cmd->function);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rtt(cmd->ref_task_tag);
    print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
    if( cmd->ref_cmd_sn != 0 )
      printk("    RefCmdSN: %u\n", be32_to_cpu(cmd->ref_cmd_sn));
    print_exp_data_sn(cmd->exp_data_sn);
    print_rsvd_u64(4, cmd->rsvd4);
}

void print_targ_task_mgt_response(struct iscsi_targ_task_mgt_response *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    if( cmd->response != 0 )
      printk("    Response: 0x%.2x\n", cmd->response);
    if( cmd->qualifier != 0 )
      printk("    Qualifier: 0x%.2x\n", cmd->qualifier);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rtt(cmd->ref_task_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

/* Print SNACK Request - SAI */

/********************************************************************************
*       Called to print the SNACK Request being sent
********************************************************************************/

void print_begrun(unsigned int begrun)
        {
        printk("    BegRun: %u\n", be32_to_cpu(begrun));
        }

void print_runlen(unsigned int runlen)
        {
        printk("    RunLength: %u\n", be32_to_cpu(runlen));
        }

void print_init_snack(struct iscsi_init_snack *cmd)
        {
        print_opcode(cmd->opcode);
        print_flags(cmd->flags);
        print_rsvd_u16(1, cmd->rsvd1);
        print_dsl(cmd->length);
        print_lun(cmd->lun);
        print_itt(cmd->init_task_tag);
        print_ttt(cmd->target_xfer_tag);
        print_rsvd_u32(2, cmd->rsvd2);
        print_expstatsn(cmd->exp_stat_sn);
        print_rsvd_u64(3, cmd->rsvd3);
        print_begrun(cmd->begrun);
        print_runlen(cmd->runlen);
        }

/*	returns pointer to printable name of SCSI op code */
/*	(Normally this is byte 1 of a SCSI CDB */
char *printable_scsi_op( unsigned int opcode )
	{
	static char		buffer[16];
	char			*result = "";


	/*	opcode symbols defined in <scsi/scsi.h> */
	/*	this table includes only those symbols we have actually seen
		being sent to/from a vendor */
	switch( opcode )
		{
	case TEST_UNIT_READY:		/* 0x00 */
		result = "TEST_UNIT_READY";
		break;
	case REZERO_UNIT:			/* 0x01 */
		result = "REWIND";
		break;
	case REQUEST_SENSE:			/* 0x03 */
		result = "REQUEST_SENSE";
		break;
	case READ_BLOCK_LIMITS:		/* 0x05 */
		result = "READ_BLOCK_LIMITS";
		break;
	case READ_6:				/* 0x08 */
		result = "READ_6";
		break;
	case WRITE_6:				/* 0x0a */
		result = "WRITE_6";
		break;
	case WRITE_FILEMARKS:		/* 0x10 */
		result = "WRITE_FILEMARKS";
		break;
	case INQUIRY:				/* 0x12 */
		result = "INQUIRY";
		break;
	case MODE_SENSE:			/* 0x1a */
		result = "MODE_SENSE";
		break;
	case READ_CAPACITY:			/* 0x25 */
		result = "READ_CAPACITY";
		break;
	case READ_10:				/* 0x28 */
		result = "READ_10";
		break;
	case WRITE_10:				/* 0x2a */
		result = "WRITE_10";
		break;
	case READ_12:				/* 0xa8 */
		result = "READ_12";
		break;
	case WRITE_12:				/* 0xaa */
		result = "WRITE_12";
		break;
	default:
		sprintf(buffer, "CDBop 0x%02x", opcode);
		result = buffer;
		break;
		}	/* switch */
	return result;
	}

#include "my_memory.h"

#endif
