/**********************************************************************
*	ref18_01/common/text_param.h
*
*	This defines the generic functions used in the Login phase
*	by both the iSCSI target for parameter negotiation
*
*   IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
*   By downloading, copying, installing or using the software you agree to this
*   license. If you do not agree to this license, do not download, install,
*   copy or use the software.
*
*   License Agreement
*
*   Copyright (C) 2001  Inter-Operability Lab.(IOL),Univ. of New Hampshire(UNH),
*   Durham, NH 03824
*
*   Redistribution and use in source and binary forms, with or without
*   modification, are permitted provided that the following conditions are met:
*
*   -Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
*
*   -Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
*
*   The name of IOL/UNH may not be used to endorse or promote products derived
*   from this software without specific prior written permission.
*
*   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
*   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IOL,UNH OR CONTRIBUTORS
*   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF
*   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*   POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************/


#ifndef TEXT_PARAM_H
#define TEXT_PARAM_H

#include "iscsi_common.h"
#include "crc.h"
/* chap support -CHONG */
#include "chap.h"

/*
 * Define the Login/Text Parameters
 */

#define HEADERDIGEST				"HeaderDigest"
#define DATADIGEST				"DataDigest"
#define MAXCONNECTIONS				"MaxConnections"
#define SENDTARGETS				"SendTargets"
#define TARGETNAME				"TargetName"
#define INITIATORNAME				"InitiatorName"
#define TARGETALIAS				"TargetAlias"
#define INITIATORALIAS				"InitiatorAlias"
#define TARGETADDRESS				"TargetAddress"
#define FMARKER					"FMarker"
#define RFMARKINT				"RFMarkInt"
#define SFMARKINT				"SFMarkInt"
#define INITIALR2T				"InitialR2T"
#define BIDIINITIALR2T				"BidiInitialR2T"
#define IMMEDIATEDATA				"ImmediateData"
#define MAXRECVPDULENGTH			"MaxRecvPDULength"
#define MAXRECVDATASEGMENTLENGTH		"MaxRecvDataSegmentLength"
#define DATAPDULENGTH                           "DataPDULength"
#define MAXBURSTSIZE				"MaxBurstSize"
#define MAXBURSTLENGTH                          "MaxBurstLength"
#define FIRSTBURSTSIZE				"FirstBurstSize"
#define FIRSTBURSTLENGTH                        "FirstBurstLength"
#define LOGOUTLOGINMAXTIME			"LogoutLoginMaxTime"
#define LOGOUTLOGINMINTIME			"LogoutLoginMinTime"
#define MAXOUTSTANDINGR2T			"MaxOutstandingR2T"
#define DATAPDUINORDER				"DataPDUInOrder"
#define DATASEQUENCEINORDER			"DataSequenceInOrder"
#define ERRORRECOVERYLEVEL			"ErrorRecoveryLevel"
#define SESSIONTYPE				"SessionType"
#define AUTHMETHOD				"AuthMethod"
#define DEFAULTTIME2WAIT			"DefaultTime2Wait"
#define DEFAULTTIME2RETAIN			"DefaultTime2Retain"
#define IFMARKER				"IFMarker"
#define OFMARKER				"OFMarker"
#define IFMARKINT				"IFMarkInt"
#define OFMARKINT				"OFMarkInt"
#define TARGETPORTALGROUPTAG                    "TargetPortalGroupTag"
#define X_EXTENSIONKEY				"X-edu.unh.iol-extension-key-1"

/* Following Macros Added for Error Recovery - SAI*/

#define PAYLOAD_DIGERR                  0x00000000
#define HEADER_DIGERR                   0x00000001
#define SEQUENCE_ERR                    0x00000010



/*
 * Define the Options available for the Parameters
 */

/* common for all */
#define NONE_LOW					"none"
#define NONE						"None"
#define REJECT						"Reject"
#define REJECT_LOW					"reject"
#define IRRELEVANT					"Irrelevant"
#define IRRELEVANT_LOW					"irrelevant"

/* For Digests - (HeaderDigest, DataDigest) */
#define CRC32C						"CRC32C"


/* Different Authentication Methods supported */
#define KRB5						"KRB5"	/* RFC1510 */
#define SPKM1						"SPKM1" /* RFC2025 */
#define SPKM2						"SPKM2" /* RFC2025 */
#define SRP						"SRP"	/* RFC2945 */
#define CHAP						"CHAP"  /* RFC1994 */

/* chap support-CHONG */
#define MAX_CHALLENGE_LENGTH		0x400 /* 1024 bytes */

/*******************************************************************/
/* For Parameter types */

/*****************************************************************
 	byte-3 		byte-2 		byte-1 		byte-0
	XX		XX		XX		XX
********************************************************************/

/* following constants use the byte-3 and byte-2 for specifying type */
#define SECURITY_PARAM		0x80000000
#define OPERATIONAL_PARAM	0x40000000
#define INFORMATIONAL_PARAM	0x20000000
#define LEADING_ONLY		0x10000000
#define INITIAL_ONLY		0x08000000
#define FULL_FEATURE_ONLY	0x04000000
#define ALL			0x02000000
#define USE_BY_INITIATOR	0x01000000
#define USE_BY_TARGET		0x00800000
#define NORMAL_RETURN		0x00400000/*TargetAddress in normal loginRPS*/
#define REDIRECT_RETURN		0x00200000/*TargetAddress for redirection*/
#define	SENDTARGET_RETURN	0x00100000/*TargetAddress in return to SendTargets*/

/* byte-2  for denoting whether an offer of the parameter can be nothing,
		just a single value, a 2-value range, or a list of values */
#define KEY_NO_VALUE		0x00080000 /* follow = with <= 1 value */
#define KEY_SINGLE_VALUE	0x00040000 /* follow = with == 1 value */
#define KEY_TWO_VALUE		0x00020000 /* follow = with range low[~top] */
#define KEY_MULTI_VALUE		0x00010000 /* follow = with list >= 1 value */

/* following constants use the byte-1 for specifying type */
#define NUMBER			0x00008000 /* numeric values only */
#define STRING			0x00004000 /* arbitrary strings */
#define ENUMERATED		0x00002000 /* set of specific strings */
#define BOOL_AND		0x00001000 /* boolean values only */
#define BOOL_OR			0x00000800 /* boolean values only */
#define MIN_NUMBER		0x00000400 /* numeric values only */
#define MAX_NUMBER		0x00000200 /* numeric values only */
#define NUMBER_RANGE		0x00000100 /* range of 2 numeric values */

/* byte-0 qualifies the byte-1 */

/* if the byte-1 specifies the type to be NUMBER */
#define ONE_TO_65535		0x00000080
#define N512_TO_16777215	0x00000040
#define ZERO_TO_3600		0x00000020
#define TWO_TO_3600		0x00000010
#define ZERO_TO_2		0x00000008
#define ZERO_TO_32767		0x00000004
#define ZERO_TO_65535           0x00000002

/* if the byte-1 specifies the type to be STRING */
#define UTF_8			0x00000080 /* <UTF-8 string> */
#define ISCSI_NAME		0x00000040 /* <iSCSI-Name> */
#define TARGET_ADDRESS_TYPE	0x00000020 /* domainname[:port][,tag] iSCSI-Name */

/* if the byte-1 specifies the type to be ENUMERATED */
#define DIGEST_PARAM		0x00000080 /* Key is a Digest key */
#define AUTH_PARAM		0x00000040 /* Key is a Authentication key */
#define DISCOVERY_NORMAL	0x00000020 /* <Session Type> */
#define SEND_RECEIVE		0x00000010 /* <send|receive|send-receive|no>	*/

/********************************************************************/

#define MAX_CONFIG_PARAMS		38

/* NOTE: Whenever a new key is added to the array
		struct parameter_type config_params[MAX_CONFIG_PARAMS]
	 then MAX_CONFIG_PARAMS must be incremented too!
*/

/********************************************************************/

/* These defines are to be used in the special_key_flag field of the
   struct parameter_type to identify only those keys that need "special"
   processing.  Introducing these flags eliminates the need for
   a lot of costly "strcmp()" operations.  The bit assigned to each
   flag is totally arbitrary as long as it is different for each key.
   Most, but not every, key has a bit assigned to it -- only the
   following 4 keys do not have a bit assigned:
		TargetAlias
		InitiatorAlias
		TargetAddress
		X-edu.unh.iol.extension-key

   Some of these defines are also used in login_flags.
   When asked for, they are set if the corresponding special key is found in
   the input by scan_input_and_process().
*/

/* The first set of 35 flags is used in the special_key_flag field of keys.
 * They must not conflict with the defines in the login_flags
 */
#define TARGETPORTALGROUPTAG_FLAG		0x0080000000LL
#define TARGETNAME_FLAG					0x0040000000LL
#define INITIATORNAME_FLAG				0x0020000000LL
#define SESSIONTYPE_FLAG				0x0010000000LL
#define HEADERDIGEST_FLAG				0x0008000000LL
#define DATADIGEST_FLAG					0x0004000000LL
#define SENDTARGETS_FLAG				0x0002000000LL
#define DATAPDULENGTH_FLAG				0x0001000000LL
#define MAXRECVPDULENGTH_FLAG			0x0000800000LL
#define LOGOUTLOGINMINTIME_FLAG			0x0000400000LL
#define LOGOUTLOGINMAXTIME_FLAG			0x0000200000LL
#define FMARKER_FLAG					0x0000100000LL
#define OFMARKER_FLAG					0x0000080000LL
#define IFMARKER_FLAG					0x0000040000LL
#define RFMARKINT_FLAG					0x0000020000LL
#define SFMARKINT_FLAG					0x0000010000LL
#define OFMARKINT_FLAG					0x0000008000LL
#define IFMARKINT_FLAG					0x0000004000LL
#define FIRSTBURSTSIZE_FLAG				0x0000002000LL
#define MAXBURSTSIZE_FLAG				0x0000001000LL
#define DEFAULTTIME2WAIT_FLAG			0x0000000800LL
#define DEFAULTTIME2RETAIN_FLAG			0x0000000400LL
#define AUTHMETHOD_FLAG					0x0000000200LL
#define INITIALR2T_FLAG					0x0000000100LL
#define BIDIINITIALR2T_FLAG				0x0000000080LL
#define IMMEDIATEDATA_FLAG				0x0000000040LL
#define DATAPDUINORDER_FLAG				0x0000000020LL
#define DATASEQUENCEINORDER_FLAG		0x0000000010LL
#define MAXCONNECTIONS_FLAG				0x0000000008LL
#define MAXOUTSTANDINGR2T_FLAG			0x0000000004LL
#define ERRORRECOVERYLEVEL_FLAG			0x0000000002LL
#define X_EXTENSIONKEY_FLAG				0x0000000001LL
#define MAXRECVDATASEGMENTLENGTH_FLAG	0x1000000000LL
#define MAXBURSTLENGTH_FLAG				0x2000000000LL
#define FIRSTBURSTLENGTH_FLAG			0x4000000000LL
#define TARGETADDRESS_FLAG				0x8000000000LL


/* The following defines are used in login_flags only.
   They are set and reset during processing, and must not
   conflict with the special key flags defined above.
*/

/* The second set of 2 flags is never used in the special_key_flag field
   of keys, only in login_flags.  Therefore, they are 64-bits, as is the
   login_flags variable.
*/
/* set when there are more keys to add to output than fit in one PDU */
#define MORE_TO_SEND_FLAG		0x0100000000LL

/* set when processing first Login Request/Response, else reset */
#define	FIRST_FLAG			0x0200000000LL

/* set when in the discovery session, else in normal session */
#define DISCOVERY_FLAG			0x0400000000LL


/************************************************************************/
#define ALLOCATE_MAX_TEXT_LEN(str)	do { \
	if( !((str) = (char *)kmalloc( MAX_TEXT_LEN, GFP_KERNEL)) ) \
		{ \
		TRACE_ERROR("unable to allocate memory for" #str "\n"); \
		return -1; \
		} \
	} \
					while(0)

#define FREE_STRING(str)		if(str) \
						{ \
						kfree(str); \
						}

/********************************************************************/

#define YES		"Yes"
#define YES_LOW		"yes"
#define NO		"No"
#define NO_LOW		"no"
#define SEND		"send"
#define RECEIVE		"receive"
#define SENDRECEIVE	"send-receive"
#define NORMAL		"Normal"
#define NORMAL_LOW	"normal"
#define DISCOVERY	"Discovery"
#define DISCOVERY_LOW	"discovery"


#define INITIAL_FIRSTBURSTLENGTH	65536
#define INITIAL_MAXBURSTLENGTH		262144

int target_num; /* target_num is defined in scsi_target.c */

/* This structure holds the parameters required for login-text
 * negotiation
 */
struct parameter_type
{
  /* Name of the parameter */
  char *parameter_name;

  /* Type of the parameter */
  /* Byte 3 - Security, operational etc.. */
  /* Byte 2 - Number of values that can be sent with the key */
  /* Byte 1 - number, string, enumerated, boolean, etc.. */
  /* Byte 0 - further qualifies the type depending on Byte-1 */
  unsigned int type;

  /* used only if the type is NUMBER or NUMBER_RANGE,
     not STRING or ENUMERATED or BOOLEAN */
  unsigned int int_value;

  /* used only if the type is STRING or ENUMERATED or BOOLEAN or NUMBER_RANGE,
     not NUMBER */
  char *str_value;

  /* used only if the type is STRING or ENUMERATED or BOOLEAN or NUMBER_RANGE,
     not NUMBER */
  /* Stores the list of allowed values (for negotiation) */
  char *value_list;

  /* Contains all the info required for negotiating this parameter */
  /* packed into bits */
  unsigned int neg_info;

  /* One of the bits in this word will be set if this key
     requires "special" processing  -- each key defines its own special bit */
  unsigned long long special_key_flag;

};

/* chap support - CHONG */
struct chap_parameter_type
{
	int chap_flags;
	CHAP_Context * local_ctx;
	CHAP_Context * peer_ctx;
};


/* Store the information about the configurable parameters in this array.

   The values in this table are ALWAYS updated to the latest draft (12).

   Older drafts with different values required dynamic modifications in
   adjust_version_diff() and dynamic checks in check_drafts().
*/
struct parameter_type config_params[MAX_CONFIG_PARAMS]
= {
  {
    parameter_name: HEADERDIGEST,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_MULTI_VALUE |
    ENUMERATED | DIGEST_PARAM,
    str_value: NONE,
    value_list: NONE,
    special_key_flag: HEADERDIGEST_FLAG,
  },
  {
    parameter_name: DATADIGEST,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_MULTI_VALUE |
    ENUMERATED | DIGEST_PARAM,
    str_value: NONE,
    value_list: NONE,
    special_key_flag: DATADIGEST_FLAG,
  },
  {
    parameter_name: MAXCONNECTIONS,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    NUMBER | ONE_TO_65535 | MIN_NUMBER,
    int_value: 1,
    special_key_flag: MAXCONNECTIONS_FLAG,
  },
  {
    parameter_name: SENDTARGETS,
    type:
    FULL_FEATURE_ONLY |
    USE_BY_INITIATOR |
    KEY_NO_VALUE |
    STRING | UTF_8,
    str_value: NULL,
    value_list: NULL,
    special_key_flag: SENDTARGETS_FLAG,
  },
  {
    parameter_name: TARGETNAME,
    type:
    INFORMATIONAL_PARAM |
    ALL |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    STRING | ISCSI_NAME,
    str_value: NULL,
    value_list: NULL,
    special_key_flag: TARGETNAME_FLAG,
  },
  {
    parameter_name: INITIATORNAME,
    type:
    INFORMATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    STRING | ISCSI_NAME,
    str_value: NULL,
    value_list: NULL,
    special_key_flag: INITIATORNAME_FLAG,
  },
  {
    parameter_name: TARGETALIAS,
    type:
    INFORMATIONAL_PARAM |
    ALL |
    USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    STRING | UTF_8,
    str_value: NULL,
    value_list: NULL,
  },
  {
    parameter_name: INITIATORALIAS,
    type:
    INFORMATIONAL_PARAM |
    ALL |
    USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    STRING | UTF_8,
    str_value: NULL,
    value_list: NULL,
  },
  {
    parameter_name: TARGETADDRESS,
    type:
    INFORMATIONAL_PARAM |
    ALL |
    USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    STRING | TARGET_ADDRESS_TYPE,
    str_value: NULL,
    value_list: NULL,
    special_key_flag: TARGETADDRESS_FLAG,
  },
  {
    parameter_name: FMARKER,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    ENUMERATED | SEND_RECEIVE,
    str_value: NO,
    value_list: NO,
    special_key_flag: FMARKER_FLAG,
  },
  {
    parameter_name: OFMARKER,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    BOOL_AND,
    str_value: NO,
    value_list: NO,
    special_key_flag: OFMARKER_FLAG,
  },
  {
    parameter_name: IFMARKER,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    BOOL_AND,
    str_value: NO,
    value_list: NO,
    special_key_flag: IFMARKER_FLAG,
  },
  {
    parameter_name: RFMARKINT,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_TWO_VALUE |
    NUMBER | ONE_TO_65535,
    int_value: 2048,
    special_key_flag: RFMARKINT_FLAG,
  },
  {
    parameter_name: SFMARKINT,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    NUMBER | ONE_TO_65535,
    int_value: 2048,
    special_key_flag: SFMARKINT_FLAG,
  },
  {
    parameter_name: OFMARKINT,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_TWO_VALUE |
    NUMBER_RANGE | ONE_TO_65535,
    int_value: 2048,
    str_value: "2048~2048",
    value_list: "2048~2048",
    special_key_flag: OFMARKINT_FLAG,
  },
  {
    parameter_name: IFMARKINT,
    type:
    OPERATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_TWO_VALUE |
    NUMBER_RANGE | ONE_TO_65535,
    int_value: 2048,
    str_value: "2048~2048",
    value_list: "2048~2048",
    special_key_flag: IFMARKINT_FLAG,
  },
  {
    parameter_name: INITIALR2T,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    BOOL_OR,
    str_value: YES,
    value_list: YES,
    special_key_flag: INITIALR2T_FLAG,
  },
  { /* removed in draft 16 and after */
    parameter_name: BIDIINITIALR2T,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    BOOL_OR,
    str_value: YES,
    value_list: YES,
    special_key_flag: BIDIINITIALR2T_FLAG,
  },
  {
    parameter_name: IMMEDIATEDATA,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    BOOL_AND,
    str_value: YES,
    value_list: YES,
    special_key_flag: IMMEDIATEDATA_FLAG,
  },
  { /* used in draft 13 and after in place of MaxRecvPDULength */
    parameter_name: MAXRECVDATASEGMENTLENGTH,
    type:
    OPERATIONAL_PARAM |
    ALL |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    NUMBER | N512_TO_16777215,
    int_value: 8192,
    special_key_flag: MAXRECVDATASEGMENTLENGTH_FLAG,
  },
  {/* used in draft 9 in place of DataPDULength in draft 8 */
    parameter_name: MAXRECVPDULENGTH,
    type:
    OPERATIONAL_PARAM |
    ALL |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    NUMBER | N512_TO_16777215,
    int_value: 8192,
    special_key_flag: MAXRECVPDULENGTH_FLAG,
  },
 { /* only used in draft 8 and before */
    parameter_name: DATAPDULENGTH,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    NUMBER | ZERO_TO_32767,
    int_value: 16,
    special_key_flag: DATAPDULENGTH_FLAG,
  },
  {/* only used in draft 15 and before */
    parameter_name: MAXBURSTSIZE,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    N512_TO_16777215 | MIN_NUMBER |
    NUMBER,
    int_value: INITIAL_MAXBURSTLENGTH,
    special_key_flag: MAXBURSTSIZE_FLAG,
  },
  { /* used in draft 16 in place of MaxBurstSize */
    parameter_name: MAXBURSTLENGTH,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    N512_TO_16777215 | MIN_NUMBER |
    NUMBER,
    int_value: INITIAL_MAXBURSTLENGTH,
    special_key_flag: MAXBURSTLENGTH_FLAG,
  },
  {/* used only in draft 15 and before */
    parameter_name: FIRSTBURSTSIZE,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    N512_TO_16777215 | MIN_NUMBER |
    NUMBER,
    int_value: INITIAL_FIRSTBURSTLENGTH,
    special_key_flag: FIRSTBURSTSIZE_FLAG,
  },
  {/* use in draft 16 in place of FirstBurstSize */
    parameter_name: FIRSTBURSTLENGTH,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    N512_TO_16777215 | MIN_NUMBER |
    NUMBER,
    int_value: INITIAL_FIRSTBURSTLENGTH,
    special_key_flag: FIRSTBURSTLENGTH_FLAG,
  },
  {/* used in darft 11 and after */
    parameter_name: DEFAULTTIME2WAIT,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    NUMBER | ZERO_TO_3600 | MAX_NUMBER,
    int_value: 2,
    special_key_flag: DEFAULTTIME2WAIT_FLAG,
  },
  {/* used in darft 11 and after */
    parameter_name: DEFAULTTIME2RETAIN,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    NUMBER | ZERO_TO_3600 | MIN_NUMBER,
    int_value: 20,
    special_key_flag: DEFAULTTIME2RETAIN_FLAG,
  },
  {/* used in darft 9 and before */
    parameter_name: LOGOUTLOGINMAXTIME,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    NUMBER | TWO_TO_3600 | MIN_NUMBER,
    int_value: 3,
    special_key_flag: LOGOUTLOGINMAXTIME_FLAG,
  },
  {/* used in darft 9 and before */
    parameter_name: LOGOUTLOGINMINTIME,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    NUMBER | TWO_TO_3600 | MAX_NUMBER,
    int_value: 3,
    special_key_flag: LOGOUTLOGINMINTIME_FLAG,
  },
  {
    parameter_name: MAXOUTSTANDINGR2T,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    NUMBER | ONE_TO_65535 | MIN_NUMBER,
    int_value: 1,
    special_key_flag: MAXOUTSTANDINGR2T_FLAG,
  },
  {
    parameter_name: DATAPDUINORDER,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_TARGET | USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    BOOL_OR,
    str_value: YES,
    value_list: YES,
    special_key_flag: DATAPDUINORDER_FLAG,
  },
  {
    parameter_name: DATASEQUENCEINORDER,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    BOOL_OR,
    str_value: YES,
    value_list: YES,
    special_key_flag: DATASEQUENCEINORDER_FLAG,
  },
  {
    parameter_name: AUTHMETHOD,
    type:
    SECURITY_PARAM |
    INITIAL_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_MULTI_VALUE |
    ENUMERATED | AUTH_PARAM,
    str_value: NONE,
    value_list: NONE,
    special_key_flag: AUTHMETHOD_FLAG,
  },
  {
    parameter_name: ERRORRECOVERYLEVEL,
    type:
    OPERATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    NUMBER | ZERO_TO_2 | MIN_NUMBER,
    int_value: 0,
    special_key_flag: ERRORRECOVERYLEVEL_FLAG,
  },
  {
    parameter_name: SESSIONTYPE,
    type:
    INFORMATIONAL_PARAM |
    LEADING_ONLY |
    USE_BY_INITIATOR |
    KEY_SINGLE_VALUE |
    ENUMERATED | DISCOVERY_NORMAL,
    str_value: NORMAL,
    value_list: NORMAL,
    special_key_flag: SESSIONTYPE_FLAG,
  },
  {
    parameter_name: TARGETPORTALGROUPTAG,
    type:
    INFORMATIONAL_PARAM |
    INITIAL_ONLY |
    USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    NUMBER | ZERO_TO_65535,
    special_key_flag: TARGETPORTALGROUPTAG_FLAG,
  },
  {/* non-standard key to text X-extension facilities */
    parameter_name: X_EXTENSIONKEY,
    type:
    OPERATIONAL_PARAM |
    ALL |
    USE_BY_INITIATOR | USE_BY_TARGET |
    KEY_SINGLE_VALUE |
    STRING | UTF_8,
    str_value: NULL,
    value_list: NULL,
    special_key_flag: X_EXTENSIONKEY_FLAG,
  },
};

/* Macros for checking the type */

#define IS_SECURITY_PARAM(p)			((p) & SECURITY_PARAM)
#define IS_OPERATIONAL_PARAM(p)			((p) & OPERATIONAL_PARAM)
#define IS_INFORMATIONAL_PARAM(p)		((p) & INFORMATIONAL_PARAM)
#define IS_LEADING_ONLY(p)			((p) & LEADING_ONLY)
#define IS_INITIAL_ONLY(p)			((p) & INITIAL_ONLY)
#define IS_FULL_FEATURE_ONLY(p)			((p) & FULL_FEATURE_ONLY)
#define IS_USE_BY_INITIATOR(p)			((p) & USE_BY_INITIATOR)
#define IS_USE_BY_TARGET(p)			((p) & USE_BY_TARGET)
#define IS_KEY_SINGLE_VALUE(p)			((p) & KEY_SINGLE_VALUE)
#define IS_KEY_TWO_VALUE(p)			((p) & KEY_TWO_VALUE)
#define IS_KEY_MULTI_VALUE(p)			((p) & KEY_MULTI_VALUE)
#define IS_KEY_NO_VALUE(p)			((p) & KEY_NO_VALUE)
#define IS_NUMBER(p)				((p) & NUMBER)
#define IS_NUMBER_RANGE(p)			((p) & NUMBER_RANGE)
#define IS_MAX_NUMBER(p)			((p) & MAX_NUMBER)
#define IS_MIN_NUMBER(p)			((p) & MIN_NUMBER)
#define IS_STRING(p)				((p) & STRING)
#define IS_ENUMERATED(p)			((p) & ENUMERATED)
#define IS_BOOL_AND(p)				((p) & BOOL_AND)
#define IS_BOOL_OR(p)				((p) & BOOL_OR)
#define IS_BOOLEAN(p)				((p) & (BOOL_AND|BOOL_OR))

/* If the type is number */
#define IS_ONE_TO_65535(p)			((p) & ONE_TO_65535)
#define IS_ZERO_TO_65535(p)                     ((p) & ZERO_TO_65535)
#define IS_N512_TO_16777215(p)			((p) & N512_TO_16777215)
#define IS_ZERO_TO_3600(p)			((p) & ZERO_TO_3600)
#define IS_TWO_TO_3600(p)			((p) & TWO_TO_3600)
#define IS_ZERO_TO_2(p)				((p) & ZERO_TO_2)
#define IS_ZERO_TO_32767(p)                     ((p) & ZERO_TO_32767)



/* If the type is string */
#define IS_UTF_8(p)				(IS_STRING((p)) && ((p) & UTF_8))
#define IS_ISCSI_NAME(p)			(IS_STRING((p)) && ((p) & ISCSI_NAME))
#define IS_TARGET_ADDRESS_TYPE(p)		(IS_STRING((p)) && ((p) & TARGET_ADDRESS_TYPE))

/* If the type is ENUMERATED */
#define IS_DIGEST_PARAM(p)			((p) & DIGEST_PARAM)
#define IS_AUTH_PARAM(p)			((p) & AUTH_PARAM)
#define IS_SEND_RECEIVE(p)			((p) & SEND_RECEIVE)
#define IS_DISCOVERY_NORMAL(p)			((p) & DISCOVERY_NORMAL)


/**********************************************************************/
/* Macros to test bits used in the "neg_info" variable and key field */

#define IS_KEY_TO_BE_NEGOTIATED(p)		((p) & KEY_TO_BE_NEGOTIATED)
#define IS_KEY_BREAK_CONN_OR_RJT(p)		((p) & KEY_BREAK_CONN_OR_RJT)
#define IS_KEY_SENT_TO_OTHER_SIDE(p)		((p) & KEY_SENT_TO_OTHER_SIDE)
#define IS_KEY_GOT_FROM_OTHER_SIDE(p)		((p) & KEY_GOT_FROM_OTHER_SIDE)
#define IS_KEY_REPLY_OPTIONAL(p)		((p) & KEY_REPLY_OPTIONAL)
#define IS_KEY_IRRELEVANT(p)			((p) & KEY_IRRELEVANT)
#define IS_KEY_BAD(p)				((p) & KEY_BAD)
#define IS_KEY_REJECT(p)			((p) & KEY_REJECT)



/*************************************************************************/
/*   Handle the different case of the first letter of some key values for*/
/*   different draft versions. 						 */
/*   There are 2 tables, each with a string with a proper lower/upper    */
/*   case for the first letter						 */
/*   Which table gets used depends on the draft number.			 */
/*  									 */
/*************************************************************************/

struct key_values {
		  char	*yes;
		  char	*no;
		  char	*none;
		  char	*reject;
		  char	*irrelevant;
		  char	*discovery;
		  char	*normal;
		  };

struct key_values  lower_case_values =
		  {
		        none: NONE_LOW,
		         yes: YES_LOW,
		          no: NO_LOW,
		      reject: REJECT_LOW,
		  irrelevant: IRRELEVANT_LOW,
		   discovery: DISCOVERY_LOW,
		      normal: NORMAL_LOW
		  };

struct key_values  upper_case_values =
		  {
		        none: NONE,
		         yes: YES,
		          no: NO,
		      reject: REJECT,
		  irrelevant: IRRELEVANT,
		   discovery: DISCOVERY,
		      normal: NORMAL
		  };


/* structure used to keep a list of unknown keys sent to us */
/* chap support -CHONG*/
struct unknown_key	{
			unsigned int	processed;
			char			* keyname;
			char			* keyvalue;
			struct unknown_key	*next;
			};

/* maximum number of unkown keys we will accept before terminating login */
#define	MAX_UNKNOWN_KEYS	8

/************************************************************************/
/* check if it is security key                                          */
/* modified by Chong Zhang for CHAP implementaion                       */
/************************************************************************/

int is_securitykey(char * keyname)
{
	if ((!strcmp(keyname, "CHAP_A")) ||
		(!strcmp(keyname, "CHAP_N")) ||
		(!strcmp(keyname, "CHAP_I")) ||
		(!strcmp(keyname, "CHAP_C")) ||
		(!strcmp(keyname, "CHAP_R")))
		return 1;
	else
		return 0;
}

/* Called to get table entry for key identified by keytext string */
/* Returns pointer to entry if found, NULL if not found */
struct parameter_type *find_parameter( char *keytext,
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS] )
{
  struct parameter_type *p, *result = NULL;
  int i;


  TRACE(TRACE_ENTER_LEAVE, "Enter find_parameter %s\n", keytext);

  for( i=0, p = p_param_tbl;  i < MAX_CONFIG_PARAMS;  i++, p++ )
    {
      if( !strcmp(p->parameter_name, keytext) )
	{
	  result = p;
	  break;
	}
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave find_parameter, p %p\n", result);

  return result;
}


/* Called to get pointer to table entry for key identified by its special flag*/
struct parameter_type *find_flag_parameter( unsigned long long key_flag,
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS] )
{
  struct parameter_type *p, *result = NULL;
  int i;


  TRACE(TRACE_ENTER_LEAVE,"Enter find_flag_parameter, flag 0x%08llx\n",key_flag);

  for( i = 0, p = p_param_tbl;  i < MAX_CONFIG_PARAMS;  i++, p++ )
    {
      if( p->special_key_flag & key_flag )
	{
	  result = p;
	  break;
	}
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave find_flag_parameter, p %p\n", result);

  return result;
}


/*	called to set max_recv_length to value we sent to target in a
	MaxRecvDataSegmentLength key */
void set_connection_recv_length(
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			int *max_recv_length,
			unsigned int draft_number )
  {
    struct parameter_type	*p;

    TRACE(TRACE_ENTER_LEAVE, "Enter set_connection_recv_length\n");

    if( draft_number >= DRAFT13 )
      {
	if( (p=find_flag_parameter(MAXRECVDATASEGMENTLENGTH_FLAG, p_param_tbl))
				&& IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info) )
	  {/* we did send this key to the other side */
	    *max_recv_length = p->int_value;
	  }
      }
    else if( draft_number >= DRAFT9 )
      {
	if( (p=find_flag_parameter(MAXRECVPDULENGTH_FLAG, p_param_tbl))
				&& IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info) )
	  {/* we did send this key to the other side */
	    *max_recv_length = p->int_value;
	  }
      }

    TRACE(TRACE_ENTER_LEAVE, "Leave set_connection_recv_length\n");

  }


void set_digestflags( struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			 unsigned int *connection_flags )
    {
    struct parameter_type	*p;


    if( (p = find_flag_parameter(HEADERDIGEST_FLAG, p_param_tbl)) )
	{
	if( strcmp(p->str_value, CRC32C) == 0 )
	    {
	    *connection_flags |= USE_HEADERDIGEST;
	    TRACE(TRACE_ISCSI, "Enabling Header Digests\n");
	    }
	}
    if( (p = find_flag_parameter(DATADIGEST_FLAG, p_param_tbl)) )
	{
	if( strcmp(p->str_value, CRC32C) == 0 )
	    {
	    *connection_flags |= USE_DATADIGEST;
	    TRACE(TRACE_ISCSI, "Enabling Data Digests\n");
	    }
	}
    }


/* check that a number is within legal bounds depending on its type */
/* returns 1 if ok, 0 if error */
int check_bounds( struct parameter_type *p, int int_value, int who_called )
  {
    int retval = 1;

    if( IS_ONE_TO_65535(p->type) )
      {
	if( (int_value < 1) || (int_value > 65535) )
	  {
	    TRACE_ERROR("illegal number %d - "
			  "should be between 1 and 65535\n", int_value);
	    retval = 0;
	  }
      }
    else if( IS_ZERO_TO_65535(p->type) )
      {
	if( (int_value < 0) || (int_value > 65535) )
	  {
	    TRACE_ERROR("illegal number %d - "
			  "should be between 0 and 65535\n", int_value);
	    retval = 0;
	  }
      }
    else if( IS_N512_TO_16777215(p->type) )
      {
	if( (int_value < 512) || (int_value > 16777215) )
	  {
	    if( who_called == MANAGEMENT && (p->special_key_flag&
	      		(MAXBURSTSIZE_FLAG|FIRSTBURSTSIZE_FLAG|
			 MAXBURSTLENGTH_FLAG|FIRSTBURSTLENGTH_FLAG)) )
	      {/* Burst limits depend on draft, do it at run time */
	      }
	    else
	      {
		TRACE_ERROR("illegal number %d - "
			  "should be between 512 and 16777215\n", int_value);
		retval = 0;
	      }
	  }
      }
    else if( IS_ZERO_TO_3600(p->type) )
      {
	if( (int_value < 0) || (int_value > 3600) )
	  {
	    TRACE_ERROR("illegal number %d - "
				  "should be between 0 and 3600\n", int_value);
	    retval = 0;
	  }
      }
    else if( IS_TWO_TO_3600(p->type) )
      {
	if( (int_value < 2) || (int_value > 3600) )
	  {
	    TRACE_ERROR("illegal number %d - "
				  "should be between 2 and 3600\n", int_value);
	    retval = 0;
	  }
      }
    else if( IS_ZERO_TO_2(p->type) )
      {
	if( (int_value < 0) || (int_value > 2) )
	  {
	    TRACE_ERROR("illegal number %d - "
				  "should be between 0 and 2\n", int_value);
	    retval = 0;
	  }
      }
    else if( IS_ZERO_TO_32767(p->type) )
      {
	if( (int_value < 0) || (int_value > 32767) )
	  {
	    TRACE_ERROR("illegal number %d - "
			  "should be between 0 and 32767\n", int_value);
	    retval = 0;
	  }
      }

    return retval;
  }


/* this is a range, check both numbers and their order */
/* Returns first number in range if ok, -1 if error */
int check_range( char *value_list, int value )
  {
    char	*endptr;
    int		lower, upper;


    lower = simple_strtoul(value_list, &endptr, 0);
    if( *endptr == '~' )
      {/* first number followed by ~ must be followed by another number */
	upper = simple_strtoul(endptr+1, &endptr, 0);
      }
    else
      {/* only 1 value was specified, which is ok */
        upper = lower;
      }

    if( strspn(endptr, WHITE_SPACE) != strlen(endptr)  ||  lower > upper )
      {
	TRACE_ERROR("illegal range \"%s\"\n", value_list);
	return -1;
      }
    if( value >= 0  &&  (value < lower  ||  value > upper ) )
      {
	TRACE_ERROR("value %d out of range \"%s\"\n", value, value_list);
	return -1;
      }

    return lower;
  }


/************************************************************************/
/* Checks correctness of the type of parameter with the value passed	*/
/* The main types are defined in "text_param.h"				*/
/* If the value is NULL, no checks made..				*/
/* On return, a flag is set in p->neg_info if a fatal error occured	*/
/* Called only from check_correctness().				*/
/************************************************************************/

void check_type_correctness( struct parameter_type *p, char *value,
				int who_called,
				struct key_values *key_table,
				int *int_value )
{
  char *dummy;
  char *endptr = NULL;
  

  TRACE(TRACE_ENTER_LEAVE, "Enter check_type_correctness\n");
  TRACE(TRACE_ENTER_LEAVE, "Parameter: %s, value: %s\n",
						  p->parameter_name, value);

  /* If the value is inquiry, ignore it */
  if( !strcmp(value, "?") )
    goto out;

  /* look for value = "{Rr}eject" or "{Ii}rrelevant" or "NotUnderstood" */
  if( !strcmp(value, key_table->reject) )
  	{
      if( !IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info) )
		{/* Reject always fatal as an offer */
          TRACE_ERROR("illegal offer: %s=%s\n", p->parameter_name, value);
      	  p->neg_info |= KEY_BAD;
		}
      else if( p->special_key_flag & (OFMARKINT_FLAG | IFMARKINT_FLAG) )
		{/* Reject is not fatal as a reply to OFMarkInt/IFMarkInt keys */
	 	/* Draft 13, Section A.3.2 OFMarkInt, IFMarkInt
	    	"When the interval is unacceptable the responder answers with
	    	"Reject".  Reject is resetting the marker function in the
	    	specified direction (Output or Input) to No."
	 	*/
		  p->neg_info |= KEY_REJECT;	/* so we can do the reset later */
		}
      else
    	{/* Reject always fatal as a reply to all other keys */
          TRACE_ERROR("got key: %s=%s\n", p->parameter_name, value);
      	  p->neg_info |= KEY_BAD;
	    }
      goto out;
    }

  if( !strcmp(value, key_table->irrelevant) )
    {
      if( !IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info) )
    	{/* Irrelevant always fatal as an offer */
          TRACE_ERROR("illegal offer: %s=%s\n", p->parameter_name, value);
      	  p->neg_info |= KEY_BAD;
		}
      else
		{/* Irrelevant is a legal reply in all negotiations */
	  	  TRACE(TRACE_ISCSI, "***Warning***: got key: %s=%s\n",
						    p->parameter_name, value);
	  	  p->neg_info |= KEY_IRRELEVANT;
		}
      goto out;
    }

  /*  Draft 13, Section 4.2 Text Mode Negotiation
      "All keys in Chapter 11, except for the X- extension format, MUST be
      supported by iSCSI initiators and targets and MUST NOT be answered with
      NotUnderstood."
  */
  if( !strcmp(value, "NotUnderstood") )
    {
      if( !IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info) )
    	{/* NotUnderstood always fatal as an offer */
        	TRACE_ERROR("illegal offer: %s=%s\n", p->parameter_name, value);
      	  	p->neg_info |= KEY_BAD;
		}
      else if( (p->special_key_flag & X_EXTENSIONKEY_FLAG) )
        {/* target not expected to understand this anyway, consider its
	    	response irrelevant */
	  		p->neg_info |= KEY_IRRELEVANT;
		}
  	  else
		{/* NotUnderstood always fatal as a reply to a standard key */
	  		TRACE_ERROR("standard key \"%s\" MUST NOT be answered with "
				  "\"NotUnderstood\"\n", p->parameter_name);
	  		p->neg_info |= KEY_BAD;
		}
      goto out;
   }

  if( IS_STRING(p->type) )
    {/*  key value is a string, just check its total length */

      /* Draft 19, Section 5.1 Text Format
 		"If not otherwise specified, the maximum length of a simple-value
		(not its encoded representation) is 255 bytes not including the
		delimiter (comma or zero byte)."
      */
	 /* Draft 19, Section 3.2.6.1  iSCSI Name Properties
	    "Each iSCSI node, whether an initiator or target, MUST have an iSCSI
		    name. Initiators and targets MUST support the receipt of iSCSI 
			names of up to the maximum length of 223 bytes. "
	  */ 
	  char *tptr=NULL;
      int len = strlen(value);
	  int t_num;
	  
	  if( len > MAX_KEY_VALUE_LENGTH )
	  {
	  	TRACE_ERROR("value of key \"%s\" longer than %d characters\n",
				p->parameter_name, (IS_ISCSI_NAME(p->type)?:
				MAX_ISCSI_NAME_LENGTH,MAX_KEY_VALUE_LENGTH));
	 	p->neg_info |= KEY_BAD;
	  }
	  
      /* check ISCSI_NAME type */
	  if ( IS_ISCSI_NAME(p->type) )
	  {
		if( who_called == TARGET && p->special_key_flag & TARGETNAME_FLAG )
		{/* check TargetName sent by the initiator */
			if( (strncmp(TARGETNAME_HEADER, value,strlen(TARGETNAME_HEADER))!= 0) 
				||(tptr=strchr(value, ':')) == NULL)
			{
				TRACE_ERROR("***Warning***:Invalid target name \"%s\". Please use Discovery "
			 	"session to get the proper target name\n", value);
				p->neg_info |= KEY_WRONG;
				goto out;
			}			
			t_num = simple_strtoul( tptr+1, &endptr, 0);

		  	if( strspn(endptr, WHITE_SPACE) != strlen(endptr) )
	  		{
	  			TRACE_ERROR("illegal target number \"%s\" in the target name \"%s\"\n", tptr+1, value);	 
		  		p->neg_info |= KEY_BAD;
				goto out;
			}
			/* check target number from the target name */ 
	  		if( t_num < 0 || ((t_num >= target_num) && t_num !=0) )
	  		{
				TRACE_ERROR("target number \"%d\" is out of range-- 0 to %d\n", t_num, target_num);
	  			*(tptr+1) = '0'; /* use 0 as default */
			}
		}
		/* check the format of the iSCSI name */
		/* Draft 19, 3.2.6.2  iSCSI Name Encoding
		  "The type designator strings currently defined are:
		           iqn.       - iSCSI Qualified name
		           eui.       - Remainder of the string is an IEEE EUI-64
		                        identifier, in ASCII-encoded hexadecimal."
		    "One of these two type strings MUST be used when constructing
		    an iSCSI name; any type string not listed here is not allowed,
		    as they cannot be guaranteed to be unique."
		 */  
		
		if( len == 20  &&  strncmp(value, "eui.", 4) == 0 )
		{/* IEEE format world wide unique name -- 16 hex digits must follow */
			if( strspn(value+4, "0123456789abcdefABCDEF") == 16 )
				/* ok, there are exactly 16 hex digits following the "eui." */
				goto out;
		}
		else if ( len > 12  &&  len <= MAX_ISCSI_NAME_LENGTH&&
						strncmp(value,"iqn.", 4) == 0 )
		{/* iSCSI qualified name -- must get "yyyy-mm." then alphanums, ., - */
			if( strspn(value+4, "0123456789") == 4  &&
				*(value+8) == '-'  &&
				strspn(value+9, "0123456789") == 2  &&
				*(value+11) == '.'  &&
				strspn(value+12,"0123456789abcdefghijklmnopqrstuvwxyz-.:")==len-12 )
		/* ok, have iqn.yyyy-mm.xxx with xxx all digits, lower-case letters, 
			 * '-', '.', ':' */
				goto out;
		}
		 /* should reach here. If so, invalid iSCSI name */
		TRACE_ERROR("\"%s\" is not a valid iSCSI name\n", value );  
		goto out;
	  }
    }

  /**************************************************************/
  /* the following are to be tested for each value in the list  */

  do
    {
      endptr = NULL;

      /* extract just a single value */
      dummy = strchr( value, ',' );
      if( dummy )
	  {/* found a comma, so value contains a list of values -- is this ok? */
	  	if( !IS_KEY_MULTI_VALUE(p->type) && !(p->special_key_flag & TARGETADDRESS_FLAG))
	    {/* this key cannot accept a list of values */
	    	TRACE_ERROR("key \"%s\" does not accept a list of values "
				     "\"%s\"\n", p->parameter_name, value);
	      	p->neg_info |= KEY_BAD;
	      	break;
	    }
	  	/* terminate the first item in the list so it is a string by itself */
	  	*dummy = '\0';
	  }

      TRACE(TRACE_DEBUG, "value: %s\n", value);

      /* Check for length of the key value. */
      /* Draft 13, Section 4.1 Text Format
	  	"If not otherwise specified, the maximum length of a simple-value
		(not its encoded representation) is 255 bytes not including the
		delimiter (comma or zero byte)."
      */
      if( strlen(value) > MAX_KEY_VALUE_LENGTH )
	  {
	  	TRACE(TRACE_ISCSI,
		"***Warning***: value of key \"%s\" exceeds %d characters\n",
		p->parameter_name, MAX_KEY_VALUE_LENGTH);
	  }

      if( *value == '\0'  &&  !IS_KEY_NO_VALUE(p->type) )
	  {
	  	TRACE_ERROR("no value after '=' for key \"%s\"\n", p->parameter_name);
	    p->neg_info |= KEY_BAD;
	    break;
	  }

      else if( IS_NUMBER(p->type)  ||  (IS_NUMBER_RANGE(p->type) &&
  					IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)))
	  {/* a number or a reply to a range */

	  	*int_value = simple_strtoul(value, &endptr, 0);

	  	if( strspn(endptr, WHITE_SPACE) != strlen(endptr) )
	  	{
	  		TRACE_ERROR("illegal number \"%s\"\n", value);
	  		p->neg_info |= KEY_BAD;
	    	break;
	  	}

	  	if( !check_bounds(p, *int_value, who_called) )
	  	{
	  		p->neg_info |= KEY_BAD;
	  		break;
	  	}
	  }

      else if( IS_BOOLEAN(p->type) )
      {
	  	if( strcmp(value, key_table->yes) && strcmp(value, key_table->no) )
	    {
	      TRACE_ERROR("illegal value \"%s\" - expected "
			"\"%s\" or \"%s\"\n",
			value, key_table->yes, key_table->no);
	      p->neg_info |= KEY_BAD;
	      break;
	    }
	  }

      else if( IS_ENUMERATED(p->type) )
	  {
	  	if( IS_AUTH_PARAM(p->type) )
	    {
	      if( strcmp(value, KRB5) && strcmp(value, SPKM1) &&
		  		strcmp(value, SPKM2) && strcmp(value, SRP)  &&
		  		strcmp(value,CHAP) && strcmp(value,key_table->none))
		  {
		  		TRACE_ERROR("illegal value \"%s\" - expected "
		  			"\"KRB5\" or "
					"\"SPKM1\" or \"SPKM2\" or \"SRP\" or "
					"\"CHAP\" or \"%s\"\n",
					value, key_table->none);
		  		p->neg_info |= KEY_BAD;
		  		break;
		  }
	  	}
	 	else if( IS_SEND_RECEIVE(p->type) )
	  	{
	  		if( strcmp(value, SEND) && strcmp(value, RECEIVE) &&
		  		strcmp(value, SENDRECEIVE) &&
		  		strcmp(value, key_table->no ) )
			{
		  		TRACE_ERROR("illegal value \"%s\" - expected "
						"\"send\" or \"receive\" or "
						"\"send-receive\" or \"%s\"\n",
						value, key_table->no);
		  		p->neg_info |= KEY_BAD;
		  		break;
			}
	  	}
	  	else if( IS_DISCOVERY_NORMAL(p->type) )
	  	{
	  		if( strcmp(value, key_table->discovery) &&
				strcmp(value, key_table->normal) )
			{
		  		TRACE_ERROR("illegal value \"%s\" - expected "
		  			"\"%s\" or \"%s\"\n",
					value, key_table->discovery,
					key_table->normal);
		  		p->neg_info |= KEY_BAD;
		  		break;
			}
	  	}
	}

    else if( IS_NUMBER_RANGE(p->type)  &&
  				!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info) )
    {/* is an offer of a numeric range */
		if( (*int_value = check_range(value, -1)) < 0  ||
				!check_bounds(p, *int_value, who_called) )
	  	{
	    	p->neg_info |= KEY_BAD;
	   		break;
	  	}
	}

    if( dummy )
	{/* restore the comma and goto the next value */
	  *dummy++ = ',';
	}

    value = dummy;
  }
  while( value );

  /**************************************************************/

out:
  TRACE(TRACE_ENTER_LEAVE, "Leave check_type_correctness\n");

  return;
}


/************************************************************************/
/* Checks the correctness of the passed string, checks whether it	*/
/* conforms to the <key>=<value> format					*/
/* extracts the value from the input string and returns it as an	*/
/* output parameter.							*/
/* Returns pointer to parameter in p_param_tbl if ok, else NULL		*/
/* (but on non-NULL p->neg_info can have flag set on fatal error)	*/
/************************************************************************/

struct parameter_type *check_correctness( char *keytext, char **p_value,
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			unsigned int who_called,
  			struct key_values *key_table,
			unsigned int when_called, int *int_value,
			struct unknown_key **unknown_key_list )
{
  char *value;
  struct parameter_type *p = NULL;
  struct unknown_key	*uptr;
  unsigned int		count;


  TRACE(TRACE_ENTER_LEAVE, "Enter check_correctness\n");

  TRACE(TRACE_ISCSI, "Got key: %s\n", keytext);


  if( (value = strchr(keytext, '=' )) == NULL )
    {/* could not find '=', don't accept this */
      TRACE_ERROR("key \"%s\" not followed by '='\n", keytext);
      goto out;
    }

  /* terminate the key name string and point at the value string */
  *value++ = '\0' ;


  /* Check for length of the key name. */

  /*  Draft 13, Section 4.1 Text Format
      "A key-name MUST begin with a capital letter and must not exceed 63
      characters."
  */
  if( strlen(keytext) > MAX_KEY_NAME_LENGTH )
    {
      TRACE(TRACE_ISCSI,
			"***Warning***: length of key name \"%s\" exceeds %d\n",
			keytext, MAX_KEY_NAME_LENGTH);
    }

  /*  Draft 13, Section 9.12.11 Login Parametersr
      "All keys in Chapter 11, except for the X- extension format, MUST be
      supported by iSCSI initiators and targets.  Keys in Chapter 10 only
      need to be supported when the function to which they refer is mandatory
      to implement."
  */
  if( !(p = find_parameter(keytext, p_param_tbl)) )
    {/* key not found in our table,
	look for value = "{Rr}eject" or "{Ii}rrelevant" or "NotUnderstood" */
      if( !strcmp(value, key_table->reject)  ||
	  !strcmp(value, key_table->irrelevant)  ||
	  !strcmp(value, "NotUnderstood") )
    	{/* These values always fatal as a value for an offer of unknown key */
          TRACE_ERROR("unknown key with illegal value: %s=%s\n", keytext,
									value);
	}
      else
	{/* value not bad, but have we seen this unknown key before? */
	  for(	count = 0, uptr = *unknown_key_list;
		uptr != NULL;
		count++, uptr = uptr->next )
	    {
	      if( !strcmp(keytext, uptr->keyname) )
		{/* yes, this unknown key is being negotiated twice */
		  TRACE_ERROR("unknown key \"%s\" received twice\n", keytext);
		  goto out;
		}
	    }

	  /* never saw this key before, should we save it for later checking? */
	  /* chap support = CHONG */
	  if (!is_securitykey(keytext))
	    {
	      TRACE(TRACE_ISCSI,"***Warning***: unknown key \"%s\"\n", keytext);
	    }

	  if( count >= MAX_UNKNOWN_KEYS )
	    {/* no, we have seen too many unknown keys, bail out */
	      TRACE_ERROR("%u unknown keys received, too many!\n", count);
	      goto out;
	    }

	  else if( (uptr=my_kmalloc(sizeof(struct unknown_key), "Unknown key"))
								== NULL )
	    {
	      goto out;
	    }
      
	  /* chap support - CHONG */
	  else if( (uptr->keyname = my_kmalloc(strlen(keytext)+1, "Key Name"))
								== NULL )
	    {
	      my_kfree((void **)&uptr, "Unknown key");
	      goto out;
	    }

	  else if( (uptr->keyvalue = my_kmalloc(strlen(value)+1, "Key Value"))
								== NULL )
	    {
	      my_kfree((void **)&uptr->keyname, "Key Name");
	      my_kfree((void **)&uptr, "Unknown key");
	      goto out;
	    }
	  else
	    {
	      strcpy(uptr->keyname, keytext);
	      strcpy(uptr->keyvalue, value);
	      uptr->processed = 0;
	      uptr->next = *unknown_key_list;
	      *unknown_key_list = uptr;

	      /* pass back Notunderstood value for unknown key */
	      *p_value = "NotUnderstood";
	    }
	}
      goto out;
    }

  if( (who_called==INITIATOR  &&  !IS_USE_BY_TARGET(p->type))  ||
      (who_called==TARGET  &&  !IS_USE_BY_INITIATOR(p->type)) )
    {
      TRACE_ERROR("key \"%s\" cannot be sent to %s\n", keytext,
			who_called == INITIATOR ? "initiator" : "target");
      p->neg_info |= KEY_BAD;
      goto out;
    }

  if( IS_KEY_GOT_FROM_OTHER_SIDE(p->neg_info) )
    {/* attempting to negotiate this key twice */
    /*  Draft 13, Section 4.3 Login Phase
	"Neither the initiator nor the target should attempt to negotiate a
	parameter more than once during login.  If detected by the target
	this MUST result in a Login reject (initiator error).  The initiator
	MUST drop the connection."
    */
      	TRACE_ERROR("key \"%s\" received twice\n", p->parameter_name);
     	p->neg_info |= KEY_BAD;
      	goto out;
    }

  if( !IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info) )
    {/* this is a new offer from other side, not a reply to one of our offers */
      if( !(p->type & when_called) )
        {/* wrong time to be negotiating this parameter */
	  TRACE_ERROR("key \"%s\" cannot be negotiated now\n", keytext);
	  p->neg_info |= KEY_BAD;
	  goto out;
	}
    }

  check_type_correctness(p, value, who_called, key_table, int_value);

  /* Return the value pointer on success */
  *p_value = value;

out:

  TRACE(TRACE_ENTER_LEAVE, "Leave check_correctness, p %p\n", p);
  return p;
}


char *strdup (char *str)
{
  int n = strlen(str)+1;

  char *s = kmalloc(n, GFP_KERNEL);

  if (!s) return NULL;
  return strcpy(s, str);
}

void strreplace( char **str, char *new_str )
{
  if( *str != NULL )
    kfree(*str);
  *str = strdup(new_str);
}


/* Copy src parameter table to dst, duplicating any strings */
void param_tbl_cpy( struct parameter_type dst[MAX_CONFIG_PARAMS],
		    struct parameter_type src[MAX_CONFIG_PARAMS] )
{
  struct parameter_type	*dptr;
  int			i;


  TRACE(TRACE_ENTER_LEAVE, "Enter param_tbl_cpy, dst %p, src %p, size %d\n",
		dst, src, sizeof(struct parameter_type)*MAX_CONFIG_PARAMS);

  /* copy everything in the src table to the dst table "as is" */
  memcpy(dst, src, sizeof(struct parameter_type)*MAX_CONFIG_PARAMS);

  /* now go back and "dup" all the strings */
  for( i = 0, dptr = dst;  i < MAX_CONFIG_PARAMS;  i++, dptr++ )
    {
    if( dptr->parameter_name )
      {
	dptr->parameter_name = strdup(dptr->parameter_name);
      }
    if( dptr->str_value )
      {
	dptr->str_value = strdup(dptr->str_value);
      }
    if( dptr->value_list )
      {
	dptr->value_list = strdup(dptr->value_list);
      }
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave param_tbl_cpy\n");
}


/* Free any strings referenced in dst parameter table */
void param_tbl_uncpy( struct parameter_type dst[MAX_CONFIG_PARAMS] )
{
  struct parameter_type	*dptr;
  int			i;


  TRACE(TRACE_ENTER_LEAVE, "Enter param_tbl_uncpy, dst %p, size %d\n",
		dst, sizeof(struct parameter_type)*MAX_CONFIG_PARAMS);

  /* go thru and free all the strings that were "duped" during copy */
  for( i = 0, dptr = dst;  i < MAX_CONFIG_PARAMS;  i++, dptr++ )
    {
    if( dptr->parameter_name )
      {
        TRACE(TRACE_DEBUG, "%2d kfree(%s)\n", i, dptr->parameter_name);
	kfree(dptr->parameter_name);
      }
    if( dptr->str_value )
      {
	kfree(dptr->str_value);
      }
    if( dptr->value_list )
      {
	kfree(dptr->value_list);
      }
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave param_tbl_uncpy\n");
}


void configure_parameter( int param_neg_info,
			  char *ptr_to_keytext,
			  struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS] )
{
  struct parameter_type *param = NULL;
  char *value_list = NULL;
  char *endptr;
  int	int_value;


  TRACE(TRACE_ENTER_LEAVE, "Enter configure_parameter\n");

  /* Get the value string */
  value_list = strchr(ptr_to_keytext, '=');
  if( value_list )
    *value_list++ = '\0';	/* terminate the key text string and skip */

  if( (param = find_parameter(ptr_to_keytext, p_param_tbl)) != NULL )
    {
      if( value_list )
	{
	  if( IS_NUMBER(param->type) )
	    {/* parameter is a number, set only the int_value field,
		don't set value_list or str_value fields */

	      int_value = simple_strtoul(value_list, &endptr, 0);
	      if( strspn(endptr, WHITE_SPACE) != strlen(endptr) )
		{
		  TRACE_ERROR("illegal number \"%s\"\n", value_list);
		  goto out;
		}
	      else
		{
		param->int_value = int_value;
		}
	  }
	  else
	    {/* parameter is a string, enumerated, boolean or range,
	   	 set both the value_list and str_value fields,
		 do not set int_value field */

	      if( IS_NUMBER_RANGE(param->type) )
	        {/* this is a range, check both numbers and their order */
		  if( check_range(value_list, -1) < 0 )
		      goto out;
		}

	      strreplace(&param->value_list, value_list);
	      endptr = strchr(value_list, ',');
	      if( endptr )
		{/* value_list really a list, use 1st on list as current value*/
		  *endptr = '\0';
		  strreplace(&param->str_value, value_list);
		  *endptr = ',';	/* repair the damage */
		}
	      else
		{/* value_list only single value, make it current val too*/
		  strreplace(&param->str_value, value_list);
		}
	    }
	}

      /* set the negotiation info only if no error */
      param->neg_info = param_neg_info;
    }

out:
  TRACE(TRACE_ENTER_LEAVE, "Leave configure_parameter\n");
}

/* in draft18, FirstBurstSize and MaxBurstSize are renamed to FirstBurstLength and MaxBurstLength */
void fix_burst_name( unsigned int draft_number,  unsigned long long *max_burst, unsigned long long *first_burst)
{
 if( draft_number <= DRAFT13 )
 {
     *max_burst = MAXBURSTSIZE_FLAG;
     *first_burst = FIRSTBURSTSIZE_FLAG;
 }
 else
 {
    *max_burst = MAXBURSTLENGTH_FLAG;
    *first_burst = FIRSTBURSTLENGTH_FLAG;
 }
}

/*	Draft 13, Section 4.2 Text Mode Negotiation
	"Some parameters may be subject to integrity rules (e.g., parameter-x
	must not exceed parameter-y or parameter-y not 1 implies parameter-v
	to by Yes).  Whenever required integrity rules are specified with the
	keys.  Checking for compliance with the integrity rule MUST NOT be
	performed before all the negotiation parameters are available (the
	existent and newly negotiated).  An iSCSI target MUST perform integ-
	rity checking before committing new values for parameters.  An
	initiator MAY perform integrity checking."
*/
void check_integrity_rules(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
						unsigned int draft_number )
  {
    struct parameter_type	*p, *p2;
    struct key_values		*key_table;
    unsigned long long          max_burst,  first_burst;


    TRACE(TRACE_ENTER_LEAVE, "Enter check_integrity_rules\n");

    if( draft_number <= DRAFT9 )
      key_table = &lower_case_values;
    else
      key_table = &upper_case_values;

    /*	Draft 13, Section A.3.2 OFMarkInt, IFMarkInt
	"Reject is resetting the marker function in the specified
	direction (Output or Input) to No."
    */
    if( (p = find_flag_parameter(OFMARKINT_FLAG, p_param_tbl)) != NULL )
      {
	if( p->neg_info & KEY_REJECT )
	  {/* got OFMarkInt=Reject, so set OFMarker=No */
	    TRACE(TRACE_DEBUG, "Checking %s=%d\n", p->parameter_name,
								p->int_value);
	    if( (p = find_flag_parameter(OFMARKER_FLAG, p_param_tbl)) != NULL )
	      {
		TRACE(TRACE_DEBUG, "Have %s=%s\n", p->parameter_name,
								p->str_value);
		if( !strcmp(p->str_value, key_table->yes) )
		  {/* have OFMarker=Yes, change value to No */
		    strreplace(&p->str_value, key_table->no);
		    TRACE(TRACE_ISCSI, "Reset %s to %s\n",
					    p->parameter_name, p->str_value);
		  }
	      }
	  }
      }
    if( (p = find_flag_parameter(IFMARKINT_FLAG, p_param_tbl)) != NULL )
      {
	if( p->neg_info & KEY_REJECT )
	  {/* got IFMarkInt=Reject, so set IFMarker=No */
	    TRACE(TRACE_DEBUG, "Checking %s=%d\n", p->parameter_name,
								p->int_value);
	    if( (p = find_flag_parameter(IFMARKER_FLAG, p_param_tbl)) != NULL )
	      {
		TRACE(TRACE_DEBUG, "Have %s=%s\n", p->parameter_name,
								p->str_value);
	        if( !strcmp(p->str_value, key_table->yes) )
		  {/* have IFMarker=Yes, change value to No */
		    strreplace(&p->str_value, key_table->no);
		    TRACE(TRACE_ISCSI, "Reset %s to %s\n",
					    p->parameter_name, p->str_value);
		  }
	      }
	  }
      }

    /*  Draft 13, Section 11.15 FirstBurstSize
    	"FirstBurstSize MUST NOT exceed MaxBurstSize."
    */
    /* in draft18, FirstBurstSize and MaxBurstSize are renamed to FirstBurstLength and MaxBurstLength */
   fix_burst_name(draft_number, &max_burst, &first_burst);
   
   if( (p = find_flag_parameter(first_burst, p_param_tbl)) != NULL )
      {
    	if( (p2 = find_flag_parameter(max_burst, p_param_tbl)) != NULL )
	  {
	    if( p->int_value > p2->int_value )
	      {
	      TRACE(TRACE_ISCSI, "***Warning***: FirstBurstLength %u exceeds "
			      "MaxBurstLength %u\n", p->int_value, p2->int_value);
	      p->int_value = p2->int_value;	/* fix FirstBurstLength */
	      }
	  }
      }

    /*	Draft 13, Section 11.22 SessionType
    	"The discovery session implies MaxConnections = 1 and overrides both
	the default and an explicit setting."
    */
    if( (p = find_flag_parameter(SESSIONTYPE_FLAG, p_param_tbl)) != NULL )
      {
        if( !strcmp(p->str_value, key_table->discovery)  &&
    	    (p2=find_flag_parameter(MAXCONNECTIONS_FLAG,p_param_tbl)) != NULL )
	  {
	    if( p2->int_value != 1 )
	      {/* do not have MaxConnections=1, change value to 1 */
		p2->int_value = 1;
		TRACE(TRACE_ISCSI, "Reset %s to %u\n",
					p2->parameter_name, p2->int_value);
	      }
	  }
      }

    /*	Draft 13, Section 11.20 DataSequenceInOrder
    	"If DataSequenceInOrder is set to Yes, a target may retry at most the
	last R2T, and an initiator may at most request retransmission for the
	last read data sequence.  For this reason if ErrorRecoveryLevel is
	not 0 and DataSequenceInOrder is set to Yes then MaxOutstandingR2T
	MUST be set to 1."
    */
    if( (p=find_flag_parameter(DATASEQUENCEINORDER_FLAG,p_param_tbl)) != NULL )
      {
	if( !strcmp(p->str_value, key_table->yes)  &&
	    (p2=find_flag_parameter(ERRORRECOVERYLEVEL_FLAG,p_param_tbl))!=NULL)
	  {/* have DataSequenceInOrder=Yes */
	    if( p2->int_value > 0  &&
	    	(p=find_flag_parameter(MAXOUTSTANDINGR2T_FLAG,p_param_tbl))
									!=NULL )
	      {/* have ErrorRecoveryLevel>0 */
	        if( p->int_value != 1 )
		  {/* do not have MaxOutstandingR2T = 1, change it to 1 */
	          TRACE(TRACE_ISCSI, "***Warning***: %s=%u reset "
						  "to 1\n", p->parameter_name, p->int_value);
		  p->int_value = 1;
	          }
	      }
	  }
      }

    TRACE(TRACE_ENTER_LEAVE, "Leave check_integrity_rules\n");
  }


/************************************************************************/
/* This function receives 'length' bytes into buffer from the socket */
/************************************************************************/

int iscsi_recv_msg( struct socket *sock, int length, char *buffer, int flags )
{
  struct msghdr msg;
  size_t size ;
  int received, retval;
  struct iovec iov;
  char address[MAX_SOCK_ADDR];
  mm_segment_t oldfs;

  TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_recv_msg, length %d bytes\n", length);

  /* following code will receive the PDU's from the other side */
  received = 0;
  size = length;
  while (received < length)
    {
      msg.msg_control=NULL;
      msg.msg_controllen=0;
      msg.msg_iovlen=1;
      msg.msg_iov=&iov;
      iov.iov_len=size;

      iov.iov_base = buffer + received;

      msg.msg_name=address;
      msg.msg_namelen=MAX_SOCK_ADDR;
      msg.msg_flags = MSG_WAITALL;

      oldfs = get_fs(); set_fs(get_ds());
      retval = sock_recvmsg(sock, &msg, size, 0);
      set_fs(oldfs);

      if( signal_pending(current) || retval <= 0 )
	{
	  TRACE_ERROR("sock_recvmsg error %d, signal_pending %x\n",
	 				 retval, signal_pending(current));
	  retval= -1;
	  goto out;
	}
      received += retval;
      size = length - received;

      TRACE(TRACE_DEBUG, "Received: %d, total_rx: %d, data: %d\n",
     						 retval, received, length);
    }

  TRACE(TRACE_BUFF, "Got buffer, length: %d\n", length);
  PRINT_BUFF(buffer, length);
  retval = 0;

out:
  TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_recv_msg, retval %d\n", retval);

  return retval;
}


/*******************************************************************/
/* returns 1 if needs a break from the loop */
/* else returns a 0 */
/*******************************************************************/
int check_out_length( int out_length, int resp_len )
{

  TRACE(TRACE_ENTER_LEAVE, "Enter check_out_length cur_len: %d, add_len: %d\n",
							out_length, resp_len);

  /* Increment the out_length */
  out_length += resp_len + 2;

  TRACE(TRACE_ENTER_LEAVE, "Leave check_out_length\n");

  if( out_length > MAX_TEXT_LEN )
    return 1;
  else
    return 0;
}


/*******************************************************************/
/* sends the PDUs to the other end. */
/*******************************************************************/

int iscsi_send_msg( struct socket *sock, unsigned int draft_number,
				struct generic_pdu *outputpdu, int flags )
{
  struct msghdr			msg;
  struct iovec			*iov, actual_iov[3];
  mm_segment_t			oldfs;
  struct iscsi_targ_login_rsp	*targ_login_rsp;

  int				total_tx, tx_loop;
  int				data_length, count, hdr_length, send_length;
  unsigned int			data_checksum;


  TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_send_msg \n");

  switch( outputpdu->opcode & (ISCSI_OPCODE) )
    {

  case ISCSI_INIT_LOGIN_CMND:

      /* Send the login Command */
      if( TRACE_TEST(TRACE_ISCSI_FULL) )
  	print_init_login_cmnd((struct iscsi_init_login_cmnd *)outputpdu);
      break;


  case ISCSI_TARG_LOGIN_RSP:

      /* Send the Login Response */
      targ_login_rsp = (struct iscsi_targ_login_rsp *)outputpdu;

      TRACE(TRACE_ISCSI, "Send Login Response\n");
      if( TRACE_TEST(TRACE_ISCSI_FULL) )
          print_targ_login_rsp(targ_login_rsp);
      break;

  default:
      TRACE_ERROR("opcode 0x%08X not allowed during Login phase\n",
		  					outputpdu->opcode);
      return -1;

    }	/* switch */

  /* all pdus store the DSL in the same place */
  outputpdu->length = cpu_to_be32(outputpdu->text_length);

  /* set up the i/o message and vectors */
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  msg.msg_iov = iov = actual_iov;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_flags = 0;

  count = 1;
  hdr_length = ISCSI_HDR_LEN;


  /* NOTE: in draft 9 and after, no packets sent or received during login phase
 	 will have digests of any kind.
	 However, in draft 8 and before, packets sent after the security
	 negotiation stage may have digests.
  */
  if( flags & USE_HEADERDIGEST )
    {/* header digests are in use, compute the crc and add it to header */
      do_crc((unsigned char *)outputpdu, ISCSI_HDR_LEN, 3,
					      	&outputpdu->header_digest);
      hdr_length += CRC_LEN;
      TRACE(TRACE_ISCSI_FULL, "Attach HeaderDigest\n");
    }

  iov->iov_base = (char *)outputpdu;
  iov->iov_len = send_length = hdr_length;

  if( (data_length = outputpdu->text_length) > 0 )
      {/* there is data attached to this pdu */

      /* Add padding to end of any attached data */
      data_length += (-data_length) & 3;

      if( data_length > outputpdu->text_length )
	{
          TRACE(TRACE_DEBUG, "Length with padding = %d, without = %d\n",
					data_length, outputpdu->text_length);
	  memset(outputpdu->text+outputpdu->text_length, 0,
					data_length-outputpdu->text_length);
	}

      count++;
      iov++;
      iov->iov_base = outputpdu->text;
      iov->iov_len = data_length;
      send_length += data_length;

      if( flags & USE_DATADIGEST )
      	{
	do_crc(iov->iov_base, iov->iov_len, 3, &data_checksum);
        count++;
	iov++;
	iov->iov_base = &data_checksum;
	iov->iov_len = CRC_LEN;
	send_length += CRC_LEN;
	TRACE(TRACE_ISCSI_FULL, "Attach DataDigest\n");
	}
      }

  /* i/o message and vector now set up */
  msg.msg_iovlen = count;

  TRACE(TRACE_DEBUG, "Sending %d bytes total, %d actual_iovs\n",
							send_length, count);

  total_tx = 0;
  do
    {
      oldfs = get_fs(); set_fs(get_ds());

      TRACE(TRACE_NET, "Sending %d bytes, %d actual_iovs on sock %p\n",
						      send_length, count, sock);
      tx_loop = sock_sendmsg(sock, &msg, send_length);

      set_fs(oldfs);

      if( tx_loop <= 0 )
	{
	  TRACE_ERROR("sock_sendmsg error %d, total_tx %d\n", tx_loop,
								total_tx );
	  return tx_loop;
	}
      total_tx += tx_loop;

      TRACE(TRACE_NET, "tx_loop = %d, total_tx: %d, data: %d\n",
						tx_loop, total_tx, send_length);
      if( total_tx != send_length )
      	{
	TRACE_ERROR("sock_sendmsg total_tx %d, expected %d\n",
							total_tx, send_length);
	}
    }
  while( total_tx < send_length );

  TRACE(TRACE_BUFF, "Sent buffer, length: %d\n", send_length);

  PRINT_BUFF((char *)outputpdu, hdr_length);
  if( count > 1 )
    {
    PRINT_BUFF((char *)outputpdu->text, data_length);
    if( count > 2 )
      PRINT_BUFF((char *)&data_checksum, CRC_LEN);
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_send_msg\n");

  return 0;
}


/************************************************************************/
/* checks whether the receiving side has recevied all the responses */
/* for the negotiations it started in the last packet. */

/* Return: */
/* 		-1 if error, didn't receive the response */
/* 		0 if success */
/************************************************************************/

int check_neg_responses(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			unsigned int print_error )
{
  int i=0;
  int retval = 0;
  struct parameter_type *p = NULL;

  TRACE(TRACE_ENTER_LEAVE, "Enter check_neg_responses\n");

  for( i=0 ; i < MAX_CONFIG_PARAMS ; i++ )
    {
      p = &p_param_tbl[i];

      if( IS_KEY_TO_BE_NEGOTIATED(p->neg_info)	&&
	  IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)  &&
	  !IS_KEY_GOT_FROM_OTHER_SIDE(p->neg_info)  &&
	  !IS_KEY_REPLY_OPTIONAL(p->neg_info) )
	{/* parameter was sent, no reply was received and reply not optional */
	  if( !IS_INFORMATIONAL_PARAM(p->type)  &&
	!(p->special_key_flag &
			(MAXRECVPDULENGTH_FLAG|MAXRECVDATASEGMENTLENGTH_FLAG)) )
	    {/* informational parameters expect no response */
	      /* and neither do MaxRecvPDULength/MaxRecvDataSegmentLength */
	      if( print_error )
		{
		TRACE_ERROR("response expected for parameter %s\n",
			    p->parameter_name );
		/* set got bit so we won't get this message again */
		p->neg_info |= KEY_GOT_FROM_OTHER_SIDE;
		}
	      retval = -1;
	    }
	}
    }


  TRACE(TRACE_ENTER_LEAVE, "Leave check_neg_responses, retval = %d\n", retval);

  return retval;
}


/************************************************************************/
/* checks whether the receiving side supports the parameter values      */
/* offered by the sender.                                               */
/* NOTE that the type of this parameter will never be Boolean or Number	*/
/* (because in this case this function is not called)			*/
/* 				                                        */
/* Return:                                                              */
/* 	If supported returns the pointer to the (terminated) value in	*/
/*	the supplied_values_from_sender list, else returns NULL		*/
/************************************************************************/

char *check_for_support( struct parameter_type *p,
			 char *supplied_values_from_sender,
			 unsigned int draft_number )
{
  char *dummy1 = NULL;
  char *dummy2 = NULL;
  char *sender_value = NULL;
  char *receiver_value = p->value_list;


  TRACE(TRACE_ENTER_LEAVE, "Enter check_for_support, sender_value %s\n",
						supplied_values_from_sender);

  /* No support */
  if( receiver_value == NULL )
    goto out;

  TRACE(TRACE_DEBUG, "%s's value_list: %s\n",p->parameter_name,receiver_value);

  /* Check whether the parameter's str_value agrees with *
   * one of the supplied values */

  do
    {
      /* Get the receiver_value from the parameter's value_list */
      dummy1 = strchr( receiver_value, ',');
      if( dummy1 )
	{
	  *dummy1 = '\0';
	}

      TRACE(TRACE_DEBUG, "receiver_value: %s\n", receiver_value);

      /* Initialise */
      sender_value = supplied_values_from_sender;

      do
	{
	  /* extract just a single value */
	  dummy2 = strchr( sender_value, ',' );
	  if( dummy2 )
	    {
	      *dummy2 = '\0' ;
	    }

	  TRACE(TRACE_DEBUG, "sender_value: %s\n", sender_value);

	  if( ! strcmp(receiver_value, sender_value) )
	    {
	      /* Found a match, return pointer to it in sender_value */
	      if( dummy1 )
		*dummy1 = ',';		/* restore the comma before leaving */
	      goto out;
	    }

	  /* goto the next value supplied by sender */
	  if( dummy2 )
	    {
	      dummy2++;
	      TRACE(TRACE_DEBUG, "dummy2++sender_value: %s\n", sender_value);
	    }
	  else
	    {
	      TRACE(TRACE_DEBUG,"dummy2 null sender_value: %s\n",sender_value);
	    }
	  sender_value = dummy2;
	}
      while( sender_value );

      /* restore the comma and goto the next value in receiver's table list */
      if( dummy1 )
	{
	  *dummy1++ = ',';
	}

      receiver_value = dummy1;
    }
  while( receiver_value );

out:

  if( sender_value )
    {
      TRACE(TRACE_ENTER_LEAVE, "Leave check_for_support, return value %s\n",
								sender_value);
    }
  else
    {
      TRACE(TRACE_ENTER_LEAVE, "Leave check_for_support, return value NULL\n");
    }

  return sender_value;
}


/* Deal with different draft versions */
void check_drafts( unsigned int draft_number, struct parameter_type *p )
{
  if( draft_number <= DRAFT8 )
    {/* this is draft 8 or before */
    if( p->special_key_flag & LOGOUTLOGINMINTIME_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" valid only in draft 9, "
			"current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }

    if( p->special_key_flag & MAXRECVPDULENGTH_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" valid only in draft 9 "
			"and after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }

  else
    {/* this is draft 9 and after */
      if( p->special_key_flag & DATAPDULENGTH_FLAG )
	{
	  TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" valid only in draft 8 "
	  		"and before, current draft is %u\n",
	     		p->parameter_name, draft_number/DRAFT_MULTIPLIER);
	}
    }

  if( draft_number >= DRAFT11 )
    {/* this is draft 11 or after, several keys are renamed */
    if( p->special_key_flag & ( LOGOUTLOGINMINTIME_FLAG |
				LOGOUTLOGINMAXTIME_FLAG |
				FMARKER_FLAG |
				RFMARKINT_FLAG |
				SFMARKINT_FLAG ) )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" not valid in draft 11 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }
  else
    {/* this is draft 9 or before */
    if( p->special_key_flag & ( DEFAULTTIME2WAIT_FLAG |
				DEFAULTTIME2RETAIN_FLAG |
				IFMARKER_FLAG |
				OFMARKER_FLAG |
				IFMARKINT_FLAG |
				OFMARKINT_FLAG ) )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" only valid in draft 11 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }

  if( draft_number < DRAFT12 )
    {/* this is before draft 12, TargetPortalGroupTag not valid yet */
    if( p->special_key_flag & TARGETPORTALGROUPTAG_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" only valid in draft 12 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }

  if( draft_number >= DRAFT18 )
    {/* this is draft 18 or after, some keys are removed or renamed */
    if( p->special_key_flag &  BIDIINITIALR2T_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" not valid in draft 18 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    if( p->special_key_flag &  MAXBURSTSIZE_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" not valid in draft 18 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    if( p->special_key_flag &  FIRSTBURSTSIZE_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" not valid in draft 18 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }
  else
    {/* prior to draft 18 */
    if( p->special_key_flag &  MAXBURSTLENGTH_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" not valid before draft 18"
			", current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    if( p->special_key_flag &  FIRSTBURSTLENGTH_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" not valid before draft 18"
			", current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }

  if( draft_number >= DRAFT13 )
    {/* this is draft 13 or after, one key is renamed */
    if( p->special_key_flag & MAXRECVPDULENGTH_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" not valid in draft 13 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }
  else
    {/* this is draft 12 or before */
    if( p->special_key_flag & MAXRECVDATASEGMENTLENGTH_FLAG )
      {
	TRACE(TRACE_ISCSI, "***Warning***: key \"%s\" only valid in draft 13 "
			"or after, current draft is %u\n",
			p->parameter_name, draft_number/DRAFT_MULTIPLIER);
      }
    }
}


/************************************************************************/
/* scan through the input and process keys                              */
/* p_param_tbl - pointer to the table of parameters                     */
/* sock - pointer to the socket structure                               */
/* process_these_types - types of parameters to be processed.           */
/* flags_to_be_set - Which flags to set                                 */
/* role - which is calling (INITIATOR/TARGET)                           */
/* Return:                                                              */
/*	returns the length of the string added to the output string	*/
/************************************************************************/

int scan_input_and_process( struct socket *sock,
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			int process_these_types,
			int flags_to_be_set, int role,
			int *max_send_length,
			unsigned int draft_number,
			unsigned int when_called,
			struct generic_pdu *inputpdu,
			struct generic_pdu *outputpdu,
			unsigned int flags,
			unsigned long long *login_flags,
			struct unknown_key **unknown_key_list )
{
  struct parameter_type *p = NULL, *p1 = NULL;
  int	resp_len, n;
  char	*dummy_string;
  char	*input_str, *this_key;
  char	*input_string;
  int	out_length = 0;
  char  *output_string = outputpdu->text + outputpdu->text_length;
  int	key_value_len;
  char	*value;
  int	int_value;
  char	*last_input;
  struct key_values	*key_table;
  int FBSize = -1, MBSize = -1;
  unsigned long long first_burst, max_burst;


  TRACE(TRACE_ENTER_LEAVE, "Enter scan_input_and_process\n");

  ALLOCATE_MAX_TEXT_LEN(dummy_string);
  ALLOCATE_MAX_TEXT_LEN(input_str);

  TRACE(TRACE_DEBUG, "scan_input_and_process draft_number %u\n", draft_number);

  if( draft_number <= DRAFT9 )
    key_table = &lower_case_values;
  else
    key_table = &upper_case_values;

  fix_burst_name( draft_number, &max_burst, &first_burst );

  /* Copy the input text to a local buffer */
  memcpy(input_str, inputpdu->text, inputpdu->text_length);
  input_string = input_str;			/* point at first char */
  last_input = input_str+inputpdu->text_length;	/* point beyond last char */
  *last_input = '\0';				/* ensure '\0' at end of last */

  while( input_string < last_input )
    {
      key_value_len = strlen(input_string);
      TRACE(TRACE_DEBUG, "key_value_len = %d\n", key_value_len);

      /* assume no reply will be sent */
      resp_len = 0;

      value = NULL;
      int_value = -1;
      if( (p = check_correctness(input_string, &value, p_param_tbl, role,
				key_table, when_called, &int_value,
				unknown_key_list)) == NULL )
	{/* We don't understand the request. */
	  if( value == NULL )
	    {/* Error is fatal */
	      out_length = -1;
	      goto out;
	    }
	  if (!((process_these_types & SECURITY_PARAM)) || (!is_securitykey(input_string)))
	    resp_len = sprintf(dummy_string, "%s=%s", input_string, value);
	}
      else if ( IS_KEY_BAD(p->neg_info) )
	{
	  out_length = -1;
          goto out;
	}
      else if( !IS_KEY_IRRELEVANT(p->neg_info) )
	{/* p is the pointer to the table entry for the received parameter
	    and value points to the string following the "=" in the input */

	  TRACE(TRACE_DEBUG, "Process input param: %s, value: %s\n",
						p->parameter_name, value);

	  /* check if this is a Full Feature Only key */
  	  if( IS_FULL_FEATURE_ONLY(p->type) )
	    {
	      /* protocol error */
	      TRACE_ERROR("%p only valid in full feature phase\n",
							p->parameter_name);
	      out_length = -1;
	      goto out;
	    }

	  /* Deal with different draft versions */
	  check_drafts(draft_number, p);

	  if( p->special_key_flag & flags_to_be_set )
	    {/* caller wants to know that this key was processed */
	      *login_flags |= p->special_key_flag;

	      if( p->special_key_flag & SESSIONTYPE_FLAG )
	        {/* set the DISCOVERY_FLAG when get session={Dd}iscovery */
		if( !strcmp(value, key_table->discovery) )
		  {
		    *login_flags |= DISCOVERY_FLAG;
		  }
		}
	    }

	  if( (p->type & process_these_types) )
	    {
	      /* handle key enquiry by replying with our current value */
	      if( !strcmp(value, "?") )
		{
		 if( draft_number >= DRAFT11 )
		  {/* inquiry no longer accepted in draft 11 and after */
	    	   TRACE_ERROR("inquiry no longer accepted "
			"param: %s, value: %s\n", p->parameter_name, value);
		   resp_len = sprintf(dummy_string,
				"%s=%s", p->parameter_name, key_table->reject);
		  }
		 else if( !IS_NUMBER(p->type)  &&  !IS_NUMBER_RANGE(p->type) )
		  {/* non-numeric parameter, send back copy of string value */
		   if( p->str_value != NULL )
			resp_len = sprintf(dummy_string,
			      "%s=%s", p->parameter_name, p->str_value);
		  }
          	 else
		  {/* numeric or range, convert and send back numeric value */
		   resp_len = sprintf(dummy_string,
			      "%s=%d", p->parameter_name, p->int_value);
		  }
		 /* do not consider this parameter as "gotten from other side"*/
		 p = NULL;
		}
	      else if( p->str_value  &&  !strcmp(p->str_value, "?") )
	      	{/* we sent an inquiry and this is the response */
		 /* Note this should never happen for draft 11 and after */
		 printk("iscsi response to inquiry: %s=%s\n",
						 p->parameter_name, value);
		}
	      else
	        {/* neither an inquiry or a reply to an inquiry */
		  /* first convert boolean to yes or no result */
		  if( IS_BOOLEAN(p->type) )
		    {
		     int receiver_value;
		     int sender_value;

		     if( !strcmp(value, key_table->yes) )
		       sender_value = 1;
		     else
		       sender_value = 0;

		     if( !strcmp(p->str_value, key_table->yes) )
		       receiver_value = 1;
		     else
		       receiver_value = 0;

		     if( IS_BOOL_AND(p->type) )
		       {/* Boolean AND function */
			 if(sender_value & receiver_value)
			   {/* both of the sides said YES so result is YES */
			     value = key_table->yes;
			   }
			 else
			   {/* either (or both) of the sides said NO */
			     value = key_table->no;
			     if( sender_value == 0 )
				{/* sender sent NO, so reply is optional */

				/* Draft 13, Section 4.2.2 Simple-value
				   negotiations
				   "Specifically, the two cases in which
				   responses are OPTIONAL are:
				   - The boolean function is "AND" and the
				     value "No" is received.  The outcome of
				     the negotiation is "No".
				   - The boolean function is "OR" and the value
				     "Yes" is received.  The outcome of the
				     negotiation is "Yes"."
				*/
			     	  p->neg_info |= KEY_REPLY_OPTIONAL;
				}
			   }
		       }
		     else
		       {/* Boolean OR function */
			 if(sender_value | receiver_value)
			   {/* either (or both) of the sides said YES */
			     value = key_table->yes;
			     if( sender_value != 0 )
			       {/* sender sent YES, so reply is optional */

					/* Draft 13, Section 4.2.2 Simple-value
					   negotiations
					   "Specifically, the two cases in
					   which responses are OPTIONAL are:
					   - The boolean function is "AND" and
					     the value "No" is received.  The
					     outcome of the negotiation is "No".
					   - The boolean function is "OR" and
					     the value "Yes" is received.  The
					     outcome of the negotiation is
					     "Yes"."
					*/
			     	  p->neg_info |= KEY_REPLY_OPTIONAL;
			       }
			   }
			 else
			   {/* both of the sides said NO so result is NO */
			     value = key_table->no;
			   }
		       }
		    }

		  if( p->special_key_flag &
			(MAXRECVPDULENGTH_FLAG|MAXRECVDATASEGMENTLENGTH_FLAG) )
		    {/* this is MaxRecvDataSegmentLength that other side wants
		    	to receive, i.e., the max size PDU we should send.
			No reply is necessary for this key. */

			if( int_value < 512  ||  int_value > 16777215 )
			  {
			    TRACE(TRACE_ISCSI,"***Warning***: %s "
				"%d out of bounds [512..16777215]\n",
				p->parameter_name, int_value);
			  }
		    	else
			  {/* value is within bounds, pass it back to caller */
			  /*  after forcing it to be a multiple of 512 */
			  int_value /= 512;
			  *max_send_length = 512*int_value;
			  }
		    }
		  else if( !IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info))
		    {/* parameter not sent previously by us,
			so this is a new offer, formulate our reply */
		      if( IS_INFORMATIONAL_PARAM(p->type) )
			{/* offered parameter is informational, send no reply */
			strreplace(&p->str_value, value);
			TRACE(TRACE_ISCSI,"Update key %s, new value %s\n",
					p->parameter_name, p->str_value);
			}
		      else if( IS_NUMBER_RANGE(p->type) )
			{/* offered parameter's value should be a range
						    <key=value1~value2> */
			  int send_irrelevant = 0;

			  if( p->special_key_flag & OFMARKINT_FLAG )
			    {/* reply with Irrelevant if OFMARKER is no */
			      if( (p1=find_flag_parameter(OFMARKER_FLAG,
						    p_param_tbl)) != NULL )
				{
				  if( !strcmp(p1->str_value,key_table->no) )
				  {
				    send_irrelevant = 1;
				  }
				}
			    }
			  else if( p->special_key_flag & IFMARKINT_FLAG )
			    {/* reply with Irrelevant if IFMARKER is no */
			      if( (p1=find_flag_parameter(IFMARKER_FLAG,
						    p_param_tbl)) != NULL )
				{
				  if( !strcmp(p1->str_value,key_table->no) )
				  {
				    send_irrelevant = 1;
				  }
				}
			    }

			  if( send_irrelevant )
			    {
			      resp_len = sprintf(dummy_string, "%s=%s",
				  p->parameter_name, key_table->irrelevant);
			    }
			  else if( int_value >= 0 )
			    {
			      if( p->int_value != int_value )
				{/* keep the offered lower end of range */
				  p->int_value = int_value;

				  TRACE(TRACE_DEBUG,
					"Update parameter %s with value %d\n",
					p->parameter_name, p->int_value);
				}
			    }
			}
		     else if( !IS_NUMBER(p->type) )
			{/* offered parameter is not informational, range or
			    number, check that we support the offered value */
			  char *param_value = NULL;

			  if( IS_BOOLEAN(p->type) )
			    {/* offered parameter is boolean, support assured */
			      param_value = value;
			    }
			  else
			    {
			      param_value = check_for_support(p, value,
								draft_number);
			    }

			  if( param_value != NULL )
			    {/* value received is supported by us, use it */
			      if( !IS_KEY_REPLY_OPTIONAL(p->neg_info)  ||
				  (flags & USE_FULL_REPLIES) )
				{/* reply not optional or full reply required */
				  resp_len = sprintf(dummy_string, "%s=%s",
						p->parameter_name, param_value);
				}
			      if( p->str_value  &&
					 !strcmp(param_value,p->str_value) )
				{/* already have this value in place */
				}
			      else
				{
				  strreplace(&p->str_value, value);

				  TRACE(TRACE_DEBUG,
					  "Update parameter %s with value %s\n",
					  p->parameter_name, p->str_value);
				}
			    }
			  else
			     {/* value not supported by us, send reject */
			       resp_len = sprintf(dummy_string, "%s=%s",
					 p->parameter_name, key_table->reject);
			     }
			}
		     else
		       {/* offered parameter is a number, apply its selection
			   function */

			  if( (IS_MIN_NUMBER(p->type)  &&
				  p->int_value <= int_value)  ||
				  (IS_MAX_NUMBER(p->type)  &&
				  p->int_value >= int_value) )
			    {/* correct min or max already in place */
			    }
			  else
			    {/* the offered value is the one we keep */
			      p->int_value = int_value;

			      TRACE(TRACE_DEBUG,
				    "Update parameter %s with value %d\n",
				    p->parameter_name, p->int_value);
			    }

			  /* Delay replying to FirstBurstLength and MaxBurstLength
			   * until end of the PDU to ensure
			   * FirstBurstLength <= MaxBurstLength
			   */
			  if( p->special_key_flag & first_burst )
			    {
			      FBSize = p->int_value;
			      p->neg_info |= KEY_GOT_FROM_OTHER_SIDE;
			      /* we will reply this key */
			      p->neg_info |= KEY_SENT_TO_OTHER_SIDE;
			    }
			  else if( p->special_key_flag & max_burst )
			    {
			      MBSize = p->int_value;
			      p->neg_info |= KEY_GOT_FROM_OTHER_SIDE;
			      /* we'll reply this key */
			      p->neg_info |= KEY_SENT_TO_OTHER_SIDE;
			    }
			  else
			    {/* always need a reply to a numeric parameter */
			      resp_len = sprintf(dummy_string, "%s=%d",
						p->parameter_name,p->int_value);
			    }
		       }
		    }
		  else
		    {/* Parameter already sent, so this is the response */
		      if( !IS_NUMBER(p->type) )
			{/* received parameter is not a number */
			 char *param_value = NULL;

			 /* response value should just be a single string */
			 if( strchr(value, ',') )
			   {
			     TRACE_ERROR("multiple values in negotiation "
					"response for parameter: %s\n",
					p->parameter_name);
			     out_length = -1;
			     goto out;
			   }

		         if( IS_BOOLEAN(p->type) )
		           {/* offered parameter is boolean, support assured */
			     param_value = value;
		           }
			 else if( !IS_NUMBER_RANGE(p->type) )
			   {
			     param_value = check_for_support(p, value,
							  draft_number);
			   }
			 if( param_value != NULL )
			   {/* received value is supported by receiver, use it*/
			     if( p->str_value && !strcmp(value, p->str_value) )
			       {/* already have this value in place */
			       }
			     else
			       {
				 strreplace(&p->str_value, value);

				 TRACE(TRACE_ISCSI,
					  "Update key %s, new value %s\n",
					  p->parameter_name, p->str_value);
			       }
			    }
			 else if( IS_NUMBER_RANGE(p->type) )
			    {/* reply should be in range <key=value1~value2> */
			      if( IS_KEY_REJECT(p->neg_info) )
			        {/* got reply of Reject to OFMarkInt/IFMarkInt*/
			   	}
			      else
			      	{/* reply must be numeric within offered range*/
				if( check_range(p->value_list, int_value) < 0 )
				  {
				  out_length = -1;
				  goto out;
				  }
			  	if( p->int_value != int_value )
				  {
				    /* assign the negotiated value */
				    p->int_value = int_value;

				    TRACE(TRACE_DEBUG,
					"Update parameter %s with value %d\n",
					p->parameter_name, p->int_value);
				  }
				}
			    }
			 else
			    {/* received value not among those we offered */
				TRACE_ERROR("bad response for key "
					"%s, response %s\n", p->parameter_name,
					value);
				out_length = -1;
				goto out;
			    }
			}
		      else
			{/* received reply parameter is a number */

			  /* Numeric param, select min or max as appropriate */
			  if( (p->int_value<int_value && IS_MIN_NUMBER(p->type))
			   || (p->int_value>int_value && IS_MAX_NUMBER(p->type))
			    )
			    {
			      /* Error, reply did not do min/max correctly */
			      TRACE_ERROR("got reply %d to offer of %d for %s "
			      		"parameter %s\n",
		      			int_value, p->int_value,
					IS_MIN_NUMBER(p->type) ? "min" : "max",
					p->parameter_name);
			      out_length = -1;
			      goto out;
			    }
			  else if( p->int_value != int_value )
			    {
			      /* assign the negotiated value */
			      p->int_value = int_value;

			      TRACE(TRACE_DEBUG,
					  "Update parameter %s with value %d\n",
					  p->parameter_name, p->int_value);
			    }
			}
		    }
		}
	    }
	  else
	    {
	    if( IS_SECURITY_PARAM(p->type) )
		{
	    	TRACE_ERROR("not in security phase when received input "
			"param: %s, value: %s\n", p->parameter_name, value);
		out_length = -1;
		goto out;
	    	}
	    else if( IS_OPERATIONAL_PARAM(p->type) )
		{
	    	TRACE_ERROR("not in operational phase when received input "
			"param: %s, value: %s\n", p->parameter_name, value);
		out_length = -1;
		goto out;
	    	}
	    else if( IS_INFORMATIONAL_PARAM(p->type) )
		{
	    	TRACE_ERROR("not processing informational parameters when "
					"received input param: %s, value: %s\n",
					p->parameter_name, value);
		out_length = -1;
		goto out;
	    	}
	    else
	    	{
		TRACE_ERROR("received unclassified input param: %s, "
				"value: %s\n", p->parameter_name, value);
		out_length = -1;
		goto out;
		}
	    }
	}

      /* Always check if the response length exceeds MAX_TEXT_LEN */
      /* Then the remaining parameters have to be sent in a   */
      /* separate text command/response */

      TRACE(TRACE_DEBUG, "resp_len = %d\n", resp_len);

      if( resp_len )
	{
	  if( check_out_length(out_length, resp_len) )
	    {
	      *login_flags |= MORE_TO_SEND_FLAG;
	      break;
	    }
	  else
	    {
	      /* Update resp_len and the pointer to output_string */
	      sprintf( output_string, "%s", dummy_string );
	      output_string += resp_len + 1;

	      TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);

	      /* update out_length */
	      out_length += resp_len + 1;

	      if( p )
		{/* set sent-to-other-side bit */
		  p->neg_info |= KEY_SENT_TO_OTHER_SIDE;
		}

	      resp_len = 0;
	    }
	}

      if( p )
        {/* set got-from-other-side bit */
	  p->neg_info |= KEY_GOT_FROM_OTHER_SIDE;
	}

      /* Goto the next key=value pair */
      this_key = input_string;		/* remember where this key started */
      input_string += key_value_len + 1;/* this should be start of next key */

      /* Drafts 9 and before, Section 3.10.4 Text
         "Every key=value pair (including the last or only pair) MUST
	 be followed by at least one null (0x00) delimiter."
      */

      /* This implied there could be more than one null (0x00) at the
	 end of a string, so skip over these extra nulls now */
      while( *input_string == '\0'  &&  input_string < last_input )
      	{
	input_string++;
	}
      /* Drafts 11 and after, Section 4.1 Text Format
      	 "Every key=value pair (including the last or only pair) MUST
	 be followed by one null (0x00) delimiter."
      */
      /* This implies that there can NOT be more than one null (0x00) at the
	 end of a string, so if there were extra nulls, give a warning now. */
      n = (input_string-this_key) - (key_value_len + 1);
      if( n > 0  &&  draft_number > DRAFT9 )
      	{
	TRACE(TRACE_ISCSI, "***Warning***: %d extra nulls (0x00) found after "
						"key \"%s\"\n", n, this_key);
	}
    }

   /* process the FirstBurstSize and MaxBurstSize if their reply is delayed */
   if( FBSize > 0 && MBSize > 0) /* both keys were offered by the other side before */
   {
	if( FBSize < MBSize ) /* ok */
	{
	}
	else /* fix the FirstBurstSize */
	{
	  FBSize = MBSize;
	}
   }
   else if( FBSize > 0) /* only FirstBurstLength is received */
   {
	if ( (p=find_flag_parameter(max_burst,p_param_tbl)) != NULL )
	{
		if (  FBSize > p->int_value && (p->neg_info&KEY_SENT_TO_OTHER_SIDE) )
		{ /* The unsent FirstBurstSize is bigger than the negotiated MaxBurstSize,
		     fix the FirstBurstSize to the negotiated MaxBurstSize
		   */
		  FBSize = p->int_value;
		}
		if( FBSize < p->int_value && !(p->neg_info&KEY_SENT_TO_OTHER_SIDE) )
		{ /* The unsent FirstBurstSize is smalled than the default(unnegotiated)
		     MaxBurstSize, offer the default MaxBurstSize to the other side
		   */
		  MBSize = p->int_value;
		}
		else if (  FBSize > p->int_value && !(p->neg_info&KEY_SENT_TO_OTHER_SIDE) )
		{
		  /* The unsent FirstBurstSize is bigger than the default(unnegotiated)
		   * MaxBurstSize, fix the FirstBurstSize and send both keys
		   */
		  FBSize = p->int_value;
		  MBSize = p->int_value;
		}
	        p->neg_info |= KEY_SENT_TO_OTHER_SIDE;
		outputpdu->flags &= (~T_BIT); /* We are expecting an answer,
						 set T bit to be 0*/
	}
   }
   else if( MBSize > 0 ) /* only MaxBurstLength is received */
   {
	if ( (p=find_flag_parameter(first_burst,p_param_tbl)) != NULL )
	{
		if( MBSize < p->int_value && (p->neg_info&KEY_SENT_TO_OTHER_SIDE ) )
		{ /* The unsent MaxFirstSize is smaller than the negotiated FirstBurstSize
		     can't be fixed
		   */
		  TRACE_ERROR("MaxBurstSize %d is smaller than the negotiated FirstBurstSize %d",
				  MBSize, p->int_value);
		  out_length=-1;
		  goto out;
		}
		if( MBSize > p->int_value && !(p->neg_info&KEY_SENT_TO_OTHER_SIDE ) )
		{ /* The unsent MaxBurstSize is bigger than default(unnegotiated) FirstBurstSize
		     offer the FirstBurstSize with the default value
		   */
		  FBSize = p->int_value;
		}
		else if( MBSize < p->int_value && !(p->neg_info&KEY_SENT_TO_OTHER_SIDE ) )
		{ /* The unsent MaxBurstSize is smaller than the default(unnegotiated)
		     FirstBurstSize, fix the FirstBurstSize and offer it to negotiate  */
		  FBSize = MBSize;
		}
 		p->neg_info |= KEY_SENT_TO_OTHER_SIDE;
		outputpdu->flags &= (~T_BIT);
 	}
   }

  if( FBSize > 0 )
  {
   /* reply the FirstBurstSize */
   if( draft_number < DRAFT18)
   	resp_len = sprintf(dummy_string, "%s=%d", FIRSTBURSTSIZE, FBSize );
   else
   	resp_len = sprintf(dummy_string, "%s=%d", FIRSTBURSTLENGTH, FBSize );
   sprintf( output_string, "%s", dummy_string );
   output_string += resp_len + 1 ;

   TRACE(TRACE_ISCSI, "Attach key, %s\n", dummy_string);

   /* update out_length */
   out_length += resp_len +1;
   resp_len = 0;
   FBSize = -1;
  }
  if ( MBSize > 0 )
  {
   /* reply the MaxBurstSize */
  if( draft_number < DRAFT18)
  	resp_len = sprintf(dummy_string, "%s=%d", MAXBURSTSIZE, MBSize );
  else
  	resp_len = sprintf(dummy_string, "%s=%d", MAXBURSTLENGTH, MBSize );
   sprintf( output_string, "%s", dummy_string );
   output_string += resp_len + 1 ;

   TRACE(TRACE_ISCSI, "Attach key, %s\n", dummy_string);

   /* update out_length */
   out_length += resp_len +1;
   MBSize = -1;
  }

out:
  FREE_STRING( dummy_string );
  FREE_STRING( input_str );

  TRACE(TRACE_ENTER_LEAVE, "Leave scan_input_and_process, out_length %d\n",
								out_length);

  /* return the length added to the output_string */
  return out_length;
}


/************************************************************************/
/* scan through the table and count the keys of each type */
/* that still remain to be negotiated */
/* p_param_tbl - pointer to the table of parameters */
/* returns counts in last 3 parameters */
/************************************************************************/

void scan_table_and_count(
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			unsigned int draft_number,
			int *nsecurity,
			int *ninformational,
			int *noperational )
{
  struct parameter_type *p;
  int i;


  TRACE(TRACE_ENTER_LEAVE, "Enter scan_table_and_count draft_number %u\n",
 								 draft_number);

  /* initialize all counts to 0 */
  *nsecurity = *ninformational = *noperational = 0;

  for( i=0, p = p_param_tbl;  i < MAX_CONFIG_PARAMS;  i++, p++ )
    {
      if( (IS_KEY_TO_BE_NEGOTIATED(p->neg_info))	&&
	  (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info))	)
	{/* only count those parameters that still remain to be negotiated */

	if( IS_SECURITY_PARAM(p->type) )
	    *nsecurity += 1;
	else if( IS_INFORMATIONAL_PARAM(p->type) )
	    *ninformational += 1;
	else if( IS_OPERATIONAL_PARAM(p->type) )
	    *noperational += 1;
	}
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave scan_table_and_count draft_number %u\n",
 								 draft_number);
}


/************************************************************************/
/* scan through the table and process keys */
/* p_param_tbl - pointer to the table of parameters */
/* process_these_types - types of parameters to be processed. */
/* Return: */
/*		returns the length of the string added to the output string */
/************************************************************************/

int scan_table_and_process(struct socket *sock,
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			int process_these_types,
			int role,
			unsigned int draft_number,
			struct generic_pdu *inputpdu,
			struct generic_pdu *outputpdu,
			unsigned int flags,
			unsigned long long *login_flags )
{
  struct parameter_type *p;
  int resp_len = 0;
  char *dummy_string;
  int out_length = 0;
  char *output_string = outputpdu->text + outputpdu->text_length;
  int i;
  struct key_values	*key_table;


  TRACE(TRACE_ENTER_LEAVE, "Enter scan_table_and_process\n");

  ALLOCATE_MAX_TEXT_LEN(dummy_string);

  TRACE(TRACE_DEBUG, "scan_table_and_process draft_number %u\n", draft_number);

  if( draft_number <= DRAFT9 )
    key_table = &lower_case_values;
  else
    key_table = &upper_case_values;

  for( i=0, p = p_param_tbl;  i < MAX_CONFIG_PARAMS;  i++, p++ )
    {
      if( (IS_KEY_TO_BE_NEGOTIATED(p->neg_info))	&&
	  (p->type & process_these_types)  	&&
	  (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info))	)
	{

	  TRACE(TRACE_DEBUG,"Process table parameter: %s\n",p->parameter_name);

	  /* Deal with different draft versions */
	  check_drafts(draft_number, p);

	  /* Add it to the outgoing data */
	  if( IS_NUMBER(p->type) )
	    {
	      resp_len = sprintf(dummy_string,
				  "%s=%d", p->parameter_name, p->int_value);
	    }
	  else if( p->value_list != NULL )
	    {
	      resp_len = sprintf(dummy_string,
				  "%s=%s", p->parameter_name, p->value_list);
	    }
	  else if( p->str_value != NULL )
	    {
	      resp_len = sprintf(dummy_string,
				  "%s=%s", p->parameter_name, p->str_value);
	    }


	  /* Always check if the Text length exceeds MAX_TEXT_LEN */
	  /* Then the remaining parameters have to be sent in a   */
	  /* separate login command/response */

	  if( check_out_length(outputpdu->text_length + out_length, resp_len) )
	    {
	      /* set the more_to_send flag */
	      *login_flags |= MORE_TO_SEND_FLAG;
	      break;
	    }
	  else
	    {
	      /* Update resp_len and the pointer to output_string */

	      strcpy(output_string, dummy_string);
	      output_string += resp_len + 1;

	      TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);

	      /* update out_length */
	      out_length += resp_len + 1;

	      /* set sent bit */
	      p->neg_info |= KEY_SENT_TO_OTHER_SIDE;

	      /* in 2 special cases, set the reply optional bit */
	      if( (IS_BOOL_AND(p->type) &&
		      strcmp(p->str_value, key_table->no) == 0)  ||
		  (IS_BOOL_OR(p->type)  &&
		      strcmp(p->str_value, key_table->yes) == 0) )
	        {
		  p->neg_info |= KEY_REPLY_OPTIONAL;
		  TRACE(TRACE_DEBUG, "Reply optional for %s\n", dummy_string);
		}
	      else
	      	{
		  TRACE(TRACE_DEBUG,"Reply required for %s\n",dummy_string);
		}
	      resp_len = 0;
	    }
	}
    }

  FREE_STRING(dummy_string);

  TRACE(TRACE_ENTER_LEAVE, "Leave scan_table_and_process, out_length %d\n",
  								out_length);

  /* return the length added to the output_string */
  return out_length;
}


/* uses the final values negotiated during a successful login of a new session
   to set up the session-wide operational values used during FFP */
void set_session_parameters(
			struct session_operational_parameters *oper_param_entry,
			struct parameter_type login_params[MAX_CONFIG_PARAMS],
			unsigned int draft_number )
{
  struct parameter_type   *p;
  unsigned int		  i;
  struct key_values	  *key_table;
  char *tptr;

  TRACE(TRACE_ENTER_LEAVE, "Enter set_session_parameters\n");

  if( draft_number <= DRAFT9 )
    key_table = &lower_case_values;
  else
    key_table = &upper_case_values;

  for( p = login_params, i = 0;  i < MAX_CONFIG_PARAMS;  p++, i++ )
    {
      TRACE(TRACE_DEBUG, "Parameter %s\n", p->parameter_name);

      /* For MaxConnections */
      if( p->special_key_flag & MAXCONNECTIONS_FLAG )
	     oper_param_entry->MaxConnections = p->int_value;

      /* For InitialR2T */
      else if( p->special_key_flag & INITIALR2T_FLAG )
	     oper_param_entry->InitialR2T =
			  !strcmp(p->str_value, key_table->yes);

      /* For BidiInitialR2T */
      else if( p->special_key_flag & BIDIINITIALR2T_FLAG )
	     oper_param_entry->BidiInitialR2T =
			  !strcmp(p->str_value, key_table->yes);

      /* For ImmediateData */
      else if( p->special_key_flag & IMMEDIATEDATA_FLAG )
	     oper_param_entry->ImmediateData =
			  !strcmp(p->str_value, key_table->yes);

      /* For DataPDULength times 512 so units are same as MaxRecvPDULength */
      else if( p->special_key_flag & DATAPDULENGTH_FLAG )
	     oper_param_entry->DataPDULength=p->int_value*512;

      /* For MaxBurstSize, forced to a multiple of 512 */
      else if( p->special_key_flag & MAXBURSTSIZE_FLAG )
	  {
	   if( draft_number >= DRAFT9 )
	    {/* draft 9 and after */
	      oper_param_entry->MaxBurstSize = (p->int_value/512)*512;
	    }
	   else
	    {/* draft 8 and before, but in draft 9 units */
	      oper_param_entry->MaxBurstSize = (p->int_value)*512;
	    }
	 }

    /* For MaxBurstLength, draft 18 and after */
    else if( p->special_key_flag & MAXBURSTLENGTH_FLAG )
	 {
	      oper_param_entry->MaxBurstLength = (p->int_value/512)*512;
	 }

     /* For FirstBurstSize, forced to a multiple of 512 */
     else if( p->special_key_flag & FIRSTBURSTSIZE_FLAG )
     {
	  if( draft_number >= DRAFT9 )
	    {/* draft 9 and after */
	      oper_param_entry->FirstBurstSize = (p->int_value/512)*512;
	    }
	  else
	    {/* draft 8 and before, but in draft 9 units */
	      oper_param_entry->FirstBurstSize = (p->int_value)*512;
	    }
	 }
	
     /* For FirstBurstLength, draft 18 and after */
     else if( p->special_key_flag & FIRSTBURSTLENGTH_FLAG )
     {
	      oper_param_entry->FirstBurstLength = (p->int_value/512)*512;
	 }

    /* For LogoutLoginMinTime, used only in draft 9 */
    else if( p->special_key_flag & LOGOUTLOGINMINTIME_FLAG )
	   oper_param_entry->LogoutLoginMinTime=p->int_value;

      /* For DefaultTime2Wait, used only in draft 11 and after */
      else if( p->special_key_flag & DEFAULTTIME2WAIT_FLAG )
	oper_param_entry->DefaultTime2Wait=p->int_value;

      /* For LogoutLoginMaxTime */
      else if( p->special_key_flag & LOGOUTLOGINMAXTIME_FLAG )
	oper_param_entry->LogoutLoginMaxTime=p->int_value;

      /* For DefaultTime2Retain, used only in draft 11 and after */
      else if( p->special_key_flag & DEFAULTTIME2RETAIN_FLAG )
	oper_param_entry->DefaultTime2Retain=p->int_value;

      /* For MaxOutstandingR2T */
      else if( p->special_key_flag & MAXOUTSTANDINGR2T_FLAG )
	oper_param_entry->MaxOutstandingR2T=p->int_value;

      /* For DataPDUInOrder */
      else if( p->special_key_flag & DATAPDUINORDER_FLAG )
	oper_param_entry->DataPDUInOrder =
			  !strcmp(p->str_value, key_table->yes);

      /* For DataSequenceInOrder */
      else if( p->special_key_flag & DATASEQUENCEINORDER_FLAG )
	oper_param_entry->DataSequenceInOrder =
			  !strcmp(p->str_value, key_table->yes);

      /* For ErrorRecoveryLevel */
      else if( p->special_key_flag & ERRORRECOVERYLEVEL_FLAG )
	oper_param_entry->ErrorRecoveryLevel=p->int_value;

      /* For SessionType */
      else if( p->special_key_flag & SESSIONTYPE_FLAG )
	  {
	  if( (oper_param_entry->SessionType =
		  !strcmp(p->str_value, key_table->discovery)) )
	    {/* discovery session */
	    /* Draft 13, Section 11.22 SessionType
	       "The discovery session implies MaxConnections = 1 and overrides
	       both the default and an explicit setting."
	    */
	      oper_param_entry->MaxConnections = 1;
	    }
	 }
     else if( p->special_key_flag & TARGETNAME_FLAG )
     {
	  if( p->neg_info & KEY_WRONG) 
	  {
		oper_param_entry->TargetName=0;
	  }
	  else if( p->str_value)
	  {
		tptr=strchr(p->str_value, ':');
		if( tptr) /* save the number of TargetName for the target*/
			oper_param_entry->TargetName=*(tptr+1)-'0';
		else
			oper_param_entry->TargetName=0;
	  }
    }
  }

  TRACE(TRACE_ENTER_LEAVE, "Leave set_session_parameters\n");
}


/* uses the final values negotiated during a successful login of a new
   connection to set up the connection-specific operational values used
   during FFP */
void set_connection_parameters(
		struct connection_operational_parameters *oper_param_entry,
		struct parameter_type login_params[MAX_CONFIG_PARAMS],
		unsigned int draft_number )
{

  TRACE(TRACE_ENTER_LEAVE, "Enter set_connection_parameters\n");

  /* for now, there are no connection-specific values used during
     FFP.  These would be the security values and the marker values.
  */

  TRACE(TRACE_ENTER_LEAVE, "Leave set_connection_parameters\n");
}


/************************************************************************/
/* Dynamically fixes up the attributes of the Burst parameters		*/
/* 	FirstBurstSize and MaxBurstSize					*/
/* Called only from adjust_version_diff()				*/
/* when the draft number is being forced.				*/
/* Note this means that when configuring parameters, the draft number	*/
/* should be forced first, before any actual parameter is configured.	*/
/************************************************************************/

void fix_burst_parameter( struct parameter_type *p, unsigned int int_value,
						unsigned int draft_number )
	{
    TRACE(TRACE_ENTER_LEAVE, "Enter fix_burst_parameter\n");
	
	p->type &= ~(ZERO_TO_32767 | N512_TO_16777215);

    if( draft_number <= DRAFT8 )
	{/* draft 8 or before */
	p->type |= ZERO_TO_32767;
	p->int_value = int_value/512;		/* in units of 512 bytes */
	}
    else
	{/* draft 9 or after */
	p->type |= N512_TO_16777215;
	p->int_value = int_value;
	}
 
   TRACE(TRACE_ENTER_LEAVE, "Leave fix_burst_parameter\n");
   
	}


/* called when starting up a new session -- after all parameters have been
   configured and copied into the new session's p_param_tbl, and before
   any of these parameters are used in the login negotiation.
  */
void adjust_version_diff( struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
			  unsigned int draft_number )
{
  struct parameter_type   *p;
  unsigned int		  i;
  unsigned long long max_burst, first_burst;


  fix_burst_name( draft_number, &max_burst, &first_burst);
 
  for( p = p_param_tbl, i = 0;  i < MAX_CONFIG_PARAMS;  p++, i++ )
    {
      /* adjust Burst parameter units to match draft */
      if( p->special_key_flag & max_burst )
	{
	   fix_burst_parameter(p, INITIAL_MAXBURSTLENGTH, draft_number);
	}

      else if( p->special_key_flag & first_burst)
	{
	   fix_burst_parameter(p, INITIAL_FIRSTBURSTLENGTH, draft_number);
	}

      else
	{/* Adjust the leading letter of some default keys for drafts 8 & 9.
	    Starting with draft 11, all values start with upper-case letters.
	    The default values in the table are set for draft 12 and after.
	   */
	if ( draft_number <= DRAFT9 )
	  {
	    if( p->special_key_flag &
		  (HEADERDIGEST_FLAG | DATADIGEST_FLAG | AUTHMETHOD_FLAG))
	      {/* 3 keys default to NONE */
		if( !strcmp(p->str_value, NONE) )
		  strreplace(&p->str_value, NONE_LOW);
		if( !strcmp(p->value_list, NONE) )
		  strreplace(&p->value_list, NONE_LOW);
	      }
	    else if( p->special_key_flag & (INITIALR2T_FLAG |
				BIDIINITIALR2T_FLAG | IMMEDIATEDATA_FLAG |
				DATAPDUINORDER_FLAG | DATASEQUENCEINORDER_FLAG))
	      {/* 5 keys default to YES */
		if( !strcmp(p->str_value, YES) )
		  strreplace(&p->str_value, YES_LOW);
		if( !strcmp(p->value_list, YES) )
		  strreplace(&p->value_list, YES_LOW);
	      }
	    else if( p->special_key_flag &
			  (FMARKER_FLAG | OFMARKER_FLAG | IFMARKER_FLAG) )
	      {/* 3 keys default to NO */
		if( !strcmp(p->str_value, NO) )
		  strreplace(&p->str_value, NO_LOW);
		if( !strcmp(p->value_list, NO) )
		  strreplace(&p->value_list, NO_LOW);
	      }
	    else if( p->special_key_flag & SESSIONTYPE_FLAG )
	      {/* 1 key defaults to NORMAL */
		if( !strcmp(p->str_value, NORMAL) )
		  strreplace(&p->str_value, NORMAL_LOW);
		if( !strcmp(p->value_list, NORMAL) )
		  strreplace(&p->value_list, NORMAL_LOW);
	      }

	    /* InitialR2T, BidiInitialR2T, ImmediateData keys are
	       "leading only" for draft 11 or after,
	       they were "all" for draft 9 or before.
	    */
	    if( p->special_key_flag &
		  (INITIALR2T_FLAG | BIDIINITIALR2T_FLAG | IMMEDIATEDATA_FLAG) )
	      {
		p->type &= ~LEADING_ONLY;
		p->type |= ALL;
	      }

	  if( draft_number <= DRAFT8 )
	    {/* this is draft 8 or before */
	    if( p->special_key_flag & (HEADERDIGEST_FLAG | DATADIGEST_FLAG) )
		{/* the 2 digest keys are security parameters in draft 8 */
		  p->type &= ~OPERATIONAL_PARAM;
		  p->type |= SECURITY_PARAM;
		}
	    }
	  }

	/* The default values for DEFAULTTIME2WAIT and DEFAULTTIME2RETAIN
	   were 3 when they first came in in draft 11 */
	if ( draft_number <= DRAFT11 )
	  {
	    if( p->special_key_flag &
			    (DEFAULTTIME2WAIT_FLAG | DEFAULTTIME2RETAIN_FLAG) )
	      {
		p->int_value = 3;
	      }
	  }
      }
    }
}



#endif

