/*****************************************************************************
*   ref13_06/target/iscsi_target.h
*
*   header file for the iSCSI Target
*
*   IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
*   By downloading, copying, installing or using the software you agree to this
*   license. If you do not agree to this license, do not download, install,
*   copy or use the software.
*
*   License Agreement
*
*   Copyright (C) 2001  Inter-Operability Lab.(IOL),Univ. of New Hampshire(UNH),
*                       Durham, NH 03824
*
*   Redistribution and use in source and binary forms, with or without
*   modification, are permitted provided that the following conditions are met:
*
*   -Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
*
*   -Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
*
*   The name of IOL/UNH may not be used to endorse or promote products derived
*   from this software without specific prior written permission.
*
*   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
*   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IOL,UNH OR CONTRIBUTORS
*   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF
*   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*   POSSIBILITY OF SUCH DAMAGE.
******************************************************************************/

#ifndef __ISCSI_TARGET_H
#define __ISCSI_TARGET_H

#include "scsi_target.h"
#include "../iscsi-unh-common/debug.h"
#include "../iscsi-unh-common/text_param.h"
#include "../iscsi-unh-common/crc.c"
#include "../iscsi-unh-common/target_negotiate.h"

/* stores everything related to a SCSI command received */
struct iscsi_cmnd
{
  /* next: pointer to the next command in the list */
  struct iscsi_cmnd	*next;

  /* conn: connection on which this command was received */
  struct iscsi_conn	*conn;

  /* session: session on which this command was received */
  struct iscsi_session	*session;

  /* unsolicited_data_sem: to control receiving immediate data */
  struct semaphore    	unsolicited_data_sem;

  /* state: execution state of the command */
  int			state;

  /* ping_data: data get from nopout and send for nopin */
  char *		ping_data;

  /* cmnd: cmnd corresponding to this struct */
  Target_Scsi_Cmnd	*cmnd;

  /* message: message corresponding to this struct */
  Target_Scsi_Message 	*message;

  /* immediate_data_present: */
  __u8    		immediate_data_present;

  /* unsolicited_data_present: */
  __u8    		unsolicited_data_present;

  /* data_length: data associated with the command */
  __u32			data_length;

  /* first_burst_size: data associated with the command */
  __u32			first_burst_len;

  /* max_burst_size: data associated with one R2T */
  __u32			next_burst_len;

  /* data_done: dat left to be transferred */
  __u32			data_done;

  /* data_dir: direction of data transfer */
  int			data_dir;

  /* scatter_list_offset: stores offset within sg entry */ 
  __u32           	scatter_list_offset;

  /* scatter_entry_count:stores current count of sg entry */    
  __u32           	scatter_list_count;

  /* init_task_tag: Initiator Tag */
  __u32			init_task_tag;

  /* target_xfer_tag: Target Assigned Tag */
  __u32			target_xfer_tag;

  /* ref_task_tag: Referenced Task Tag used when aborting tasks*/
  __u32           	ref_task_tag;

  /* ref_task_lun: Referenced Task lun used when aborting luns*/
  __u32           	ref_task_lun;

  /* ref_cmd_sn: Referenced command sn used when aborting command*/
  __u32           	ref_cmd_sn;

  /* lun and cdb: as their name mean */
  __u64   lun;
  __u8    cdb[16];	

  /* cmd_sn of this command and stat_sn of this command response */
  __u32			cmd_sn;
  __u32			stat_sn;

  /* Command WIDE COUNTERS: data_sn and r2t_sn */
  __u32			data_sn;
  __u32			r2t_sn;
  __u32			outstanding_r2t;

  /* Added r2t cookie details - SAI */
  struct iscsi_cookie *first_r2t_cookie;
  struct iscsi_cookie *last_r2t_cookie;

  /* queue out-of-order data pdus - SAI */
  struct iscsi_cookie *first_data_q;
  struct iscsi_cookie *last_data_q;

  /* Added timestamp for r2t retransmissions - SAI */
  int           timestamp;

  /* Store the previously sent data_sn for recovery - SAI */
  int           prev_data_sn;

  /* Data PDU re-transmission on Data SNACK - SAI */
  __u8                    retransmit_flg;

};


/* global struct to store properties corresponding to an iscsi server */
struct iscsi_global
{
  /* conn_id: id assigned to connections within session */
  int			conn_id;

  /* ntsih: id assigned to target sessions */
  __u16			ntsih;

  /* nsessions: no of sessions active */
  int			nsessions;

  /* session_list: list of sessions */
  struct iscsi_session	*session_list;

  /* session_sem: semaphore to add and remove sessions */
  struct semaphore	session_sem;

  /* server_thr: task struct for the server thread */
  struct task_struct	*server_thr;

  /* server_socket: socket on which the server listens */
  struct socket		*server_socket;

  /* server_sem: semaphore for killing the server thread */
  struct semaphore	server_sem;

  /* device: device returned by the STML */
  Scsi_Target_Device	*device;

  /* force: manageable bits to initialize connection flags */
  unsigned int            force;

  /* nop_period: timeout period to send periodic NopIns to initiator */
  unsigned int            nop_period;
        
  /* Data and Status snack support - SAI */
  __u8                  targ_snack_flg;

  /* Added r2t retransmit period - SAI */
  int                   r2t_period;

  /* draft_number: draft number target will enforce */   
  unsigned int            draft_number;

  /* text parameters accepted by the target */
  struct parameter_type   (*param_tbl)[MAX_CONFIG_PARAMS];

  /* chap parameter - CHONG */
  struct chap_parameter_type chap_parameter;
};


struct	iscsi_global *devdata;

int	iscsi_detect		(Scsi_Target_Template*);
int	iscsi_release		(Scsi_Target_Device*);
int	iscsi_xmit_response	(Target_Scsi_Cmnd*);
int	iscsi_rdy_to_xfer	(Target_Scsi_Cmnd*);
int 	iscsi_proc_info		(char *, char **, off_t, int, int, int);
void 	iscsi_task_mgt_fn_done 	(Target_Scsi_Message *msg);

# define ISCSI {\
  name: "iscsi_target", \
  detect:iscsi_detect, \
  release:iscsi_release,\
  xmit_response:iscsi_xmit_response, \
  rdy_to_xfer:iscsi_rdy_to_xfer,\
  proc_info:iscsi_proc_info, \
  task_mgmt_fn_done:iscsi_task_mgt_fn_done \
};

Scsi_Target_Template 	my_template = ISCSI;
#include "scsi_target_module.c"

int 	init_target( struct iscsi_global*);
void	iscsi_server_thread (void*);
void	iscsi_rx_thread	(void*);
void	iscsi_tx_thread (void*);
void	clean_bad_session(struct iscsi_session *session);

static inline int iscsi_release_session (struct iscsi_session *session);
static inline int iscsi_release_connection (struct iscsi_conn *conn);
static int rx_data (struct iscsi_conn *conn, struct msghdr *msg, int data);
static int tx_data (struct iscsi_conn *conn, struct msghdr *msg, int data);

static inline int handle_login (struct iscsi_conn *conn, unsigned char *buffer);
static inline int handle_discovery (struct iscsi_conn *conn, unsigned char *buffer);
static inline int handle_logout (struct iscsi_conn *conn, unsigned char *buffer);
static inline int handle_nopout (struct iscsi_conn *conn, unsigned char *buffer);
static inline int handle_cmnd (struct iscsi_conn *conn, unsigned char *buffer);
static inline int handle_task_mgt_command(struct iscsi_conn *conn, unsigned char *buffer);
static inline int handle_data (struct iscsi_conn *conn, unsigned char *buffer);
static inline int handle_immediate_data(struct iscsi_cmnd *cmd, int length);

static int fill_msg_hdr (struct msghdr *msg);
static inline struct iscsi_cmnd* search_iscsi_cmnd (Target_Scsi_Cmnd *cmnd);
static inline struct iscsi_cmnd* search_task_mgt_command(Target_Scsi_Message *message);
static inline struct iscsi_cmnd* search_tags (struct iscsi_conn *conn, __u32 init_task_tag, __u32 target_xfer_tag);

static inline int handle_discovery_rsp (struct iscsi_cmnd *cmnd);
static inline int handle_logout_rsp (struct iscsi_cmnd *cmnd);
static inline int handle_nopin (struct iscsi_cmnd *cmnd);
static inline int handle_iscsi_done (struct iscsi_cmnd *cmnd, int startsn, int endsn);
static inline int handle_iscsi_mgt_fn_done(struct iscsi_cmnd *cmnd);
static inline int iscsi_dequeue (struct iscsi_cmnd **cmnd_pptr);
static inline int iscsi_tx_rjt (struct iscsi_conn *conn, __u8 reason);
static inline int iscsi_tx_login_rsp (struct iscsi_conn *conn, __u8 status_class, __u8 status_detail);
static inline int iscsi_tx_r2t (struct iscsi_cmnd *cmnd);

static int find_iovec_needed (unsigned int data_len, struct scatterlist *st_list, unsigned int offset);
static int fill_iovec (struct iovec *iov, int niov, struct scatterlist *st_list, int *offset, unsigned int data);

/* Added for handling SNACK requests - SAI */
static inline int handle_snack(struct iscsi_conn *conn, unsigned char *buffer);
static inline int send_iscsi_response(struct iscsi_cmnd *cmnd);
    
#define PDU_SENSE_LENGTH_SIZE 2
#define SENSE_STRUCTURE_SIZE 18
				
#endif
