/******************************************************************************
* 	ref13_06/target/sep_target_module.c
*
* 	This file must be included by every front-end target driver that is
* 	trying to register with the mid-level. This is the common portion
* 	that gets included with all the low-level drivers.
* 
* 	To use this the low-level driver must define the variable my_template
* 	of the type Scsi_Target_Template with all the needed functions and
* 	values defined in scsi_target.h
*
*   IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
*   By downloading, copying, installing or using the software you agree to this
*   license. If you do not agree to this license, do not download, install,
*   copy or use the software.
*
*   License Agreement
*
*   Copyright (C) 2001  Inter-Operability Lab.(IOL),Univ. of New Hampshire(UNH),
*                       Durham, NH 03824
*
*   Redistribution and use in source and binary forms, with or without
*   modification, are permitted provided that the following conditions are met:
*
*   -Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
*
*   -Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
*
*   The name of IOL/UNH may not be used to endorse or promote products derived
*   from this software without specific prior written permission.
*
*   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
*   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IOL,UNH OR CONTRIBUTORS
*   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF
*   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*   POSSIBILITY OF SUCH DAMAGE.
******************************************************************************/	

# include <linux/module.h>
# include <linux/init.h>
# include <linux/kernel.h>
# include "scsi_target.h"

int scsi_target_module_init (void)
{

	debugInit ("scsi_target_module_init: Initializing: %s\n", my_template.name);
	if (register_target_template (&my_template) < 0)
		return (-1);

	return (0);
}


void scsi_target_module_cleanup (void)
{

#ifdef CONFIG_PROC_FS
    /* Don't show the /proc/scsi_target files */
    remove_proc_entry( "scsi_target/scsi_target", 0 );
    remove_proc_entry( "scsi_target", 0 );
#endif

	if (deregister_target_template (&my_template) < 0)
		printk ("Error in the deregistration process\n");

	debugInit ("scsi_target_module_cleanup exit\n");
}


module_init (scsi_target_module_init);
module_exit (scsi_target_module_cleanup);
