/*****************************************************************************
*
*	ref18_01/initiator/iscsi_initiator.h
*
*	This file contains all the required header files, struct definitions and
*	constants for iscsi initiator code.
*
* 	IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING. 
* 	By downloading, copying, installing or using the software you agree to this 
* 	license. If you do not agree to this license, do not download, install,
* 	copy or use the software.
*
* 	License Agreement
*
*   Copyright (C) 2001  Inter-Operability Lab.(IOL),Univ. of New Hampshire(UNH),
*                       Durham, NH 03824
*
* 	Redistribution and use in source and binary forms, with or without 
*	modification, are permitted provided that the following conditions are met:
*
* 	-Redistributions of source code must retain the above copyright notice, 
*	this list of conditions and the following disclaimer.
*
*	-Redistributions in binary form must reproduce the above copyright notice, 
*	this list of conditions and the following disclaimer in the documentation 
*	and/or other materials provided with the distribution.
*
*	The name of IOL/UNH may not be used to endorse or promote products derived 
* 	from this software without specific prior written permission.
*
*	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* 	PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IOL,UNH OR CONTRIBUTORS
*	BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* 	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF 
* 	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*	POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/

#include <linux/blk.h>
#include <linux/kernel.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/unistd.h>
#include <linux/spinlock.h>
#include <asm/io.h>   
#include <asm/irq.h>
#include <linux/smp_lock.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <asm/uaccess.h>

#include <net/sock.h>
#include <net/tcp.h>
#include "sd.h"
#include "hosts.h"
#include <linux/module.h>

#include "../iscsi-unh-common/iscsi_common.h"
#include "../iscsi-unh-common/crc.h"
#include "../iscsi-unh-common/text_param.h"
#include "../iscsi-unh-common/initiator_negotiate.h"

#ifndef _ISCSI_INITIATOR_H
#define _ISCSI_INITIATOR_H

/*	the name of our module (driver, HBA) */
#define	OUR_NAME	"UNH-IOL iSCSI reference initiator"

/*	maximum number of iov slots iscsi_initiator can set up for I/O	*/
#define MAX_IOV_SLOTS		68

/*	maximum number of simultaneously outstanding commands per LUN */\
#define MAX_COMMANDS       	8

/*	maximum number of targets this iscsi_initiator host will register for */
#define MAX_TARGET_IDS		4

/*	maximum number of luns each target can handle */
#define MAX_LUNS			8

/*  Multiple connections Round Robin Scheduling Scheme - SAI */
#define CONN_SCHED_RR           1
#define CONN_SCHED_LUN          2

/* Add Maximum Connections allowed per session - SAI */
#define MAX_CONNECTIONS         20

int iscsi_initiator_proc_info(char *, char **, off_t, int, int, int);
int iscsi_initiator_detect(Scsi_Host_Template *);
int iscsi_initiator_release(struct Scsi_Host *);
const char * iscsi_initiator_info(struct Scsi_Host *);
int iscsi_initiator_queuecommand(Scsi_Cmnd *, void (* done)(Scsi_Cmnd *));
int iscsi_initiator_abort(Scsi_Cmnd *);
int iscsi_initiator_reset(Scsi_Cmnd *);


/* Moved from iscsi_initiator.c - SAI */
/*      holds pointer to private data part of Scsi_Host structure */
/*      this is set up once, by iscsi_detect()
        which is called when the initiator module is loaded */
static struct iscsi_hostdata *global_hostdata = NULL;

/*      holds pointer to the registered Scsi_Host */
struct Scsi_Host *global_host = NULL;


/*	initializor for static Scsi_Host_Template struct as defined in
			/usr/src/linux/drivers/scsi/hosts.h
	which has good comments about each field.

	There is one such struct for each different type of host adaptor,
	and since there is only one iscsi type, we only need 1 static instance.
	(This is normal for each different adaptor type.)
*/

#define ISCSI_INITIATOR {\
		proc_info:				iscsi_initiator_proc_info,\
		name:					OUR_NAME,\
        detect:                 iscsi_initiator_detect,\
        release:                iscsi_initiator_release,\
        info:                   NULL, /*iscsi_initiator_info,*/\
		ioctl:					NULL,\
        queuecommand:           iscsi_initiator_queuecommand,\
		eh_strategy_handler:	NULL,\
        eh_abort_handler:       iscsi_initiator_abort,\
        eh_device_reset_handler:iscsi_initiator_reset,\
        eh_bus_reset_handler:	NULL,\
        eh_host_reset_handler:	NULL,\
		/* use default scsicam_bios_param() in linux/drivers/scsi/scsicam.c */\
        bios_param:             NULL,\
		/* max no. of simultaneously active SCSI commands driver can accept */\
        can_queue:              1,\
        this_id:                -1, /* our host has no id on the SCSI bus */\
		/* max no. of slots driver can accept in a scatter-gather list */\
        sg_tablesize:           MAX_IOV_SLOTS-4,\
		/* max no. of simultaneously outstanding commands per LUN */\
        cmd_per_lun:            MAX_COMMANDS,\
		/* no. of cards of this type found on this machine */\
        present:                0,\
        unchecked_isa_dma:      0,\
		/* mid-level should merge requests for adjacent blocks of memory */\
        use_clustering:         ENABLE_CLUSTERING,\
		use_new_eh_code:		1,\
		emulated:				0,\
		/*	proc_name xxx becomes the name for the directory /proc/scsi/xxx */\
		proc_name:				"iscsi_initiator"\
}

#define STATUS_NOT_PRESENT	0x0
#define STATUS_CONNECTED	0x1
#define STATUS_LOGGED_IN	0x2
#define STATUS_DISCONNECTED	0x3

#define TX_RESTART_PERIOD	10	/* no. of ticks tx_thread waits after EAGAIN */

#define WHITE_SPACE " \t\v\f\n\r"

/*	Holds information to control out-of-order sequences and/or data pdus */
struct order_range
	{
	__u32				offset;			/* base offset of range */
	__u32				limit;			/* == offset + length of range */
	struct order_range	*next;			/* next range element in list */
	};


/*	Holds information extracted from a real or implied R2T from target.
	Used by initiator to control sending DataOut PDUs to target.
*/
struct r2t_cookie
	{
    __u32   target_xfer_tag;	/*opaque -- in network order (from target r2t)*/
	__u32   offset;				/* SCSI buffer offset of next InOrder Data Out*/
								/* but is limit of next Out-of-Order Data Out */
    __u32   r2t_xfer_length;	/* no. of bytes left to send in r2t sequence  */
								/* or no. received so far in DataIn sequence  */
	__u32	data_out_sn;		/* incremented from 0 by DataOuts for this r2t*/
	struct r2t_cookie	*next;
	};


struct command
    {
    __u8    task_mgt_response;
    __u8    task_mgt_function;
	__u8	header_opcode;		/* copy of opcode when sending a pdu */
	__u8	header_flags;		/* copy of flags when sending a pdu */
    __u32   init_task_tag;
	__u32   data_offset;		/* offset to next data to receive or r2t */
    __u32   recvd_length;		/* bytes received so far in DataIn sequence */
    __u32   tx_size;			/* no. of bytes of data to send in 1 sendmsg */
	__u32	data_in_sn;			/* incremented from 0 by DataIns for this cmd */
	struct r2t_cookie	*first_cookie, *last_cookie;	/* outstanding r2ts */
	struct order_range	r2t_range_list;		/* for out-of-order r2ts */
	struct order_range	pdu_range_list;		/* for out-of-order DataIn pdus */
	__u32	n_r2t_cookies;		/* no. of real (not implied) r2ts in list */
	__u32	r2t_sn;				/* incremented from 0 by r2ts */
	__u32	data_checksum;		/* holds data checksum for attached data */
	void	*buffer_to_free;	/* free this buffer after xmit (if not NULL) */
    __u32   ref_task_tag;       /* TASK_REASSIGN to work - SAI */
	__u8   snack_flags;                /* Command response and SNACK reject bits - SAI */
	__u32   time_stamp;             /* Stores the time when command was transmitted - SAI */
	__u8    activity_flg;       /* Stores Activity from target for a command - SAI */
	__u32   retransmit_tx_size;     /* Stores the first comamnd tx_size - SAI */
    Scsi_Cmnd *SCpnt;			/* pointer to SCSI command */
    struct iscsi_init_scsi_data_out data;	/* DataOut PDU header built here */
    struct iovec iov[MAX_IOV_SLOTS];		/* socket iovector built here */
    struct msghdr tx_msghdr;				/* socket message headr built here*/
    struct iscsi_init_scsi_cmnd iscsi_cmd;	/* SCSI Command PDU headr blt here*/
    struct command * next;
    };


/*	one of these per connection */
struct connection
    {
    __u32	target_address;			/* 32-bit ip address of target */
    __u32	target_port;			/* 16-bit port number of target */
	__u32	connection_id;			/* CID within session */
	__u32	max_send_length;		/* MaxRecvDataSegmentLength from target */
	__u32	max_recv_length;		/* initiator's MaxRecvDataSegmentLength */
	__u32	nop_counter;			/* counter used in NopOut ping data */
    __u32	exp_stat_sn;			/* next StatSN expected from target */
    struct session *my_session;		/* pointer to session this conn belongs to*/
	__u32	connection_flags;		/* bits for header/data digests, activity */
    __u32	connection_state;		/*	STATUS_NOT_PRESENT,
										STATUS_CONNECTED,
										STATUS_LOGGED_IN */
    struct semaphore rx_done_sem;	/* wait here for rx_thread to start/stop */

	/* the following field is manipulated by add_(remove_)session only */
    struct connection	*next;		/* list of connections in this session */

    struct socket		*sock;		/* pointer to open socket to target */
    __u8  				*rx_buf;	/* PDU headers from target read into here */

	/* the following 1 field is only set up after login succeeds */
    struct task_struct	*rx_thread;		/* pointer to TCB of tx_thread */

	int			rec_tests;		/* Indicator to initiate conn. recovery - SAI */

	/* the following 3 fields are used by the rx_thread during operation */
    struct command *pending_commands;	/* SCSI commands active on this conn */
	struct msghdr	rx_msg;			/* header for sock_recvmsg on this conn */
	struct iovec	rx_iov;			/* 1-slot vector for sock_recvmsg on conn*/
	};


/*	one of these structures for each initiator session */
struct session
    {
    __u32   scsi_target_id;				/* target in [host,bus,target,lun] */
	__u32	lun_bits;					/*bits for lun in[host,bus,target,lun]*/
    __u32   session_state;				/*	STATUS_NOT_PRESENT,
											STATUS_LOGGED_IN */
	__u32	cur_task_tag;				/* generator for unique ITTs */
    __u32   cur_cmd_sn;					/* generator for sequential CmdSNs */
    __u32   exp_cmd_sn;					/* last ExpCmdSN from target */
    __u32   max_cmd_sn;					/* last MaxCmdSN from target */
    struct semaphore	tx_sem;			/* tx_thread waits here for work to do*/
    struct semaphore	tx_done_sem;	/* wait here for tx_thread start/stop */
    struct semaphore	task_mgt_sem;	/* wait here for task mgt resp/timeout*/
    struct semaphore    logout_resp_sem;        /* wait here logout resp - SAI*/

	/* the following 5 fields are initialized from configured hostdata */
	__u32	session_flags;				/* flags for session-wide control */
										/* only flag so far is NEED_TX_WAKEUP */
	__u32	init_conn_flags;			/* initial connection-specific flags */
	__u32	nop_period;					/* no. of secs between sending NopOuts*/
	__u32	draft_number;				/* number of the draft to enforce */
	struct parameter_type (*session_params)[MAX_CONFIG_PARAMS];

	/* the following 3 fields are manipulated by add_(remove_)session only */
	__u32	nconnections;				/* no. of connections in next list */
    struct connection *connection_head;	/* list of connections in this session*/

    struct session *next;				/* for linked-list of sessions */

	/* the following 2 fields are allocated late */
    struct session_operational_parameters *oper_param;
	struct timer_list	*tx_timer;		/* timer for sending nops */

	/* the following field is returned by target if first login succeeds */
	__u32	tsih;						/* target's handle for this session */

	__u8    init_snack_flg;                         /* Store SNACK Related flags here - SAI*/
	struct timer_list       *retran_timer;  /* required for command retransmissions - SAI */
	int                     xmit_chk_interval;      /* stores the retransmission check interval in msec - SAI */
	int                     retran_period;          /* stores the retransmission threshold in msec - SAI */
	struct timer_list       *connrec_timer; /* Time2wait timer - SAI */
	struct connection		*rr_conn_head; /* round robin connection head - SAI */
	int         sched_scheme;       /* stores the connection scheduling scheme to be used */

	int 		n_rec_tasks;		/* stores the no. of tasks that were transfered from a failing connection - SAI */

	/* the following 1 field is only set up after first login succeeds */
    struct task_struct	*tx_thread;		/* pointer to TCB of tx_thread */
    };


/*	One of these structures forms the private data part for each
	iscsi adaptor.  Since we have only 1 such adaptor, we will
	instantiate this structure only once, when the iscsi_initiator
	module is first loaded.  At that time the variable global_hostdata
	will be set to point to the instantiation, and the fields will be
	set up to the defaults (all 0 except for param_tbl, which gets a copy
	of config_params[]).  After that, whenever the user changes parameter
	values using iscsi_manage, those changes are made to the values in this
	structure.
*/
struct iscsi_hostdata
	{
	unsigned int			nsessions;		/* no. of targets/sessions so far */
	unsigned int			nop_period;		/* timer value to force NopOuts */
	unsigned int			draft_number;	/* number of draft to enforce */
	unsigned int			init_sess_flags;/* initial session-wide flags */
	unsigned int			init_conn_flags;/*initial connection-specific flgs*/
	struct session			*session_head;	/* list of struct session */
	int						init_sessrec_flg; /* indicates if sess recovery is in progress - SAI */
	struct semaphore		init_sessrec_sem; /* puts scsi to sleep in sess recovery - SAI */
	__u8                    init_snack_flg; /* Store SNACK Related flags here - SAI*/
	struct command          *pending_commands; /* Indicate Session recovery to scsi - SAI */
	int                     xmit_chk_interval;  /* stores the retransmission check interval in msec - SAI */
	int                     retran_period;  /* stores the retransmission threshold in msec - SAI */
	int             		sched_scheme; /* indicates scheduling algorithm to be used with multiple connections - SAI*/
    struct parameter_type	(*param_tbl)[MAX_CONFIG_PARAMS];
/* chap support -CHONG */
	struct chap_parameter_type chap_parameter;
    /* recovery session list - new 18_04 SAI */
    struct session          *rec_sess_head;
	};


/* Add the Error Recovery Structure - SAI */

struct init_error_rec
        {
        struct connection       *curr_conn;
        struct  command         *related_cmd;
        __u8                    err_type;
        };

void iscsi_initiator_tx_thread(void *);
void iscsi_initiator_rx_thread(void *);
int iscsi_initiator_login( struct connection *current_connection,
						unsigned int when_called,
						struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
						struct chap_parameter_type chap_param,
						unsigned int *redirect_ip, unsigned int *redirect_port);
int rx_data( struct connection *current_connection, unsigned int local_itt,
						struct command *related_command, unsigned int size );
int rx_rsp( struct connection * current_connection, unsigned int local_itt,
			struct command *related_command, unsigned int size, char *buffer );
int rx_r2t( struct connection * current_connection, unsigned int local_itt,
										struct command *related_command);
int rx_async_msg( struct connection *current_connection, unsigned int size,
																char *buffer );
int rx_rjt( struct connection * current_connection );
int rx_nopin( struct connection * current_connection, unsigned int local_itt,
			struct command *related_command, unsigned int size, char **buffer );
int rx_text_rsp( struct connection * current_connection, unsigned int local_itt,
			struct command *related_command, unsigned int size, char *buffer );
int rx_logout_rsp( struct connection * current_connection,
					unsigned int local_itt, struct command *related_command);
int rx_task_mgt_rsp( struct connection *current_connection,
					unsigned int local_itt, struct command *related_command);
int scsi_sendmsg(struct socket *, struct msghdr *, int);
int set_up_unsolicited_data_pdu(struct connection *related_connection, 
										struct command *related_command);
int set_up_subsequent_data_pdu(struct connection *related_connection,
                                            struct command *related_command);
void close_session(struct session *sess);

int add_session_to_hostdata(struct connection *new_connection, unsigned int lun,
						struct session *new_session, struct Scsi_Host *host);
int remove_session_from_hostdata( __u32 target, __u32 lun, __u32 cid,
											struct iscsi_hostdata *hostdata );
/* Declaration needed as these functions used by Error Recovery - SAI */

struct connection* get_connection(struct session *sess);
int drive_logout(struct session *my_session, int cid, int reason);
int create_session( __u32 target, __u32 ip, __u32 port, __u32 lun, __u32 cid, struct Scsi_Host *host );
void close_connection(struct connection *conn);
void create_rec_conn(struct connection *conn);
int find_connection( unsigned int target, unsigned int cid,
                        struct iscsi_hostdata *hostdata,
                        struct connection **old_connection,
                        struct session **old_session );
void clear_free_conn(struct session *sess);

int init_session( struct session *sess, struct connection *conn,
    unsigned int lun, struct parameter_type param_tbl[MAX_CONFIG_PARAMS],
	struct chap_parameter_type chap_param, 
    unsigned int *redirect_ip, unsigned int *redirect_port );

#endif /* _ISCSI_INITIATOR_H */
